// ==============================================================
//                 ORBITER MODULE: slc40abort
//
// slc40abort.cpp
// Control module for slc40abort vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PB_FUELMASS = 750.0;
const double PB_ISP = 5e4;
const double PB_MAXHOVERTH = 0;


class slc40abort: public VESSEL3 {
public:
	slc40abort (OBJHANDLE hVessel, int flightmodel)
		: VESSEL3 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostCreation();


	THRUSTER_HANDLE th_water[2], th_group[2];
	ATTACHMENTHANDLE fh_attach;
	PROPELLANT_HANDLE ph_vernier;

private:

};

// ==============================================================
// slc40abort class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void slc40abort::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical specs
	SetEmptyMass (1000000);
	SetCW (1, 1, 1, 1);
	SetSize (200);
	SetCrossSections (_V(1000,1000,10000));
	SetPMI (_V(100,100,100));
	SetTouchdownPoints (_V(0,-0.01,50), _V(-100,-0.01,-50), _V(100,-0.01,-50));
	SetRotDrag (_V(1,1,1));

	SetMeshVisibilityMode ((AddMesh ("DM2/slc40Abort")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(0,10,-20));

	ph_vernier = CreatePropellantResource (PB_FUELMASS,PB_FUELMASS);

	PARTICLESTREAMSPEC contrail_water = {
		0, 0.3, 10, 6, 0.1, 0.75, 2, 0.25, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	

	th_water[0] = CreateThruster (_V(0,0,0), _V(0.866,0.5,0), PB_MAXHOVERTH, ph_vernier, PB_ISP);
	AddExhaustStream (th_water[0],_V( 7.75,16.8, 5), &contrail_water);
	AddExhaustStream (th_water[0],_V( 7.75,16.8, 2.5), &contrail_water);
	AddExhaustStream (th_water[0],_V( 7.75,16.8, 0), &contrail_water);
	AddExhaustStream (th_water[0],_V( 7.75,16.8,-2.5), &contrail_water);
	AddExhaustStream (th_water[0],_V( 7.75,16.8,-5), &contrail_water);


	th_water[1] = CreateThruster (_V(0,0,0), _V(-0.866,0.5,0), PB_MAXHOVERTH, ph_vernier, PB_ISP);
	AddExhaustStream (th_water[1],_V(-7.15,16.8, 5), &contrail_water);
	AddExhaustStream (th_water[1],_V(-7.15,16.8, 2.5), &contrail_water);
	AddExhaustStream (th_water[1],_V(-7.15,16.8, 0), &contrail_water);
	AddExhaustStream (th_water[1],_V(-7.15,16.8,-2.5), &contrail_water);
	AddExhaustStream (th_water[1],_V(-7.15,16.8,-5), &contrail_water);

	th_group[0] = th_water[0];
	th_group[1] = th_water[1];
	CreateThrusterGroup (th_group, 2, THGROUP_MAIN);

	fh_attach = CreateAttachment (false, _V( -0.01, 14.915, -0.127), _V( 0,1,0), _V(0,0,-1), "DR");


}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new slc40abort (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (slc40abort*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void slc40abort::clbkPostCreation ()
{
	SetPropellantMass(ph_vernier, PB_FUELMASS);
}

void slc40abort::clbkPreStep(double SimT, double SimDT, double MJD) {
	// next block of code
	ENGINESTATUS es;
	es.main = 0;
	
	OBJHANDLE hDragon = GetAttachmentStatus(fh_attach);
	if(hDragon)
	{
		oapiGetEngineStatus(hDragon, &es);
		if (es.main > 0.95)
		{
			DetachChild(fh_attach, 0.0);
			SetThrusterGroupLevel(THGROUP_MAIN, 0);
			
		}

	}
	else SetThrusterGroupLevel(THGROUP_MAIN, 0);


	}
