// ==============================================================
//                 ORBITER MODULE: CrewDragonMFD
//                  Part of the ORBITER SDK
//           Copyright (C) 2002-2003 Martin Schweiger
//                   All rights reserved
//
// CrewDragonMFD.cpp
// Reference implementation of generic user-defined MFD mode
// ==============================================================

#define STRICT
#define ORBITER_MODULE
#include <windows.h>
#include <stdio.h>
#include <cmath>
#include "orbitersdk.h"
#include "CrewDragonMFD.h"

// ==============================================================
// Global variables

VECTOR3 flight_vector = _V(0,0,1);
double steptime = 0;
double mu = 398600439968871.2;
double rt = 6371010;
VECTOR3 GBOD_AXIS = _V(0,0,0);
double earth_rot_rate = PI2/86164;
ELEMENTS el;
ORBITPARAM prm;
int status_data;
int add_orbits;
int add_orbits_sw;
VESSEL2* vme;
char mfdtgtname[256];
char vesname[256];
char gbodyname[256];
OBJHANDLE hTgt;
OBJHANDLE MyVes;
OBJHANDLE gBod;

double tgt_dist = 0;
double tgt_arc = 0;
double req_inc = 0;
double flt_time = 0;
double crossrange_error = 0;
double tgt_hdg = 0;
double req_hdg = 0;
double curr_hdg = 0;
double hud_crossrange_error = 0;
	
double bank_angle = 0;
int auto_reentry_sw = 0;
//int bank_phase_sw = 0;
//int bank_0_sw = 0;
int ttb_check_sw = 0;
double bank0_lim = 0.25*RAD;
double bank_timer = 0;
double new_bank_angle = 0;

double lambda_2 = 0;
double arc_to_burn = 0;
int auto_burn_sw = 0;
int burn_sw = 0;
double time_to_burn = 0;
double v_2 = 0;
double v_2pb = 0;
double ei_radius = 0;
double hud_range = 0;

//int data_sw = 0;
//double intv = 200000;
//double range_0 = 5000000;
double post_step = 0;
//double tot_energy_old = 0;
//double dst_past_tld_old = 0;
//double bank_correction = 0;
//double digi_dar = 0;
//double digi_dar_old = 0;
//int digi_dar_check_sw = 0;


double r_3 = 6466000;
double r_3_fact = 0;
double dv_remaining = 0;
double dv_required = 0;
int display_sw = 0;
int button_sw = 0;
const DWORD COMM_SEND_STS = 2306;
const DWORD COMM_SEND_STS2 = 2308;
bool ves_is_dragon = false;
bool has_parent = false;

int bank_dir_sw = 1;
int	bank_mode_sw = 0;
double old_vSpeed = 0;
double	old_hSpeed = 0;
double	vSpeed_adjust = 0;
double deadband = (1*RAD);
double deadband2 = (1*RAD);

int reentry_mode = 0;
double hud_omega = 0;
double hud_lambda = 0;

static struct {  // "Ascent MFD" parameters
	int mode;      // identifier for new MFD mode
	
} g_CrewDragonMFD;

CrewDragonMFD::SavePrm CrewDragonMFD::saveprm = {0,0,0,0,0,0};

// ==============================================================
// API interface



DLLCLBK void InitModule (HINSTANCE hDLL)
{
	
	static char *name = "CrewDragonMFD";
	MFDMODESPEC spec;
	spec.name    = name;
	spec.key     = OAPI_KEY_P;
//	spec.context = NULL;
	spec.msgproc = CrewDragonMFD::MsgProc;
	g_CrewDragonMFD.mode = oapiRegisterMFDMode (spec);
	/////////////////////////////////////
	//sprintf(oapiDebugString(), "%d", spec.msgproc);

	
}


DLLCLBK void ExitModule (HINSTANCE hDLL)
{
	oapiUnregisterMFDMode (g_CrewDragonMFD.mode);
}

DLLCLBK void opcCloseRenderViewport() // called before StoreStatus()
{

	CrewDragonMFD::saveprm.sp1 = 0;
	CrewDragonMFD::saveprm.sp2 = 0;
	CrewDragonMFD::saveprm.sp3 = 0;
	CrewDragonMFD::saveprm.sp4 = 0;
	CrewDragonMFD::saveprm.sp5 = 0;
	CrewDragonMFD::saveprm.sp6 = 0;

	strcpy(mfdtgtname,"NULL");
	strcpy(vesname,"NULL");
	strcpy(gbodyname,"NULL");

//	bank_phase_sw = 0;
//	bank_0_sw = 0;
	auto_reentry_sw = 0;
    auto_burn_sw = 0;
	burn_sw = 0;
	add_orbits = 0;
	tgt_dist = 0;
	tgt_arc = 0;
	req_inc = 0;
	flt_time = 0;
	crossrange_error = 0;
	tgt_hdg = 0;
	req_hdg = 0;
	curr_hdg = 0;
	hud_crossrange_error = 0;
	dv_remaining = 0;
	dv_required = 0;
	time_to_burn = 0;
	hud_range = 0;
	add_orbits_sw = 0;
	status_data = -1;
	ttb_check_sw = 0;
	bank0_lim = 0.25*RAD;
	display_sw = 0;
	button_sw = 0;

	ves_is_dragon = false;
	has_parent = false;

	MyVes = NULL;
	gBod = NULL;
	hTgt = NULL;

	bank_dir_sw = 1;
	bank_mode_sw = 0;
	old_vSpeed = 0;
	old_hSpeed = 0;
	vSpeed_adjust = 0;

	reentry_mode = 0;
}


DLLCLBK void opcPostStep (double simt, double simdt, double mjd)
{
post_step = simdt;
}

DLLCLBK void opcPreStep (double SimT, double SimDT, double mjd)
{
	
	
	/////////// init vessel /////////////////////
	//if (strlen(vesname) == 0)
	//{
	//MyVes = oapiGetFocusObject();
	//oapiGetObjectName(MyVes, vesname, 30);
	//}
	//else
	//{
	//MyVes = oapiGetVesselByName(vesname);
	//}
	//if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	//else
	//{
	//MyVes = oapiGetFocusObject();
	//oapiGetObjectName(MyVes, vesname, 30);
	//if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	//}
	

	
///////////// init tgt gbody ////////////////////
if(MyVes)
{

	if (strlen(gbodyname) == 0)
	{
	gBod = vme->GetGravityRef();
	oapiGetObjectName(gBod, gbodyname, 30);
	CrewDragonMFD::GetGbodData();
	}
	else if (!strnicmp (gbodyname, "NULL", 4))
	{
	gBod = vme->GetGravityRef();
	oapiGetObjectName(gBod, gbodyname, 30);
	CrewDragonMFD::GetGbodData();
	}
	else// if(!gBod)
	{
	gBod = oapiGetObjectByName(gbodyname);
	if(!gBod)
		{
		gBod = vme->GetGravityRef();
		oapiGetObjectName(gBod, gbodyname, 30);
		}
	CrewDragonMFD::GetGbodData();
	}

	
///////////// init tgt //////////////////////

	if (strlen(mfdtgtname) == 0) 
	{
		hTgt = NULL;
		strcpy(mfdtgtname,"NULL");
	}
	else if (!strnicmp (mfdtgtname, "NULL", 4)) hTgt = NULL;
	else
	{
	hTgt = oapiGetVesselByName(mfdtgtname);
	if(hTgt == NULL) hTgt = oapiGetBaseByName(gBod, mfdtgtname);
//	sprintf(oapiDebugString(), "tgt %s", mfdtgtname);
	if(hTgt == NULL) strcpy(mfdtgtname,"NULL");
	}

////////////////////////////////////////////////////////////////////////////


	steptime = SimDT;

//char class_name[256];
//strcpy(class_name, vme->GetClassName());
//sprintf(oapiDebugString(), "%s", class_name);
//if (!strnicmp (class_name, "dm2", 3))
//	if(ves_is_dragon == false)
//	{
//	int drg_chk = 0;
//	drg_chk = vme->SendBufferedKey(COMM_SEND_STS, true, NULL);
//	if(drg_chk > 1)ves_is_dragon = true;
//	else ves_is_dragon = false;
//	sprintf(oapiDebugString(), "sts %d", drg_chk);
//	}

if(ves_is_dragon == true)
{
//	ves_is_dragon = true;

	
	CrewDragonMFD::DoNav();

	if(has_parent == false)
	{

	/////////////// get dv remaining ///////////////////////
	PROPELLANT_HANDLE hProp = vme->GetPropellantHandleByIndex(0);
	double prop_mass = vme->GetPropellantMass(hProp);
	double tot_mass = vme->GetMass();
	THRUSTER_HANDLE hThruster = vme->GetThrusterHandleByIndex(0);
	double my_isp = vme->GetThrusterIsp(hThruster);
	double mass_fraction = (tot_mass/(tot_mass - prop_mass));
	dv_remaining = my_isp*(log(mass_fraction));
	
	//////////////// do auto reentry
if(auto_burn_sw < 2 && reentry_mode == 0)CrewDragonMFD::DoReentryBurn();

if(auto_burn_sw == 1 && reentry_mode == 0)
		{
		if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);
		//CrewDragonMFD::DoReentryBurn();
		CrewDragonMFD::GoToFlightVector();
		CrewDragonMFD::RollToBankAngle();
		if(auto_burn_sw == 2)CrewDragonMFD::KillAllRCS ();
		}

else if(auto_reentry_sw == 1)
	{
	//add_orbits = 0;
	if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);
	CrewDragonMFD::CalculateFlightVectorReentry();
//	if (auto_reentry_sw == 2)
//		{
//		CrewDragonMFD::CalculateBankNoTarget();
//		CrewDragonMFD::CalculateFlightVector();
//		}
	CrewDragonMFD::GoToFlightVectorReentry();
	//CrewDragonMFD::RollToBankAngle();
	if(vme->GetAltitude() < 15000) 
		{
		CrewDragonMFD::KillAllRCS ();
		auto_reentry_sw = 3;
		}
	}
else if(auto_reentry_sw == 2)
	{
	//add_orbits = 0;
	if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);
//	CrewDragonMFD::CalculateBankNoTarget();
	CrewDragonMFD::CalculateFlightVectorReentryNT();	
	CrewDragonMFD::GoToFlightVectorReentry();
	if(vme->GetAltitude() < 15000) 
		{
		CrewDragonMFD::KillAllRCS ();
		auto_reentry_sw = 3;
		}
	}

	}
}
else 
	{
	ves_is_dragon = false;
//	bank_phase_sw = 0;
//	bank_0_sw = 0;
	auto_reentry_sw = 0;
    auto_burn_sw = 0;
	burn_sw = 0;
	add_orbits = 0;
	tgt_dist = 0;
	tgt_arc = 0;
	req_inc = 0;
	flt_time = 0;
	crossrange_error = 0;
	tgt_hdg = 0;
	req_hdg = 0;
	curr_hdg = 0;
	hud_crossrange_error = 0;
	dv_remaining = 0;
	dv_required = 0;
	time_to_burn = 0;
	hud_range = 0;
	add_orbits_sw = 0;
	//status_data = -1;
	ttb_check_sw = 0;
	bank0_lim = 0.25*RAD;
	reentry_mode = 0;
	}
}
else ves_is_dragon = false;
}

//void  CrewDragonMFD::CalculateFlightVector()
//{
//	VECTOR3 haspdv = _V(0,0,0);
//	vme->GetGroundspeedVector(FRAME_HORIZON,haspdv);
//	haspdv = -haspdv;
//	normalise(haspdv);
//
//	VECTOR3 wv = _V(0,-1,0);
//	//GetWeightVector(wv);
//	//normalise(wv);
//
//	VECTOR3 vel_wv_nml = crossp(wv, haspdv);
//	normalise(vel_wv_nml);
//	double rotangle = 10*RAD;
//	MATRIX3 rm = rotm(vel_wv_nml, rotangle);
//	VECTOR3 fv_hrz = mul(rm,haspdv);
//	rm = rotm(haspdv, bank_angle);
//	VECTOR3 fv_hrz_banked = mul(rm,fv_hrz);
//
//	vme->HorizonInvRot(fv_hrz_banked,flight_vector);
//}

void  CrewDragonMFD::GoToFlightVector()
{
	vme->DeactivateNavmode(NAVMODE_KILLROT);
	vme->DeactivateNavmode(NAVMODE_HLEVEL);
	vme->DeactivateNavmode(NAVMODE_PROGRADE);
	vme->DeactivateNavmode(NAVMODE_RETROGRADE);
	vme->DeactivateNavmode(NAVMODE_NORMAL);
	vme->DeactivateNavmode(NAVMODE_ANTINORMAL);
	vme->DeactivateNavmode(NAVMODE_HOLDALT);

	double vrot_lim = 5*RAD;

	double mass = vme->GetMass();
	VECTOR3 pmi = _V(0,0,0);
	vme->GetPMI(pmi);

	THRUSTER_HANDLE thrcs = vme->GetThrusterHandleByIndex(0);
	double rcs_th = vme->GetThrusterMax(thrcs);
	double max_torque_x = 2*rcs_th;
	double max_torque_y = 2*rcs_th;
	double max_torque_z = 4*rcs_th;

	double angoa = atan2(-flight_vector.y, flight_vector.z);
	double slipage = atan2(flight_vector.x, flight_vector.z);

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double inertia_x = pmi.x*mass;
	double inertia_y = pmi.y*mass;
	
	VECTOR3 avel = _V(0,0,0);
	vme->GetAngularVel(avel);
	double max_acc_x = max_torque_x/inertia_x;
	double max_drot_x = max_acc_x*steptime;
	double max_acc_y = max_torque_y/inertia_y;
	double max_drot_y = max_acc_y*steptime;

	double reqxv = 1*abs(angoa); //1
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot_x;
	if((angoa*DEG) < 0.5 && (angoa*DEG) > -0.5) thx_lvl = 0.25*thx_lvl;

	double reqyv = 1*abs(slipage);//1
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot_y;
	if((slipage*DEG) < 0.5 && (slipage*DEG) > -0.5) thy_lvl = 0.25*thy_lvl;
	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));


	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;


	if(thx_lvl > 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}


	if(thy_lvl > 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}


}

void  CrewDragonMFD::GoToFlightVectorReentry()
{
	vme->DeactivateNavmode(NAVMODE_KILLROT);
	vme->DeactivateNavmode(NAVMODE_HLEVEL);
	vme->DeactivateNavmode(NAVMODE_PROGRADE);
	vme->DeactivateNavmode(NAVMODE_RETROGRADE);
	vme->DeactivateNavmode(NAVMODE_NORMAL);
	vme->DeactivateNavmode(NAVMODE_ANTINORMAL);
	vme->DeactivateNavmode(NAVMODE_HOLDALT);

	double vrot_lim = 20*RAD;
	if(vme->GetAltitude() > 120000) vrot_lim = 10*RAD;

	VECTOR3 fv_unit = flight_vector;
	normalise(fv_unit);

	double mass = vme->GetMass();
	VECTOR3 pmi = _V(0,0,0);
	vme->GetPMI(pmi);

	THRUSTER_HANDLE thrcs = vme->GetThrusterHandleByIndex(0);
	double rcs_th = vme->GetThrusterMax(thrcs);
	double max_torque_x = 2*rcs_th;
	double max_torque_y = 2*rcs_th;
	double max_torque_z = 4*rcs_th;

	double inertia_x = pmi.x*mass;
	double inertia_y = pmi.y*mass;
	double inertia_z = pmi.z*mass;
	
	VECTOR3 avel = _V(0,0,0);
	vme->GetAngularVel(avel);

	double max_acc_x = max_torque_x/inertia_x;
	double max_drot_x = max_acc_x*post_step*2; // 0.1
	double max_acc_y = max_torque_y/inertia_y;
	double max_drot_y = max_acc_y*post_step*2;
	double max_acc_z = max_torque_z/inertia_z;
	double max_drot_z = max_acc_z*post_step*2;

	//double stop_angle_x = 0.5*avel.x*(avel.x/max_acc_x);
	//double stop_angle_y = 0.5*avel.y*(avel.y/max_acc_y);
	//double stop_angle_z = 0.5*avel.z*(avel.z/max_acc_z);

	
	//double max_drot_x = max_acc_x*steptime*4;
	//double max_drot_y = max_acc_y*steptime*4;
	//double max_drot_z = max_acc_z*steptime*4;
	//sprintf(oapiDebugString(), "steptime %0.4f post_step %0.4f", step_time_old*100,post_step*100);
	

	VECTOR3 aspdv;
	vme->GetAirspeedVector(FRAME_LOCAL, aspdv);
	normalise(aspdv);

	VECTOR3 relvel;
	//OBJHANDLE gbod = GetGravityRef();
	vme->GetRelativeVel(gBod, relvel);
	double relspeed = length(relvel);
	VECTOR3 relpos;
	vme->GetRelativePos(gBod, relpos);
	double rdist = length(relpos);
	double orb_ang_speed = relspeed/rdist;
	VECTOR3 relpos_hrz = _V(0,1,0);
	VECTOR3 relpos_loc;
	vme->HorizonInvRot(relpos_hrz, relpos_loc);
	normalise(relpos_loc);
	VECTOR3 rot_off = crossp(aspdv, relpos_loc);
	normalise(rot_off);
	rot_off.x = rot_off.x*orb_ang_speed;
	rot_off.y = rot_off.y*orb_ang_speed;
	rot_off.z = rot_off.z*orb_ang_speed;

	//sprintf(oapiDebugString(), "rot_off %0.4f %0.4f %0.4f ", rot_off.x*DEG, rot_off.y*DEG, rot_off.z*DEG);

	VECTOR3 zaxis_vel_nml = crossp(_V(0,0,1),aspdv);
	normalise(zaxis_vel_nml);
	VECTOR3 fv_vel_nml = crossp(fv_unit, aspdv);
	normalise(fv_vel_nml);
	VECTOR3 rot_axis = crossp(fv_vel_nml,zaxis_vel_nml);
	normalise(rot_axis);

	double zaxis_fv_angle = 1*acos(dotp(fv_vel_nml, zaxis_vel_nml));
//	if(abs(zaxis_fv_angle*DEG) < 40) bank_0_sw = bank_phase_sw;

//	double zaxis_fv_angle = acos(dotp(fv_vel_nml, zaxis_vel_nml));
	if(bank_mode_sw == 2 && zaxis_fv_angle < (10*RAD))bank_mode_sw = 0;
	
	if(zaxis_fv_angle > vrot_lim)zaxis_fv_angle = vrot_lim;
	if(abs(zaxis_fv_angle) < (0.1*RAD))deadband = (1*RAD);
 	else if(abs(zaxis_fv_angle) > (1*RAD))deadband = (0.1*RAD);
	if(abs(zaxis_fv_angle) < deadband)zaxis_fv_angle = 0;

	double pitch_angle = acos(dotp(aspdv,fv_unit));
	double current_pitch_angle = acos(dotp(aspdv,_V(0,0,1)));
	double cos_cpa = cos(current_pitch_angle);
	double sin_cpa = sin(current_pitch_angle);

	MATRIX3 rm = rotm(zaxis_vel_nml, -(pitch_angle-PI05));
	VECTOR3 headsup_vector = mul(rm, aspdv);
	normalise(headsup_vector);

	double hu_error = 1*acos(dotp(_V(0,1,0), headsup_vector));
	
	if(hu_error > vrot_lim)hu_error = vrot_lim;
	if(abs(hu_error) < (0.1*RAD))deadband2 = (1*RAD);
 	else if(abs(hu_error) > (1*RAD))deadband2 = (0.1*RAD);
	if(abs(hu_error) < deadband2)hu_error = 0;
	VECTOR3 hu_rot_axis = crossp(headsup_vector, _V(0,1,0));
	normalise(hu_rot_axis);

	//sprintf(oapiDebugString(), "hu_error %0.3f zaxis_fv_angle %0.3f avel %0.3f %0.3f %0.3f rot_off %0.3f %0.3f %0.3f", hu_error*DEG, zaxis_fv_angle*DEG, avel.x*DEG, avel.y*DEG, avel.z*DEG, rot_off.x*DEG,rot_off.y*DEG,rot_off.z*DEG);

	VECTOR3 req_acc_hu;
	req_acc_hu.x = hu_rot_axis.x*hu_error*0.8;
	req_acc_hu.y = hu_rot_axis.y*hu_error*0.8;
	req_acc_hu.z = -hu_rot_axis.z*hu_error*0.8;


	VECTOR3 req_acc;
	req_acc.x = (rot_axis.x*zaxis_fv_angle*0.8) - avel.x + rot_off.x;
	req_acc.y = (rot_axis.y*zaxis_fv_angle*0.8) - avel.y + rot_off.y;
	req_acc.z = (-rot_axis.z*zaxis_fv_angle*0.8) + avel.z - rot_off.z;

	req_acc = req_acc+req_acc_hu;

	double thx_lvl = req_acc.x/max_drot_x;
	double thy_lvl = req_acc.y/max_drot_y;
	double thz_lvl = req_acc.z/max_drot_z;
	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));

	if(vme->GetAltitude() < 20000)
	{
		double frf = 0.333/post_step;
		if(abs(thx_lvl) < frf)thx_lvl = 0;
		if(abs(thy_lvl) < frf)thy_lvl = 0;
		if(abs(thz_lvl) < frf)thz_lvl = 0;
	}

	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;
//	if(abs(thx_lvl) < 0.2)thx_lvl = 0;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;
//	if(abs(thy_lvl) < 0.2)thy_lvl = 0;

	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;
//	if(abs(thz_lvl) < 0.2)thz_lvl = 0;

	


	////////////// x pitch
	if(thx_lvl > 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}

	//////////////////////Y YAW
	if(thy_lvl > 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}

	/////////////// z roll
	if(thz_lvl > 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, thz_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		}
	else if(thz_lvl < 0)
		{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, -thz_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
		}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
}

void  CrewDragonMFD::RollToBankAngle()
{
double vrot_lim = 10*RAD;

VECTOR3 haspdv = _V(0,0,1);
vme->GetGroundspeedVector(FRAME_HORIZON,haspdv);
haspdv = -haspdv;
normalise(haspdv);

if (auto_burn_sw == 1) haspdv = _V(0,-1,0);

VECTOR3 haspdv_loc = _V(0,0,1);
vme->HorizonInvRot(haspdv, haspdv_loc);

double roll_angle = atan2(haspdv_loc.x,-haspdv_loc.y);
if (roll_angle > PI) roll_angle = roll_angle - PI2;
else if (roll_angle < -PI) roll_angle = roll_angle + PI2;

//sprintf(oapiDebugString(), "roll_angle %0.1f", roll_angle*DEG);

	double mass = vme->GetMass();

	THRUSTER_HANDLE hThruster = vme->GetThrusterHandleByIndex(0);
	double RCSTH = vme->GetThrusterMax(hThruster);
	VECTOR3 pmi = _V(0,0,0);
	vme->GetPMI(pmi);

	double max_rcs = 1*RCSTH;
	double max_torque = 2*max_rcs*2;
	double inertia = pmi.z*mass;

	VECTOR3 avel = _V(0,0,0);
	vme->GetAngularVel(avel);


	double max_acc = max_torque/inertia;
	double max_drot = max_acc*steptime;


	double reqzv = 1*abs(roll_angle);
	if (reqzv > vrot_lim)reqzv = vrot_lim;
	if (roll_angle > 0)reqzv = -reqzv;
	double req_drot = reqzv - avel.z;
	double thz_lvl = req_drot/max_drot;
	if((roll_angle*DEG) < 0.5 && (roll_angle*DEG) > -0.5) thz_lvl = 0.25*thz_lvl;

	/*if((roll_angle*DEG) < 1 && (roll_angle*DEG) > -1)
	{
	reqzv = 1*-roll_angle;
	double acc_reqz = (reqzv-avel.z)/0.1;
	thz_lvl = acc_reqz/max_acc;
	}*/


	if(thz_lvl > 1)thz_lvl = 1;
	else if(thz_lvl < -1)thz_lvl = -1;


	if(thz_lvl > 0)
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, thz_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}
	else if(thz_lvl < 0)
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -thz_lvl);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	}
	else 
	{
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
	}


}

void  CrewDragonMFD::DoNav ()
{	
	int nav_add_orbits = add_orbits;
	if(auto_reentry_sw == 1)nav_add_orbits = 0;

	DWORD attcount = vme->AttachmentCount(true);
	ATTACHMENTHANDLE hAtt;
	OBJHANDLE hasparent;
	for (int i = 0; i < attcount; i++)
	{
	hAtt = vme->GetAttachmentHandle(true, i);
	hasparent = vme->GetAttachmentStatus(hAtt);
		if (hasparent)
		{
		flt_time = 0;
		tgt_arc = 0;
		req_inc = 0;
		req_hdg = 0;
		curr_hdg = 0;
		hud_crossrange_error = 0;
		hud_range = 0;
		time_to_burn = 0;
		auto_burn_sw = 0;
		if(auto_reentry_sw < 2)auto_reentry_sw = 0;
		has_parent = true;
		return;
		}
		else has_parent = false;
	}
	
	
	////////////////////////////////
	int dock_count = vme->DockCount();
	DOCKHANDLE hDock;
	for (int i = 0; i < dock_count; i++)
	{
	hDock = vme->GetDockHandle(i);
	hasparent = vme->GetDockStatus(hDock);
	if (hasparent)
		{
		flt_time = 0;
		tgt_arc = 0;
		req_inc = 0;
		req_hdg = 0;
		curr_hdg = 0;
		hud_crossrange_error = 0;
		hud_range = 0;
		time_to_burn = 0;
		auto_burn_sw = 0;
		if(auto_reentry_sw < 2)auto_reentry_sw = 0;
		has_parent = true;
		return;
		}
		else has_parent = false;
	}

	//////////////////////////////

	vme->GetElements (gBod,el,&prm,0,FRAME_EQU);

	if(reentry_mode > 0)
	{
	double ip_sma = el.a;
	double ip_Rp = prm.PeD;
	double ip_Ra = (2*ip_sma) - ip_Rp;
	double ip_Rp2 = 6436010;
	if(reentry_mode == 1)ip_Rp2 = 6431010;
	double ip_Rei = 6491010;
	double ip_Vp = sqrt((2*mu*ip_Ra)/(ip_Rp2*(ip_Ra + ip_Rp2)));
	double ip_ecc = (ip_Rp2*ip_Vp*ip_Vp/mu) - 1;
	//double cos_ta = ip_sma*(1-(ip_ecc*ip_ecc))/(ip_ecc*ip_Rei);
	//double omega = acos(cos_omega);
	//double ip_ta = acos(cos_ta);
	//double tan_omega = (ip_ecc*sin(ip_ta))/(1 + (ip_ecc*cos(ip_ta)));
	//double omega = atan(tan_omega);

	double ip_V2 = sqrt((2*mu*((1/ip_Rei) - (1/ip_Rp2))) + (ip_Vp*ip_Vp));
	double sin_omega = (ip_Rp2*ip_Vp)/(ip_Rei*ip_V2);
	hud_omega = PI05 - asin(sin_omega);

	double C = (ip_Rp2 * ip_Vp * ip_Vp)/mu;
	double tan_v = (C * cos(hud_omega) * sin(hud_omega))/((C * cos(hud_omega) * cos(hud_omega))-1);
	double v = atan(tan_v);

	hud_lambda = 46.2 - (DEG*v);
	if(reentry_mode == 1)
		{
		hud_lambda = 48.2 - (DEG*v);
		if(hud_lambda < 28)hud_lambda = 28;
		}
//	sprintf(oapiDebugString(), "v %0.1f tan_v %0.4f", DEG*v, tan_v);
	}

	if(reentry_mode == 2)
	{
		if(hud_lambda < 28)hud_lambda = 28;
	}
		
		//////////////////// check if targeting valid
	
	if (hTgt == NULL)
	{
		flt_time = 0;
		tgt_arc = 0;
		req_inc = 0;
		req_hdg = 0;
		curr_hdg = 0;
		hud_crossrange_error = 0;
		hud_range = 0;
		if (auto_reentry_sw == 1)auto_reentry_sw = 2;
		return;
	}
	else if (auto_reentry_sw == 2)auto_reentry_sw = 1;

	/////////////////////////////////
	

	double alti = vme->GetAltitude();
	
	VECTOR3 rpos_glob = _V(0,0,0);
	vme->GetRelativePos(gBod,rpos_glob);

	VECTOR3 tgt_rpos_glob = _V(0,0,0);
	oapiGetRelativePos(hTgt,gBod,&tgt_rpos_glob);

	VECTOR3 tgt_pos_glob = _V(0,0,0);
	oapiGetGlobalPos(hTgt, &tgt_pos_glob);

	double tgt_off_angle = flt_time*earth_rot_rate;
	MATRIX3 rm = rotm(GBOD_AXIS, -tgt_off_angle);
	VECTOR3 tgt_rpos_glob_rot = mul(rm, tgt_rpos_glob);
	
	VECTOR3 tgt_rpos_glob_rot_nzd = tgt_rpos_glob_rot;
	normalise(tgt_rpos_glob_rot_nzd);
	VECTOR3 rpos_glob_nzd = rpos_glob;
	normalise(rpos_glob_nzd);
	VECTOR3 tgt_rpos_glob_nzd = tgt_rpos_glob;
	normalise(tgt_rpos_glob_nzd);

	//tgt_arc = acos(dotp(tgt_rpos_glob_rot_nzd,rpos_glob_nzd));
	

	VECTOR3 rvel_glob = _V(0,0,0);
	vme->GetRelativeVel(gBod, rvel_glob);
	VECTOR3 rvel_glob_nzd = rvel_glob;
	normalise(rvel_glob_nzd);

	VECTOR3 ptr_req_orb_nml_ecl = crossp(tgt_rpos_glob, rpos_glob);
	normalise(ptr_req_orb_nml_ecl);

	VECTOR3 orb_nml_ecl = crossp(rvel_glob_nzd,rpos_glob_nzd);
	normalise(orb_nml_ecl);

	VECTOR3 target_ref_nml = crossp(orb_nml_ecl,tgt_rpos_glob_rot_nzd);
	normalise(target_ref_nml);

	VECTOR3 target_ref_in_pln = crossp(target_ref_nml,orb_nml_ecl);
	normalise(target_ref_in_pln);

	double schk8 = (dotp(target_ref_in_pln,rpos_glob_nzd));
	if (schk8 < -1)schk8 = -1;
	else if (schk8 > 1)schk8 = 1;

	tgt_arc = acos(schk8);

	VECTOR3 ves_tgt_nml = crossp(target_ref_in_pln, rpos_glob_nzd);
	normalise(ves_tgt_nml);

	double app_away = dotp(ves_tgt_nml,orb_nml_ecl);

	VECTOR3 ves_tgt_nml2 = crossp(tgt_rpos_glob_nzd, rpos_glob_nzd);
	normalise(ves_tgt_nml2);

	double app_away2 = dotp(ves_tgt_nml2,orb_nml_ecl);

	int aa1 = 0;
	int aa2 = 0;
	if(app_away > 0)aa1 = 1;
	else aa1 = -1;
	if(app_away2 > 0)aa2 = 1;
	else aa2 = -1;


	if(aa2 == -1)
	{
		if(add_orbits_sw == 1)
		{
			add_orbits = add_orbits - 1;
			if(add_orbits < 0)add_orbits = 0;
		}
	}
	add_orbits_sw = aa2;
	

	if (alti < 80000)tgt_arc = tgt_arc;
	else if (auto_burn_sw == 2)tgt_arc = tgt_arc;
	else if (auto_reentry_sw > 0)tgt_arc = tgt_arc;
//	else if(tgt_arc < (90*RAD) && (aa1 > 0)) tgt_arc = tgt_arc + PI2;
	else if (aa1 < 0 && aa2 < 0) tgt_arc = PI2 - tgt_arc;
	else if(aa1 > 0 && aa2 < 0)tgt_arc = tgt_arc + PI2;
	else if(aa1 < 0 && aa2 > 0)tgt_arc = PI2 - tgt_arc;
	tgt_arc = tgt_arc + (PI2 * (nav_add_orbits));

	VECTOR3 req_orb_nml_ecl = crossp(tgt_rpos_glob_rot_nzd, rpos_glob);
	normalise(req_orb_nml_ecl);

	VECTOR3 req_tan_vel_vect_glob = crossp(rpos_glob, req_orb_nml_ecl);
	normalise(req_tan_vel_vect_glob);

	VECTOR3 ptr_req_tan_vel_vect_glob = crossp(rpos_glob, ptr_req_orb_nml_ecl);
	normalise(ptr_req_tan_vel_vect_glob);
	
	VECTOR3 pos_glob;
	vme->GetGlobalPos(pos_glob);

	VECTOR3 ptr_req_tan_vel_vect_glob_pos = ptr_req_tan_vel_vect_glob+pos_glob;
	VECTOR3 ptr_req_tan_vel_vect_loc = _V(0,0,0);
	vme->Global2Local(ptr_req_tan_vel_vect_glob_pos,ptr_req_tan_vel_vect_loc);
	VECTOR3 ptr_req_tan_vel_vect_hrz = _V(0,0,0);
	vme->HorizonRot(ptr_req_tan_vel_vect_loc, ptr_req_tan_vel_vect_hrz);
	tgt_hdg = atan2(ptr_req_tan_vel_vect_hrz.x, ptr_req_tan_vel_vect_hrz.z);

	VECTOR3 req_tan_vel_vect_glob_pos = req_tan_vel_vect_glob+pos_glob;
	VECTOR3 req_tan_vel_vect_loc = _V(0,0,0);
	vme->Global2Local(req_tan_vel_vect_glob_pos,req_tan_vel_vect_loc);
	VECTOR3 req_tan_vel_vect_hrz = _V(0,0,0);
	vme->HorizonRot(req_tan_vel_vect_loc, req_tan_vel_vect_hrz);
	req_hdg = atan2(req_tan_vel_vect_hrz.x, req_tan_vel_vect_hrz.z);

	


	VECTOR3 rvel_hrz = _V(0,0,0);
	VECTOR3 rvel_loc = _V(0,0,0);
	VECTOR3 rvel_glob_pos = rvel_glob+pos_glob;
	vme->Global2Local(rvel_glob_pos,rvel_loc);
	vme->HorizonRot(rvel_loc, rvel_hrz);
	curr_hdg = atan2(rvel_hrz.x, rvel_hrz.z);

	double schk9 = (dotp(GBOD_AXIS,req_orb_nml_ecl));
	if (schk9 < -1)schk9 = -1;
	else if (schk9 > 1)schk9 = 1;

	req_inc = acos(schk9);
	if (req_inc > PI)req_inc =  req_inc-PI;
	else if (req_inc < -PI)req_inc =  req_inc+PI;
	if (req_inc > PI05)req_inc =  PI-req_inc;

	tgt_dist = tgt_arc*rt;

	hud_range = tgt_dist;

	double landing_range = 0;
	if(alti < 30000)
//	if(tgt_dist < 250000 && alti < 50000)
	{
	VECTOR3 tgt_pos_loc = _V(0,0,0);
	vme->Global2Local(tgt_pos_glob,tgt_pos_loc);
	VECTOR3 tgt_pos_hrz = _V(0,0,0);
	vme->HorizonRot(tgt_pos_loc, tgt_pos_hrz);
	req_hdg = atan2(tgt_pos_hrz.x, tgt_pos_hrz.z);

	VECTOR3 aspd_hrz = _V(0,0,0);
	vme->GetAirspeedVector(FRAME_HORIZON, aspd_hrz);
	curr_hdg = atan2(aspd_hrz.x, aspd_hrz.z);
	if(curr_hdg < 0)curr_hdg = curr_hdg + PI2;

	double tgt_angle = acos(dotp(rpos_glob_nzd, tgt_rpos_glob_nzd));
	landing_range = rt*tgt_angle;
	hud_range = landing_range;
	}
	else if(auto_reentry_sw > 0)
	{
	VECTOR3 tgt_pos_loc = _V(0,0,0);
	vme->Global2Local(tgt_pos_glob,tgt_pos_loc);
	VECTOR3 tgt_pos_hrz = _V(0,0,0);
	vme->HorizonRot(tgt_pos_loc, tgt_pos_hrz);
	req_hdg = atan2(tgt_pos_hrz.x, tgt_pos_hrz.z);

	VECTOR3 aspd_hrz = _V(0,0,0);
	vme->GetAirspeedVector(FRAME_HORIZON, aspd_hrz);
	curr_hdg = atan2(aspd_hrz.x, aspd_hrz.z);
	if(curr_hdg < 0)curr_hdg = curr_hdg + PI2;

	double tgt_angle = acos(dotp(rpos_glob_nzd, tgt_rpos_glob_nzd));
	landing_range = rt*tgt_angle;
	hud_range = landing_range;
	}
	
	
	

	GetFlightTime();
	if(alti < 25000) flt_time = 0;

	double schk10 = (dotp(tgt_rpos_glob_rot_nzd,orb_nml_ecl));
	if (schk10 < -1)schk10 = -1;
	else if (schk10 > 1)schk10 = 1;

	crossrange_error = acos(schk10);
	crossrange_error = crossrange_error - PI05;
	//if (crossrange_error < 0)crossrange_error = -crossrange_error;
	crossrange_error = -crossrange_error*rt; // -ve for orbit anti normal dir

	hud_crossrange_error = crossrange_error;

	/*if(tgt_dist < 2000000 && auto_reentry_sw == 1)
	{
		vme->GetGroundspeedVector(FRAME_HORIZON,rvel_hrz);
		curr_hdg = atan2(rvel_hrz.x, rvel_hrz.z);
		double hdgang = req_hdg - curr_hdg;
		hud_crossrange_error = -1*sin(hdgang)*hud_range;
	}*/
//	sprintf(oapiDebugString(), "aa1 %d aa2 %d rng %0.1f ttb %0.1f xrng %0.1f", aa1, aa2, tgt_dist*0.001, time_to_burn, crossrange_error*0.001);

	double arc_tgt_to_3000 = 26.9796*RAD;
	double my_ta = prm.TrA;
	double tgt_ta = my_ta+tgt_arc;
	double ei_ta = tgt_ta-arc_tgt_to_3000;

	double ecc = el.e;
	double sma = el.a;

	ei_radius = sma*(1-(ecc*ecc))/(1+(ecc*(cos(ei_ta))));
	double ei_omega = atan((ecc*(sin(ei_ta)))/(1+(ecc*(cos(ei_ta)))));

	double ang_past_lan = prm.TrA + el.omegab - el.theta;
	double end_lan_ang = ang_past_lan+tgt_arc;//-arc_tgt_to_3000;
	if (end_lan_ang > PI2)end_lan_ang = end_lan_ang - PI2;
	if (end_lan_ang > PI)end_lan_ang = end_lan_ang - PI;

	//sprintf(oapiDebugString(), "ei_radius %0.1f ei_omega %0.1f epl %0.1f", ei_radius*0.001, ei_omega*DEG, end_lan_ang*DEG);

	//sprintf(oapiDebugString(), "tgtdst %0.1f tgt_hdg %0.1f req_hdg %0.1f crs_hdg %0.1f flttime %0.1f tgt_off_angle %0.1f cre %0.1f appretrl %d apprettg %d",tgt_dist*0.001, tgt_hdg*DEG, req_hdg*DEG, curr_hdg*DEG, flt_time, tgt_off_angle*DEG, crossrange_error*0.001, appret_real, appret_tgt);

	////////////////////// reentry burn ///////////////////////////////

	double schk11 = (dotp(rvel_glob_nzd, rpos_glob_nzd));
	if (schk11 < -1)schk11 = -1;
	else if (schk11 > 1)schk11 = 1;

	double lambda_1 = acos(schk11);
	double phi_1 = PI05 - lambda_1;
	double v_1 = length(rvel_glob);
	double r_1 = alti+rt;

	double r_2 = sma*(1-(ecc*ecc))/(1+(ecc*(cos(my_ta+arc_to_burn))));

	double schk2 = (mu*((2/r_2)-(1/sma)));
	if (schk2 < 0)schk2 = -schk2;

	v_2 = sqrt(schk2);
	double phi_2 = atan((ecc*(sin(my_ta+arc_to_burn)))/(1+(ecc*(cos(my_ta+arc_to_burn)))));
	lambda_2 = PI05 - phi_2;

	double lambda_3 = 91.05*RAD;
	double phi_3 = PI05 - lambda_3;
	double r_3_tgt = r_3 + r_3_fact;

	double c = (r_1*sin(lambda_1))/(r_3_tgt*sin(lambda_3));
	double r_av = (r_1+r_3_tgt)/2;
	double g_av = mu/(r_av*r_av);

	double schk3 = (2*g_av*(r_1-r_3_tgt)/((c*c)-1));
	if (schk3 < 0)schk3 = -schk3;

	v_2pb = sqrt(schk3);
	double v_3 = c*v_2pb;

	double k_2 = r_2*v_2pb*v_2pb/mu;
	double k_3 = r_3_tgt*v_3*v_3/mu;

	double ta_2 = atan2((k_2*cos(phi_2)*sin(phi_2)),((k_2*cos(phi_2)*cos(phi_2))-1));
	if (ta_2 < 0)ta_2 = ta_2+PI2;
	double ta_3 = atan2((k_3*cos(phi_3)*sin(phi_3)),((k_3*cos(phi_3)*cos(phi_3))-1));
	if (ta_3 < 0)ta_3 = ta_3+PI2;

	double reentry_arc = abs(ta_3-ta_2);
	double burn_ta = ei_ta - reentry_arc;
	arc_to_burn = burn_ta - my_ta;

	//sprintf(oapiDebugString(), "ei_rad %0.1f ei_om %0.1f arc_to_burn %0.1f v2 %0.1f v2pb %0.1f", ei_radius*0.001, ei_omega*DEG, arc_to_burn*DEG, v_2, v_2pb);

}

void  CrewDragonMFD::GetFlightTime ()
{
	double flt_tgt_arc = tgt_arc;
	int fttasw = 0;
	int num_orb = 0;
	if (flt_tgt_arc > PI2) 
	{
		num_orb = (int) (flt_tgt_arc/PI2);
		flt_tgt_arc = flt_tgt_arc-(num_orb*PI2);
		fttasw = 1;
	}

	double apd = prm.ApD;
	double ped = prm.PeD;

	double sma = 0.5*(apd + ped);
	double ecc = 1-(2/((apd/ped)+1));
	
	double to = 0;
	double tf = 0;
	double vo = prm.TrA;
	double vf = vo+flt_tgt_arc;

	double schk6 = ((ecc + cos(vo))/(1+(ecc*cos(vo))));
	if (schk6 < -1)schk6 = -1;
	else if (schk6 > 1)schk6 = 1;

	double eo = acos(schk6);
	if (vo > PI) eo = PI2 - eo;

	double schk7 = (ecc + cos(vf))/(1+(ecc*cos(vf)));
	if (schk7 < -1)schk7 = -1;
	else if (schk7 > 1)schk7 = 1;

	double ef = acos(schk7);

	if (vf > 3*PI) ef = 4*PI - ef;
	else if (vf > PI2) ef = ef + PI2;
	else if (vf > PI) ef = PI2 - ef;

	
	double mo = eo - (ecc*sin(eo));
	double mf = ef - (ecc*sin(ef));
	//sprintf(oapiDebugString(), "eca %0.2f eo %0.2f ef %0.1f mo %0.2f mf %0.1f",prm.EcA*DEG, eo*DEG, ef*DEG, mo*DEG, mf*DEG);
//	if (mf < mo) mf = mf + PI2;

	double schk1 = (mu/(sma*sma*sma));
	if (schk1 < 0)schk1 = -schk1;

	double n = sqrt(schk1);
	tf = to + ((mf-mo)/n);
	flt_time = tf;

	if (fttasw == 1) flt_time = flt_time+(prm.T*num_orb);

}

void  CrewDragonMFD::DoReentryBurn()
{

	int no_tgt_sw = 0;
	//////////////////// check if targeting valid
	//OBJHANDLE hTgt = oapiGetObjectByName(tgt_name);
	//if (hTgt == NULL)hTgt = oapiGetBaseByName(hGref, tgt_name);
	if (hTgt == NULL)
	{
		no_tgt_sw = 1;
		tgt_dist = 0;
		hud_range = 0;
		crossrange_error = 0;
		GetdVNoTarget();
	}

	/////////////////////////////////
	THRUSTER_HANDLE hThruster = vme->GetThrusterHandleByIndex(0);
	double ISP = vme->GetThrusterIsp(hThruster);
	double RCSTH = vme->GetThrusterMax(hThruster);

	VECTOR3 rvel_glob = _V(0,0,0);
	vme->GetRelativeVel(gBod, rvel_glob);

	VECTOR3 pos_glob;
	vme->GetGlobalPos(pos_glob);

	VECTOR3 rvel_glob_pos = rvel_glob+pos_glob;
	VECTOR3 rvel_glob_pos_loc = _V(0,0,0);
	vme->Global2Local(rvel_glob_pos,rvel_glob_pos_loc);

	VECTOR3 rvel_glob_pos_loc_nzd = rvel_glob_pos_loc;
	normalise(rvel_glob_pos_loc_nzd);

	VECTOR3 rvel_glob_hrz = _V(0,0,0);
	vme->HorizonRot(rvel_glob_pos_loc, rvel_glob_hrz);
	
	flight_vector = rvel_glob_pos_loc_nzd;
	bank_angle = PI;

	double schk12 = (dotp(flight_vector,_V(0,0,1)));
	if (schk12 < -1)schk12 = -1;
	else if (schk12 > 1)schk12 = 1;

	double pointing_error = acos(schk12);

	if (no_tgt_sw == 1 && auto_burn_sw == 1) ////////////////////// no tgt
	{
		if (prm.PeD > (rt+50000))
		{
			if (pointing_error < 2*RAD) vme->SetThrusterGroupLevel(THGROUP_MAIN, 1);
			else vme->SetThrusterGroupLevel(THGROUP_MAIN, 0);
			burn_sw = 1;
		}
		else
		{
		vme->SetThrusterGroupLevel(THGROUP_MAIN, 0);
		auto_burn_sw = 2;
		burn_sw = 0;
		}
	}

	else /////////////// targeting
	{
	double req_dv = length(rvel_glob) - v_2pb;
	if(req_dv < 0)req_dv = 0;
	if (no_tgt_sw == 0) dv_required = req_dv;
	double tot_mass = vme->GetMass();

	double fraction = exp(req_dv/ISP);
	double fuelmass = tot_mass - (tot_mass/fraction);
	double fuelrate = 4*RCSTH/ISP;
	double dvtime = fuelmass/fuelrate;

	double ang_past_lan = prm.TrA + el.omegab - el.theta;
	double sin_inc = sin(el.i);
	double inc_factor = abs(sin_inc*sin_inc*sin_inc*2);// 1.626); // 1 at 51.65 inc

	//r_3_fact = ((sin(end_lan_ang))*38330*inc_factor*sma_factor)-11000;
	//sprintf(oapiDebugString(), "r_3_fact %0.1f", r_3_fact);

	//////////////////////////////////////////////////////////// alternative ending
	double tgt_arc_fact = (tgt_arc-(RAD*26.9796))/(120*RAD);
	double om_off = -13.7*RAD;
	r_3_fact = inc_factor*tgt_arc_fact*(-10457*0.5*(cos((2*((ang_past_lan+tgt_arc)-(RAD*26.9796)))+ om_off) - cos((2*(ang_past_lan)) + om_off))) + 1930;
	if (vme->NonsphericalGravityEnabled() == false)r_3_fact = 0;


	//sprintf(oapiDebugString(), "r_3_fact %0.1f", r_3_fact);

	double r_3_lim = r_3+r_3_fact;

	GetTimeToBurn ();
	time_to_burn = time_to_burn - (0.5*dvtime);
	if(no_tgt_sw == 1)time_to_burn = -0.001;

	if (time_to_burn < 0 && auto_burn_sw == 1)
	{
	if (ttb_check_sw == 1)burn_sw = 1;
	ttb_check_sw = ttb_check_sw+1;
	}
	else ttb_check_sw = 0;
		

	if (burn_sw == 1 && auto_burn_sw == 1)
		{
		if (ei_radius > r_3_lim)
			{
			if (pointing_error < 2*RAD) vme->SetThrusterGroupLevel(THGROUP_MAIN, 1);
			else vme->SetThrusterGroupLevel(THGROUP_MAIN, 0);
			}
		else
			{
			vme->SetThrusterGroupLevel(THGROUP_MAIN, 0);
			burn_sw = 0;
			CrewDragonMFD::KillAllRCS ();
			auto_burn_sw = 2;
			}
		}
	}

}

void  CrewDragonMFD::GetTimeToBurn ()
{
	
	double flt_tgt_arc = arc_to_burn;
	int fttasw = 0;
	int num_orb = 0;
	if (flt_tgt_arc > PI2) 
	{
		num_orb = (int)(flt_tgt_arc/PI2);
		flt_tgt_arc = flt_tgt_arc-(num_orb*PI2);
		fttasw = 1;
	}

	double apd = prm.ApD;
	double ped = prm.PeD;

	double sma = 0.5*(apd + ped);
	double ecc = 1-(2/((apd/ped)+1));
	
	double to = 0;
	double tf = 0;
	double vo = prm.TrA;
	double vf = vo+flt_tgt_arc;

	double schk13 = ((ecc + cos(vo))/(1+(ecc*cos(vo))));
	if (schk13 < -1)schk13 = -1;
	else if (schk13 > 1)schk13 = 1;

	double eo = acos(schk13);
	if (vo > PI) eo = PI2 - eo;

	double schk14 = ((ecc + cos(vf))/(1+(ecc*cos(vf))));
	if (schk14 < -1)schk14 = -1;
	else if (schk14 > 1)schk14 = 1;

	double ef = acos(schk14);

	if (vf > 3*PI) ef = 4*PI - ef;
	else if (vf > PI2) ef = ef + PI2;
	else if (vf > PI) ef = PI2 - ef;
	
	
	double mo = eo - (ecc*sin(eo));
	double mf = ef - (ecc*sin(ef));
	if(vf < 0)mf = -mf;
//	sprintf(oapiDebugString(), "eca %0.2f eo %0.2f ef %0.1f mo %0.2f mf %0.1f",prm.EcA*DEG, eo*DEG, ef*DEG, mo*DEG, mf*DEG);
	//if (mf < mo) mf = mf + PI2;
	//sprintf(oapiDebugString(), "vf %0.2f flt_tgt_arc %0.2f mo %0.2f mf %0.2f",vf,flt_tgt_arc,mo,mf);

	double schk4 = (mu/(sma*sma*sma));
	if (schk4 < 0)schk4 = -schk4;

	double n = sqrt(schk4);
	tf = to + ((mf-mo)/n);
	time_to_burn = tf;

	if (fttasw == 1) time_to_burn = time_to_burn + (num_orb*prm.T);

}

// ==============================================================
// Ascent MFD implementation

CrewDragonMFD::CrewDragonMFD (DWORD w, DWORD h, VESSEL *vessel)
: GraphMFD (w, h, vessel)
{
status_data = -1;
chute_status = 0;
nos_htc_status = 0;
sid_htc_status = 0;
nos_con_status = 0;
pld_status = 0;
tnk_status = 0;
rcs_status = 0;
dok_status = 0;
cab_status = 0;
nav_status = 0;
lmp1_status = 0;
lmp2_status = 0;
//sprintf(oapiDebugString(), "Implmentation");

///////// init vessel /////////////////////
	if (strlen(vesname) == 0)
	{
	MyVes = oapiGetFocusObject();
	oapiGetObjectName(MyVes, vesname, 30);
	}
	else
	{
	MyVes = oapiGetVesselByName(vesname);
	}
	if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	else
	{
	MyVes = oapiGetFocusObject();
	oapiGetObjectName(MyVes, vesname, 30);
	if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	}

	if(MyVes)
	{
	int drg_chk = 0;
	drg_chk = vme->SendBufferedKey(COMM_SEND_STS, true, NULL);
	if(drg_chk > 1)ves_is_dragon = true;
	else ves_is_dragon = false;
//	sprintf(oapiDebugString(), "sts %d", drg_chk);
	}
CrewDragonMFD::StoreStatus ();
}

CrewDragonMFD::~CrewDragonMFD ()
{
status_data = -1;
chute_status = 0;
nos_htc_status = 0;
sid_htc_status = 0;
nos_con_status = 0;
pld_status = 0;
tnk_status = 0;
rcs_status = 0;
dok_status = 0;
cab_status = 0;
nav_status = 0;
lmp1_status = 0;
lmp2_status = 0;
//sprintf(oapiDebugString(), "Destruction");
}



// message parser
int CrewDragonMFD::MsgProc (UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam)
{
	switch (msg) {
	case OAPI_MSG_MFD_OPENED:
		return (int)(new CrewDragonMFD (LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
		sprintf(oapiDebugString(), "mfd", mfd);
	}
	return 0;
}


bool CrewDragonMFD::ConsumeKeyBuffered (DWORD key)
{
	

	bool cbSelectTarget(void *id, char *str, void *data);
	bool cbSelectVessel(void *id, char *str, void *data);
	bool cbSelectPlanet(void *id, char *str, void *data);

	switch (key) {
	case OAPI_KEY_R:
		oapiOpenInputBox("Enter Reference Planet", cbSelectPlanet, 0, 30, (void *)this);
		return true;
	case OAPI_KEY_T:
		oapiOpenInputBox("Enter Landing Target", cbSelectTarget, 0, 30, (void *)this);
		return true;
	case OAPI_KEY_W:
		oapiOpenInputBox("Enter Vessel Name", cbSelectVessel, 0, 30, (void *)this);
		return true;
	case OAPI_KEY_N:
		if(ves_is_dragon == false)return false;
		else if(reentry_mode > 0)return false;
		else if(auto_burn_sw == 0)
		{
		auto_burn_sw = 1;
		auto_reentry_sw = 0;
		}
		else if(auto_burn_sw == 1)
		{
		KillAllRCS ();
		auto_burn_sw = 0;
		burn_sw = 0;
		}
		InvalidateDisplay();
		return true;
	case OAPI_KEY_Q:
		if(ves_is_dragon == false)return false;
		if(auto_reentry_sw == 0)
		{
		auto_reentry_sw = 1;
		if(auto_burn_sw == 1)auto_burn_sw = 0;
		}
		else
		{
		KillAllRCS ();
		auto_reentry_sw = 0;
	//	bank_mode_sw = 0;
	//	bank_dir_sw = 1;
		}
		InvalidateDisplay();
		return true;
	case OAPI_KEY_L:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_L, true, NULL); // const DWORD COMM_SEND_STS = 2306;
		InvalidateDisplay();
		return true;
	case OAPI_KEY_C:
		if (display_sw == 0) display_sw = 1;
		else display_sw = 0;
		button_sw = 2;
		InvalidateButtons();
		InvalidateDisplay();
		return true;
	case OAPI_KEY_K:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_K, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_G:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_G, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_U:
		if(ves_is_dragon == false) return false;
		char kstate[256];
		strcpy(kstate, "KEYMOD_CONTROL");
		vme->SendBufferedKey(OAPI_KEY_U, true, kstate);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_J:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_J, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_B:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_B, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_V:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_V, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_D:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_D, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_8:
		if(ves_is_dragon == false) return false;
		char kstate2[256];
		strcpy(kstate2, "KEYMOD_CONTROL");
		vme->SendBufferedKey(OAPI_KEY_8, true, kstate2);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_1:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_1, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_2:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_2, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_3:
		if(ves_is_dragon == false) return false;
		vme->SendBufferedKey(OAPI_KEY_3, true, NULL);
		InvalidateDisplay();
		return true;
	case OAPI_KEY_LBRACKET:
		if(ves_is_dragon == false) return false;
	//	else if(auto_reentry_sw == 1)return false;
		add_orbits = add_orbits+1;
		InvalidateDisplay();
		return true;
	case OAPI_KEY_RBRACKET:
		if(ves_is_dragon == false) return false;
	//	else if(auto_reentry_sw == 1)return false;
		add_orbits = add_orbits-1;
		if(add_orbits < 0)add_orbits = 0;
		InvalidateDisplay();
		return true;

	case OAPI_KEY_O:
		if(ves_is_dragon == false) return false;
		if(reentry_mode == 0)
			{
			reentry_mode = 1;
			if(auto_burn_sw == 1)auto_burn_sw = 0;
			burn_sw = 0;
			}
		else if(reentry_mode == 1)
			{
			reentry_mode = 2;
			if(auto_burn_sw == 1)auto_burn_sw = 0;
			burn_sw = 0;
			}
		else reentry_mode = 0;
		InvalidateDisplay();
		return true;
	}
	return false;
}

bool CrewDragonMFD::ConsumeButton (int bt, int event)
{
	if (!(event & PANEL_MOUSE_LBDOWN)) return false;
	if(display_sw == 0)
	{
	static const DWORD btkey[12] = { OAPI_KEY_W, OAPI_KEY_T, OAPI_KEY_R, OAPI_KEY_N, OAPI_KEY_Q, OAPI_KEY_L,
									OAPI_KEY_O,NULL,NULL,OAPI_KEY_LBRACKET,OAPI_KEY_RBRACKET,OAPI_KEY_C};
	if (bt < 12) return ConsumeKeyBuffered (btkey[bt]);
	else return false;
	}
	else if(display_sw == 1)
	{
	static const DWORD btkey[12] = { OAPI_KEY_K,OAPI_KEY_G, OAPI_KEY_B, OAPI_KEY_J, OAPI_KEY_U, OAPI_KEY_D,
									OAPI_KEY_V,OAPI_KEY_8,OAPI_KEY_1,OAPI_KEY_2,OAPI_KEY_3,OAPI_KEY_C};
	if (bt < 12) return ConsumeKeyBuffered (btkey[bt]);
	else return false;
	}
	else return false;
}

char *CrewDragonMFD::ButtonLabel (int bt)
{
	
	char *label[12] = {"VES", "TGT", "REF", "DEO", "RNT", "CHT",
						"MOD","","","OB+","OB-","DSP"};
	char *labelM[12] = {"NOH", "SID", "NOC", "PLD", "TNK", "RCS",
						"DOK","CAB","NAV","LMP","LMP","DSP"};
	if(display_sw == 0)return (bt < 12 ? label[bt] : 0);
	else return (bt < 12 ? labelM[bt] : 0);
}

int CrewDragonMFD::ButtonMenu (const MFDBUTTONMENU **menu) const
{
	if(display_sw == 0)
	{
	static const MFDBUTTONMENU mnu[12] = {
		{"Select Vessel", 0, 'W'},
		{"Select Target", 0, 'T'},
		{"Select Ref.Planet", 0, 'R'},
		{"Auto De-orbit", 0, 'N'},
		{"Auto Re-entry", 0, 'Q'},
		{"Chutes Arm/Safe", 0, 'L'},
		{"Reentry Mode", 0, 'O'},
		{"", 0, ' '},
		{"", 0, ' '},
		{"Orbits+1", 0, '['},
		{"Orbits-1", 0, ']'},
		{"Display", 0, 'C'}
	};
	if (menu) *menu = mnu;
	return 12;
	}
	else
	{
	static const MFDBUTTONMENU mnu[12] = {
		{"Nose Hatch", 0, 'K'},
		{"Side Hatch", 0, 'G'},
		{"Nose Cone", 0, 'B'},
		{"Jettison Payload", 0, 'J'},
		{"Jettison Trunk", 0, 'U'},
		{"RCS Mode", 0, 'D'},
		{"Extend/Retract Dock", 0, 'V'},
		{"Cabin Lights", 0, '8'},
		{"NAV Beacons", 0, '1'},
		{"Fwd.Lamp 1", 0, '2'},
		{"Fwd.lamp 2", 0, '3'},
		{"Display", 0, 'C'}
	};
	if (menu) *menu = mnu;
	return 12;
	}
}

bool cbSelectTarget(void *id, char *str, void *data)
{
	return (((CrewDragonMFD *)data)->SelectTarget(str));
}

bool cbSelectVessel(void *id, char *str, void *data)
{
	return (((CrewDragonMFD *)data)->SelectVessel(str));
}

bool cbSelectPlanet(void *id, char *str, void *data)
{
	return (((CrewDragonMFD *)data)->SelectPlanet(str));
}

bool CrewDragonMFD::SelectTarget(char *Name)
{
	char new_tgt[256];
	strcpy(new_tgt,Name);
	hTgt = oapiGetVesselByName(new_tgt);
	if(hTgt == NULL)hTgt = oapiGetBaseByName(gBod,new_tgt);

	if(hTgt == NULL)
	{
	if (!strnicmp (Name, "NULL", 4))
		{
		strcpy(mfdtgtname,Name);
		
		InvalidateDisplay();
		return true;
		}
	}

	//if (hTgt)
	else 
	{	
	strcpy(mfdtgtname,Name);
//	sprintf(oapiDebugString(), "tgt %s", mfdtgtname);
	InvalidateDisplay();
	return true;
	}

	return false;
}

bool CrewDragonMFD::SelectVessel(char *Name)
{
	MyVes = oapiGetVesselByName(Name);

	if (MyVes)
	{	
	strcpy(vesname,Name);
	vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	int drg_chk = 0;
	drg_chk = vme->SendBufferedKey(COMM_SEND_STS, true, NULL);
	if(drg_chk > 1)ves_is_dragon = true;
	else ves_is_dragon = false;
//	sprintf(oapiDebugString(), "sts %d", drg_chk);
	KillAllRCS();
	InvalidateDisplay();
	return true;
	}

	return false;
}

bool CrewDragonMFD::SelectPlanet(char *Name)
{
	gBod = oapiGetObjectByName(Name);

	if (gBod)
	{	
	strcpy(gbodyname,Name);
	GetGbodData();
	if(oapiIsVessel (hTgt) == false)strcpy(mfdtgtname,"NULL");
	InvalidateDisplay();
	return true;
	}

	
	return false;

}


void CrewDragonMFD::GetGbodData()
{
	
	double ob = oapiGetPlanetObliquity (gBod);
	GBOD_AXIS.x = 0;
	GBOD_AXIS.y = cos(ob);
	GBOD_AXIS.z = sin(ob);
	double eqlan = oapiGetPlanetTheta(gBod);
	MATRIX3 rm = rotm(_V(0,1,0),-eqlan);
	GBOD_AXIS = mul(rm,GBOD_AXIS);

	double prd = oapiGetPlanetPeriod(gBod);
	earth_rot_rate = PI2/prd;

	double mass = oapiGetMass(gBod);
	mu = GGRAV*mass;

	rt = oapiGetSize(gBod);
	//sprintf(oapiDebugString(), "ob %0.6f eqlan %0.6f", obl*DEG, eqlanan*DEG);
	
}



void CrewDragonMFD::Update (HDC hDC)
{
	if(ves_is_dragon == true)
	{
	status_data = vme->SendBufferedKey(COMM_SEND_STS, true, NULL);
	char dbuf[256];
	sprintf(dbuf, "%d", status_data);
	char sts_string[256];
	strcpy(sts_string, "                          ");
	strcpy(&sts_string[0], &dbuf[0]);
	strcpy(&sts_string[1], " ");
	strcpy(&sts_string[2], &dbuf[1]);
	strcpy(&sts_string[3], " ");
	strcpy(&sts_string[4], &dbuf[2]);
	strcpy(&sts_string[5], " ");
	strcpy(&sts_string[6], &dbuf[3]);
	strcpy(&sts_string[7], " ");
	strcpy(&sts_string[8], &dbuf[4]);
	strcpy(&sts_string[9], " ");
	strcpy(&sts_string[10], &dbuf[5]);
	strcpy(&sts_string[11], " ");
	sscanf_s(sts_string, "%d%d%d%d%d%d", &chute_status,  &nos_htc_status, &sid_htc_status, &nos_con_status,
			&pld_status, &tnk_status);// &rcs_status, &dok_status,
			//&cab_status, &nav_status, &lmp1_status, &lmp2_status);

	status_data = vme->SendBufferedKey(COMM_SEND_STS2, true, NULL);
	char ebuf[256];
	sprintf(ebuf, "%d", status_data);
	char sts2_string[256];
	strcpy(sts2_string, "                          ");
	strcpy(&sts2_string[0], &ebuf[0]);
	strcpy(&sts2_string[1], " ");
	strcpy(&sts2_string[2], &ebuf[1]);
	strcpy(&sts2_string[3], " ");
	strcpy(&sts2_string[4], &ebuf[2]);
	strcpy(&sts2_string[5], " ");
	strcpy(&sts2_string[6], &ebuf[3]);
	strcpy(&sts2_string[7], " ");
	strcpy(&sts2_string[8], &ebuf[4]);
	strcpy(&sts2_string[9], " ");
	strcpy(&sts2_string[10], &ebuf[5]);
	strcpy(&sts2_string[11], " ");
	sscanf_s(sts2_string, "%d%d%d%d%d%d",  &rcs_status, &dok_status, &cab_status, &nav_status, &lmp1_status, &lmp2_status);
	
	if(rcs_status == 2)
		{
			if(auto_reentry_sw > 0)vme->SendBufferedKey(OAPI_KEY_D, true, NULL);
		}
	
	}
	else
	{
	chute_status = 1;
	nos_htc_status = 1;
	sid_htc_status = 1;
	nos_con_status = 1;
	pld_status = 1;
	tnk_status = 1;
	rcs_status = 1;
	dok_status = 1;
	cab_status = 2;
	nav_status = 1;
	lmp1_status = 1;
	lmp2_status = 1;
	}

	////////////////////////////////

	if(button_sw > 0)
	{
	InvalidateButtons();
	button_sw = button_sw - 1;
	}
////////////////////////////////////////////
	SelectDefaultFont(hDC,0);	
	Title (hDC, "CrewDragonMFD");
if(display_sw == 0)
{
		
		SelectDefaultFont(hDC,0);
		SetTextColor(hDC, 0x0033FF33);	// green

		TextOut(hDC,(0.5*cw),(ch*2.7),"Vessel", 6);

		TextOut(hDC,(0.5*cw),(ch*6),"Target", 6);

		TextOut(hDC,(0.5*cw),(ch*9.3),"Ref.Planet", 10);

		SetTextColor(hDC, 0x0033FFFF);	// Yellow

		{
		char kbuf[256];
		sprintf(kbuf,"%s", (vesname) );
		TextOut(hDC,(cw*0.5),(ch*3.7),kbuf,strlen(kbuf));
		}

		{
		char dbuf[256];
		sprintf(dbuf,"%s", (mfdtgtname) );
		TextOut(hDC,(cw*0.5),(ch*7),dbuf,strlen(dbuf));
		}

		{
		char kbuf[256];
		sprintf(kbuf,"%s", (gbodyname));
		TextOut(hDC,(cw*0.5),(ch*10.3),kbuf,strlen(kbuf));
		}

		SetTextColor(hDC, 0x0033FF33);	// green
		TextOut(hDC,(0.5*cw),(ch*12.6),"Auto De-Orbit", 13);
		if(reentry_mode > 0)
		{
			SetTextColor(hDC, 0x0033FF33);
			TextOut(hDC,(0.5*cw),(ch*13.6),"N/A", 3);
		}
		else if(auto_burn_sw == 0)
		{
			SetTextColor(hDC, 0x0033FF33);
			TextOut(hDC,(0.5*cw),(ch*13.6),"OFF", 3);
		}
		else if(auto_burn_sw == 1)
		{
			SetTextColor(hDC, 0x005555FF);
			TextOut(hDC,(0.5*cw),(ch*13.6),"ACTIVE", 6);
		}
		else if(auto_burn_sw == 2)
		{
			SetTextColor(hDC, 0x00FFFFFF);
			TextOut(hDC,(0.5*cw),(ch*13.6),"FINISHED", 8);
		}

		SetTextColor(hDC, 0x0033FF33);	// green
		TextOut(hDC,(0.5*cw),(ch*15.9),"Auto Re-Entry", 13);

		if(auto_reentry_sw == 0)
		{
			SetTextColor(hDC, 0x0033FF33);
			TextOut(hDC,(0.5*cw),(ch*16.9),"OFF", 3);
		}
		else if(auto_reentry_sw == 3)
		{
			SetTextColor(hDC, 0x0033FF33);
			TextOut(hDC,(0.5*cw),(ch*16.9),"FINISHED", 8);
		}
		else
		{
			SetTextColor(hDC, 0x005555FF);
			TextOut(hDC,(0.5*cw),(ch*16.9),"ACTIVE", 6);
		}

		SetTextColor(hDC, 0x0033FF33);	// green
		TextOut(hDC,(0.5*cw),(ch*19.2),"Chutes Arm/Safe", 15);
		if(chute_status == 2)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(0.5*cw),(ch*20.2),"ARMED", 5);
		}
		//else if(chute_status == 3)
		//{
		//SetTextColor(hDC, 0x0033FFFF); 
		//TextOut(hDC,(0.5*cw),(ch*20.2),"DEPLOYED", 8);
		//}
		else TextOut(hDC,(0.5*cw),(ch*20.2),"SAFE", 4);

		///////// right hand side ////////
		SetTextAlign(hDC, TA_RIGHT);

		SetTextColor(hDC, 0x0033FF33);	// green
		TextOut(hDC,(37*cw),(ch*19.7),"MFD Display", 11);

		SetTextAlign(hDC, TA_LEFT);
		//////////////////////////////////////////////
		//SetTextColor(hDC, 0x0033FF33);	// green
		SetTextColor(hDC, 0x00FF6666);	// blue
		
		TextOut(hDC,(16.5*cw),(ch*7.7),"T-to-Burn",9);
		TextOut(hDC,(16.5*cw),(ch*8.7),"Tgt.Range",9);
		TextOut(hDC,(16.5*cw),(ch*9.7),"Cross-Range",11);
		TextOut(hDC,(16.5*cw),(ch*10.7),"Required Inc",12);
	//	TextOut(hDC,(16.5*cw),(ch*6.7),"Required Hdg",12);
	//	TextOut(hDC,(16.5*cw),(ch*7.7),"Current Hdg",11);
		TextOut(hDC,(16.5*cw),(ch*17),"Available dV",12);
		TextOut(hDC,(16.5*cw),(ch*18),"De-Orbit dV",11);
		
		{
		char kbuf[256];
		sprintf(kbuf,"Orbits to Tgt.");
		TextOut(hDC,(16.5*cw),(ch*14.7),kbuf,strlen(kbuf));
		}

		{
		char kbuf[256];
		sprintf(kbuf,"Reentry Mode");
		TextOut(hDC,(16.5*cw),(ch*2.7),kbuf,strlen(kbuf));
		}


		SetTextColor(hDC, 0x00FFFFFF);	// white
		SetTextAlign(hDC, TA_RIGHT);

		if(auto_reentry_sw > 0)
		{
		TextOut(hDC,(32*cw),(ch*7.7),"N/A   ", 3);
		}
		else if(reentry_mode > 0)
		{
		TextOut(hDC,(32*cw),(ch*7.7),"N/A   ", 3);
		}
		else if(auto_burn_sw < 2)
		{
		if(burn_sw == 0)
			{
			char kbuf[256];
			sprintf(kbuf,"%0.1fs", time_to_burn);
			TextOut(hDC,(36*cw),(ch*7.7),kbuf,strlen(kbuf));
			}
		else
			{
			SetTextColor(hDC, 0x005555FF);
			TextOut(hDC,(35*cw),(ch*7.7),"BURNING", 7);
			}
		SetTextColor(hDC, 0x00FFFFFF);
		}
		else
		{
		TextOut(hDC,(35*cw),(ch*7.7),"FINISHED", 8);
		}

		{
		char kbuf[256];
		sprintf(kbuf,"%0.1fkm", hud_range*0.001);
		TextOut(hDC,(37*cw),(ch*8.7),kbuf,strlen(kbuf));
		}

		{
		char kbuf[256];
		sprintf(kbuf,"%0.1fkm", hud_crossrange_error*0.001);
		TextOut(hDC,(37*cw),(ch*9.7),kbuf,strlen(kbuf));
		}

		{
		char kbuf[256];
		sprintf(kbuf,"%0.2f", req_inc*DEG);
		TextOut(hDC,(37*cw),(ch*10.7),kbuf,strlen(kbuf));
		}

		/*{
		char kbuf[256];
		if(req_hdg < 0)req_hdg = req_hdg+PI2;
		sprintf(kbuf,"%0.1f", req_hdg*DEG);
		TextOut(hDC,(36*cw),(ch*6.7),kbuf,strlen(kbuf));
		}*/

		/*{
		char kbuf[256];
		sprintf(kbuf,"%0.1f", curr_hdg*DEG);
		TextOut(hDC,(36*cw),(ch*7.7),kbuf,strlen(kbuf));
		}*/

		{
		char kbuf[256];
		sprintf(kbuf,"%0.1fm/s", dv_remaining);
		TextOut(hDC,(37*cw),(ch*17),kbuf,strlen(kbuf));
		}

		//if(auto_burn_sw < 2  && reentry_mode == 0)
		{
		char kbuf[256];
		if(auto_burn_sw == 2)sprintf(kbuf,"N/A     ");
		else if(reentry_mode > 0)sprintf(kbuf,"N/A     ");
		else if(auto_reentry_sw > 0)sprintf(kbuf,"N/A     ");
		else sprintf(kbuf,"%0.1fm/s", dv_required);
		TextOut(hDC,(37*cw),(ch*18),kbuf,strlen(kbuf));
		}

		{
		char kbuf[256];
		sprintf(kbuf,"+%d", add_orbits);
		TextOut(hDC,(34*cw),(ch*14.7),kbuf,strlen(kbuf));
		}

		{
		char kbuf[256];
		if(reentry_mode == 0)sprintf(kbuf,"LEO");
		else if(reentry_mode == 1)sprintf(kbuf,"GEO");
		else if(reentry_mode == 2)sprintf(kbuf,"HYPER");
		SetTextAlign(hDC, TA_LEFT);
		TextOut(hDC,(29*cw),(ch*2.7),kbuf,strlen(kbuf));
		}

		if(reentry_mode > 0 && auto_reentry_sw == 0)
		{
		SetTextColor(hDC, 0x00FF6666);	// blue
		SetTextAlign(hDC, TA_LEFT);
		char kbuf[256];
		sprintf(kbuf,"Alt.");
		TextOut(hDC,(16.5*cw),(ch*3.7),kbuf,strlen(kbuf));

		sprintf(kbuf,"ReA.");
		TextOut(hDC,(16.5*cw),(ch*4.7),kbuf,strlen(kbuf));

		sprintf(kbuf,"Ant.");
		TextOut(hDC,(16.5*cw),(ch*5.7),kbuf,strlen(kbuf));

		SetTextColor(hDC, 0x00FFFFFF);	// white

		sprintf(kbuf,"120km");
		TextOut(hDC,(29*cw),(ch*3.7),kbuf,strlen(kbuf));

		sprintf(kbuf,"%0.3f", hud_omega*DEG);
		TextOut(hDC,(29*cw),(ch*4.7),kbuf,strlen(kbuf));

		sprintf(kbuf,"%0.1f", hud_lambda);
		TextOut(hDC,(29*cw),(ch*5.7),kbuf,strlen(kbuf));
		}

	//	{
	//	char kbuf[256];
	//	sprintf(kbuf,"%d", chute_status);
	//	TextOut(hDC,(35*cw),(ch*12.7),kbuf,strlen(kbuf));
	//	}

		
//		SetTextColor(hDC, 0x00FF6666);	// blue

//		SetTextColor(hDC, 0x00999999);	// white
	}
	else /////////////////////////////////////// page 2
	{
	SelectDefaultFont(hDC,0);
	SetTextColor(hDC, 0x0033FF33);	// green

	TextOut(hDC,(0.5*cw),(ch*2.7),"Nose Hatch", 10);

	TextOut(hDC,(0.5*cw),(ch*6),"Side Hatch", 10);

	TextOut(hDC,(0.5*cw),(ch*9.3),"Nose Cone", 9);

	TextOut(hDC,(0.5*cw),(ch*12.6),"Jettison Payload", 16);
	
	TextOut(hDC,(0.5*cw),(ch*15.9),"Jettison Trunk", 14);

	TextOut(hDC,(0.5*cw),(ch*19.2),"RCS Normal/Reentry", 18);

	//////////////////////////// status
	if(nos_htc_status == 3)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(0.5*cw),(ch*3.7),"OPEN", 4);
		}
	else if(nos_htc_status == 1)
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*3.7),"CLOSED", 6);
		}
	else if(nos_htc_status == 2)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*3.7),"OPENING", 7);
		}
	else if(nos_htc_status == 4)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*3.7),"CLOSING", 7);
		}

	if(sid_htc_status == 3)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(0.5*cw),(ch*7),"OPEN", 4);
		}
	else if(sid_htc_status == 1)
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*7),"CLOSED", 6);
		}
	else if(sid_htc_status == 2)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*7),"OPENING", 7);
		}
	else if(sid_htc_status == 4)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*7),"CLOSING", 7);
		}

	if(nos_con_status == 3)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(0.5*cw),(ch*10.3),"OPEN", 4);
		}
	else if(nos_con_status == 1)
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*10.3),"CLOSED", 6);
		}
	else if(nos_con_status == 2)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*10.3),"OPENING", 7);
		}
	else if(nos_con_status == 4)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(0.5*cw),(ch*10.3),"CLOSING", 7);
		}

	if(pld_status > 1)
		{
		SetTextColor(hDC, 0x0033FF33); 
		TextOut(hDC,(0.5*cw),(ch*13.6),"STOWED", 6);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*13.6),"JETTISONED", 10);
		}

	if(tnk_status > 1)
		{
		SetTextColor(hDC, 0x0033FF33); 
		TextOut(hDC,(0.5*cw),(ch*16.9),"JETTISONED", 10);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*16.9),"ATTACHED", 8);
		}

	if(rcs_status > 1)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(0.5*cw),(ch*20.2),"REENTRY", 7);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(0.5*cw),(ch*20.2),"NORMAL", 6);
		}

	

	/////// righthand side ////////////////////
	SetTextAlign(hDC, TA_RIGHT);
	SetTextColor(hDC, 0x0033FF33);
	TextOut(hDC,(37*cw),(ch*2.7),"Extend/Retract Dock", 19);

	TextOut(hDC,(37*cw),(ch*6),"Cabin Lights", 12);

	TextOut(hDC,(37*cw),(ch*9.3),"NAV Beacons", 11);

	TextOut(hDC,(37*cw),(ch*12.6),"Fwd.Lamp 1", 10);

	TextOut(hDC,(37*cw),(ch*15.9),"Fwd.Lamp 2", 10);

	TextOut(hDC,(37*cw),(ch*19.7),"MFD Display", 11);

	///////////////////////////// status
	if(dok_status == 3)
		{
		SetTextColor(hDC, 0x005555FF); 
		TextOut(hDC,(37*cw),(ch*3.7),"EXTENDED", 8);
		}
	else if(dok_status == 1)
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(37*cw),(ch*3.7),"RETRACTED", 9);
		}
	else if(dok_status == 2)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(37*cw),(ch*3.7),"EXTENDING", 9);
		}
	else if(dok_status == 4)
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(37*cw),(ch*3.7),"RETRACTING", 10);
		}

	if(cab_status > 1)
		{
		SetTextColor(hDC, 0x0033FF33); 
		TextOut(hDC,(37*cw),(ch*7),"OFF", 3);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FFFF);
		TextOut(hDC,(37*cw),(ch*7),"ON", 2);
		}

	if(nav_status > 1)
		{
		SetTextColor(hDC, 0x0033FFFF); 
		TextOut(hDC,(37*cw),(ch*10.3),"ON", 2);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(37*cw),(ch*10.3),"OFF", 3);
		}

	if(lmp1_status > 1)
		{
		SetTextColor(hDC, 0x0033FFFF); 
		TextOut(hDC,(37*cw),(ch*13.6),"ON", 2);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(37*cw),(ch*13.6),"OFF", 3);
		}

	if(lmp2_status > 1)
		{
		SetTextColor(hDC, 0x0033FFFF); 
		TextOut(hDC,(37*cw),(ch*16.9),"ON", 2);
		}
	else
		{	
		SetTextColor(hDC, 0x0033FF33);
		TextOut(hDC,(37*cw),(ch*16.9),"OFF", 3);
		}

	}

}


void  CrewDragonMFD::KillAllRCS ()
{
	vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	vme->SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	vme->SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	vme->SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	vme->SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,0);
	vme->SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
	vme->SetThrusterGroupLevel(THGROUP_MAIN, 0);
}

//void  CrewDragonMFD::CalculateBank()
//{
//	double altd = vme->GetAltitude();
//
//	
//	//if (tgt_dist < 5000000)DoEnergyManagement();
//
//	double angle_diff = req_hdg - curr_hdg;
//	if(tgt_dist > 500000) angle_diff = atan2(-crossrange_error, tgt_dist);
//
//	if (angle_diff > PI) angle_diff = angle_diff-PI2;
//	else if (angle_diff < -PI) angle_diff = angle_diff+PI2;
//
//
//	double crossrange_deadband = 4;
//
//	double req_bank_angle = 80*RAD;
//
//
//	if(altd < 95000)
//	{
//		crossrange_deadband = 3.5;
//		req_bank_angle = (70+bank_correction)*RAD;
//		if (tgt_dist < 5000000)DoEnergyManagement();
//	}
//
//	if(altd < 85000) 
//	{
//		
//	}
//
//	if(altd < 70000)
//	{
//		
//		crossrange_deadband = 2.5;
//		req_bank_angle = (55+bank_correction)*RAD;
//	}
//
//	if(altd < 50000)
//	{
//		crossrange_deadband = 2.00;
//		req_bank_angle = (60+bank_correction)*RAD;
//	}
//
//	if(altd < 30000)
//	{
//		crossrange_deadband = 1.00;
//		req_bank_angle = (30+bank_correction)*RAD;
//	}
//
//
//	if(altd < 25000)
//	{
//		req_bank_angle = 10*angle_diff;
//		if (req_bank_angle > 30*RAD) req_bank_angle = 30*RAD;
//		else if (req_bank_angle < -30*RAD) req_bank_angle = -30*RAD;
//		bank_phase_sw = 3;
//	}
//
//	if(altd < 18000)
//	{
//		bank_phase_sw = 3;
//		req_bank_angle = 0;
//	}
//
//	if (req_bank_angle > 110*RAD) req_bank_angle = 110*RAD;
//	else if (req_bank_angle < 0) req_bank_angle = 0;
//
//	
//	if(bank_phase_sw == 0) // init
//	{
//		if (crossrange_error < 0)bank_phase_sw = 2;
//		else bank_phase_sw = 1;
//		//sprintf(oapiDebugString(), "initbp %d bank_phase_sw %d bank_correction %0.1f",initbp, bank_phase_sw, bank_correction);
//	}
//
//
//	if (bank_phase_sw == 1)
//	{
//		if (angle_diff > (crossrange_deadband*RAD)) bank_phase_sw = 2;
//		bank_angle = -req_bank_angle;
//	}
//	else if (bank_phase_sw == 2)
//	{
//		if (angle_diff < -(crossrange_deadband*RAD)) bank_phase_sw = 1;
//		bank_angle = req_bank_angle;
//	}
//	else if (bank_phase_sw == 3)
//	{
//	bank_angle = req_bank_angle;
//	}
//
//	if(bank_0_sw == 0)bank_0_sw = bank_phase_sw;
//	else if (bank_0_sw != bank_phase_sw)bank_angle = 0;
//		
//	//sprintf(oapiDebugString(), "bank_phase_sw %d bank_correction %0.1f", bank_phase_sw, bank_correction);
//	
//}

//void  CrewDragonMFD::CalculateBankNoTarget()
//{
//	double req_bank_angle = 80*RAD;
//	double myalt = vme->GetAltitude();
//
//	if(myalt < 95000) req_bank_angle = -70*RAD;
//	if(myalt < 85000) req_bank_angle = 70*RAD;
//	if(myalt < 70000) req_bank_angle = -55*RAD;
//	if(myalt < 60000) req_bank_angle = 55*RAD;
//	if(myalt < 50000) req_bank_angle = -60*RAD;
//	if(myalt < 40000) req_bank_angle = 60*RAD;
//	if(myalt < 30000) req_bank_angle = -30*RAD;
//	if(myalt < 25000) req_bank_angle = 15*RAD;
//	if(myalt < 20000) req_bank_angle = 0*RAD;
//
//	bank_angle = req_bank_angle;
//}

void CrewDragonMFD::GetdVNoTarget()
{
	
	VECTOR3 rvel_glob;
	vme->GetRelativeVel(gBod, rvel_glob);
	double v = length(rvel_glob);
	VECTOR3 rpos_glob;
	vme->GetRelativePos(gBod,rpos_glob);
	double r = length(rpos_glob);
	double r2 = r*r;

	normalise(rvel_glob);
	normalise(rpos_glob);
	double l = acos(dotp(rvel_glob,rpos_glob));
	double sinl = sin(l);
	double sinl2 = sinl*sinl;

	//sprintf(oapiDebugString(), "lambda %0.2f ", l*DEG);

	double rp = rt+50000;
	double rp2 = rp*rp;

	double vf2 = 2*mu*((1/rp) - (1/r))*rp2/((r2*sinl2)- rp2);
	double vf = sqrt(vf2);

	dv_required = v - vf;

	//sprintf(oapiDebugString(), "dv_required %0.2f ", dv_required);

}


void CrewDragonMFD::WriteStatus (FILEHANDLE scn) const
{
	
	for (int i = 0; i < (int)strlen(mfdtgtname); i++) 
	{
		if (mfdtgtname[i] == ' ') 
		{
			mfdtgtname[i] = '$';
		}
	}
	char cbuf[256];
	sprintf (cbuf, "%s", mfdtgtname);
	oapiWriteScenario_string (scn, "TGT", cbuf);


	for (int i = 0; i < (int)strlen(vesname); i++) 
	{
		if (vesname[i] == ' ') 
		{
			vesname[i] = '$';
		}
	}
	char bbuf[256];
	sprintf (bbuf, "%s", vesname);
	oapiWriteScenario_string (scn, "VES", bbuf);

	for (int i = 0; i < (int)strlen(gbodyname); i++) 
	{
		if (gbodyname[i] == ' ') 
		{
			gbodyname[i] = '$';
		}
	}
	char abuf[256];
	sprintf (abuf, "%s", gbodyname);
	oapiWriteScenario_string (scn, "GBD", abuf);

	char dbuf[256];
	sprintf (dbuf, "%d %d %d %d %d %d", auto_reentry_sw, auto_burn_sw, burn_sw, bank_mode_sw, bank_dir_sw, reentry_mode);
	oapiWriteScenario_string (scn, "PRM", dbuf);

	char fbuf[256];
	sprintf (fbuf, "%d", add_orbits);
	oapiWriteScenario_string (scn, "ORB", fbuf);

}

void CrewDragonMFD::ReadStatus (FILEHANDLE scn)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "TGT", 3))
			sscanf_s (line+3, "%s", &mfdtgtname);
		else if (!strnicmp (line, "VES", 3))
			sscanf_s (line+3, "%s", &vesname);
		else if (!strnicmp (line, "GBD", 3))
			sscanf_s (line+3, "%s", &gbodyname);
		else if (!strnicmp (line, "PRM", 3))
			sscanf_s (line+3, "%d%d%d%d%d%d", &auto_reentry_sw, &auto_burn_sw, &burn_sw, &bank_mode_sw, &bank_dir_sw, &reentry_mode);
		else if (!strnicmp (line, "ORB", 3))
			sscanf_s (line+3, "%d", &add_orbits);
	}



	for (int i = 0; i < (int)strlen(mfdtgtname); i++) 
	{
		if (mfdtgtname[i] == '$') 
		{
			mfdtgtname[i] = ' ';
			//break;
		}
	}


	for (int i = 0; i < (int)strlen(vesname); i++) 
	{
		if (vesname[i] == '$') 
		{
			vesname[i] = ' ';
			//break;
		}
	}

	for (int i = 0; i < (int)strlen(gbodyname); i++) 
	{
		if (gbodyname[i] == '$') 
		{
			gbodyname[i] = ' ';
			//break;
		}
	}

	///////// init vessel /////////////////////
	if (strlen(vesname) == 0)
	{
	MyVes = oapiGetFocusObject();
	oapiGetObjectName(MyVes, vesname, 30);
	}
	else
	{
	MyVes = oapiGetVesselByName(vesname);
	}
	if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	else
	{
	MyVes = oapiGetFocusObject();
	oapiGetObjectName(MyVes, vesname, 30);
	if(MyVes)vme = (VESSEL2*)oapiGetVesselInterface(MyVes);
	}

	if(MyVes)
	{
	int drg_chk = 0;
	drg_chk = vme->SendBufferedKey(COMM_SEND_STS, true, NULL);
	if(drg_chk > 1)ves_is_dragon = true;
	else ves_is_dragon = false;
//	sprintf(oapiDebugString(), "sts %d", drg_chk);
	}

}


void CrewDragonMFD::StoreStatus (void) const
{
	saveprm.sp1     = add_orbits;
	saveprm.sp2     = ttb_check_sw;
	saveprm.sp3		 = auto_reentry_sw; 
	saveprm.sp4		 = auto_burn_sw;
	saveprm.sp5		 = burn_sw;
	saveprm.sp6		= reentry_mode;
	//sprintf(oapiDebugString(), "storestatus");
}

void CrewDragonMFD::RecallStatus (void)
{
	
	add_orbits		= saveprm.sp1;
	ttb_check_sw	= saveprm.sp2;
	auto_reentry_sw = saveprm.sp3;
	auto_burn_sw	= saveprm.sp4;
	burn_sw			= saveprm.sp5;
	reentry_mode	= saveprm.sp6;
//	sprintf(oapiDebugString(), "recallstatus %0.1f", oapiGetSimTime());
}

void CrewDragonMFD::CalculateFlightVectorReentry()
{
	
	double alt = vme->GetAltitude(ALTMODE_MEANRAD);
	
	VECTOR3 vel_loc, pos_loc; 
	vme->GetAirspeedVector(FRAME_LOCAL, vel_loc);
	double speed = length(vel_loc);

	VECTOR3 vel_unit, pos_loc_unit, nml_unit;

	vel_unit = vel_loc;
	normalise(vel_unit);
	vel_unit = -vel_unit; ////////////////////////////

	vme->HorizonInvRot(_V(0,1,0),pos_loc);
	VECTOR3 vel_pos_nml = crossp(vel_unit, pos_loc);
	
	pos_loc_unit = pos_loc;
	normalise(pos_loc_unit);
	nml_unit = vel_pos_nml;
	normalise(nml_unit);

	double bank_angle = 0;
	double pitch_angle = 10*RAD;

	
	
	
	if(hTgt == NULL) 
	{
		auto_reentry_sw = 2;
		return;
	}
	double gbod_rad = oapiGetSize(gBod);
	VECTOR3 pos_rel,pos_rel_unit, vel_rel, vel_rel_unit;
	vme->GetRelativePos(gBod, pos_rel);
	pos_rel_unit = pos_rel;
	normalise(pos_rel_unit);
	vme->GetGroundspeedVector(FRAME_GLOBAL, vel_rel);
	vel_rel_unit = vel_rel;
	normalise(vel_rel_unit);
	VECTOR3 tgt_pos_rel, tgt_pos_rel_unit;
	oapiGetRelativePos(hTgt, gBod, &tgt_pos_rel);
	tgt_pos_rel_unit = tgt_pos_rel;
	normalise(tgt_pos_rel_unit);

	///////// range
	double tgt_angle = acos(dotp(tgt_pos_rel_unit, pos_rel_unit));
	double tgt_range = tgt_angle*gbod_rad;
	tgt_dist = tgt_range;


	VECTOR3 vel_hrzn;
	vme->GetGroundspeedVector(FRAME_HORIZON, vel_hrzn);

	
	VECTOR3 orb_nml = crossp(vel_rel_unit, pos_rel_unit);
	normalise(orb_nml);
	double cross_range_angle = PI05 - acos(dotp(orb_nml, tgt_pos_rel_unit));
	double cross_range_dist = gbod_rad*cross_range_angle;
	hud_crossrange_error = cross_range_dist;

	if(tgt_dist < 1000000)
	{
	VECTOR3 tgt_pos_glob;
	oapiGetGlobalPos(hTgt, &tgt_pos_glob);
	VECTOR3 tgt_pos_loc;
	vme->Global2Local(tgt_pos_glob, tgt_pos_loc);
	VECTOR3 tgt_pos_hrz;
	vme->HorizonRot(tgt_pos_loc, tgt_pos_hrz);

	double vel_az = atan2(vel_hrzn.x,vel_hrzn.z);
	double tgt_az = atan2(tgt_pos_hrz.x,tgt_pos_hrz.z);
	cross_range_angle = vel_az - tgt_az;
	if(cross_range_angle > PI)cross_range_angle = cross_range_angle - PI2;
	else if(cross_range_angle < -PI)cross_range_angle = cross_range_angle + PI2;

	cross_range_dist = sin(cross_range_angle)*tgt_dist;
	hud_crossrange_error = cross_range_dist;
	}
	

	////// cross range limit ///////////////////////////////////////////////////////

	double cross_range_angle_limit = (0.27*RAD);

	if(tgt_dist < 300000) cross_range_angle_limit = (2*RAD);

	if(abs(cross_range_angle) > cross_range_angle_limit)
	{
	bank_dir_sw = int(abs(cross_range_angle)/-cross_range_angle);
	//sprintf(oapiDebugString(), "bank_mode_sw %d bank_dir_sw %d cra %0.2f", bank_mode_sw, bank_dir_sw, cross_range_angle*DEG);
	}

	if(bank_mode_sw == 0)bank_mode_sw = bank_dir_sw;
	else if (bank_mode_sw != bank_dir_sw) bank_mode_sw = 2;

	//sprintf(oapiDebugString(), "bank_mode_sw %d bank_dir_sw %d cra %0.2f", bank_mode_sw, bank_dir_sw, cross_range_angle*DEG);

	/////////////////////////////////////////////////////////////////////////////////////////////////////

	double hSpeed = length(_V(vel_hrzn.x,0,vel_hrzn.z));
	double vSpeed = vel_hrzn.y;

	double current_h_acc =  (hSpeed - old_hSpeed)/post_step;
	double current_v_acc = (vSpeed - old_vSpeed)/post_step;
	old_hSpeed = hSpeed;
	old_vSpeed = vSpeed;

	//sprintf(oapiDebugString(), "hSpeed %0.2f old_hSpeed %0.2f post_step %0.4f", hSpeed,old_hSpeed,post_step);
	

	double reentry_time = (tgt_dist-200000)/((hSpeed+2750)/2);

	double h_acc_req = 2*((tgt_dist-200000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
	double h_acc_error = h_acc_req - current_h_acc;

	if(reentry_mode == 1)
	{
	if(tgt_dist > 1900000)
		{
		reentry_time = (tgt_dist-1800000)/((hSpeed+7000)/2);
		h_acc_req = 2*((tgt_dist-1800000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
		}
	else if(tgt_dist > 1100000)
		{
		reentry_time = (tgt_dist-1000000)/((hSpeed+6000)/2);
		h_acc_req = 2*((tgt_dist-1000000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
		}
	}
	else if(reentry_mode == 2)
	{
	if(tgt_dist > 2100000)
		{
		reentry_time = (tgt_dist-2000000)/((hSpeed+7850)/2);
		h_acc_req = 2*((tgt_dist-2000000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
		}
	else if(tgt_dist > 1100000)
		{
		reentry_time = (tgt_dist-1000000)/((hSpeed+6000)/2);
		h_acc_req = 2*((tgt_dist-1000000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
		}
	}

	if(tgt_dist < 20000)
	{
		reentry_time = (tgt_dist-0)/((hSpeed+0)/2);
		h_acc_req = 2*((tgt_dist-0) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
	}
	else if(tgt_dist < 250000)
	{
		reentry_time = (tgt_dist-14000)/((hSpeed+0)/2);
		h_acc_req = 2*((tgt_dist-14000) - (hSpeed*reentry_time))/(reentry_time*reentry_time);
		h_acc_error = h_acc_req - current_h_acc;
	}

	double v_speed_req = (10*h_acc_error) + vSpeed_adjust;

	if(reentry_mode == 2 && tgt_dist > 2100000)
	{
	if(v_speed_req > 50)v_speed_req = 50;
	else if(v_speed_req < -150)v_speed_req = -150; // 100
	}
	else
	{
	if(v_speed_req > 200)v_speed_req = 200;
	else if(v_speed_req < -200)v_speed_req = -200;
	}

	double v_speed_error = v_speed_req - vSpeed;
	if(abs(v_speed_error) < 1) vSpeed_adjust = (4*h_acc_error); // 1
//	vSpeed_adjust = (4*h_acc_error);
	double v_acc_req = v_speed_error/20;
	double v_acc_error = v_acc_req - current_v_acc;

	if(alt < 45000)
	{
	v_speed_req = vSpeed + (20*h_acc_error);
	v_speed_error = v_speed_req - vSpeed;
	v_acc_req = v_speed_error/10;
	v_acc_error = v_acc_req - current_v_acc;
	}

	double bank_adjust = 10*abs(v_acc_error);
	if(bank_adjust > 30)bank_adjust = 30;

	double current_bank_angle = DEG*(acos(dotp(nml_unit, _V(-1,0,0))));

	if(v_acc_error > 0.1)bank_angle = current_bank_angle - bank_adjust;
	else if(v_acc_error < -0.1)bank_angle = current_bank_angle + bank_adjust;
	else bank_angle = current_bank_angle;
	
	double min_bank_angle = 15;
	if(abs(cross_range_angle) > (0.25*RAD))bank0_lim = 0.125*RAD;
	else if(abs(cross_range_angle) < (0.125*RAD))bank0_lim = 0.25*RAD;

	if(abs(cross_range_angle) < bank0_lim)min_bank_angle = 0;

	if(tgt_dist < 210000)
	{
		min_bank_angle = 15;
		if(abs(cross_range_angle) > (2*RAD))bank0_lim = 1*RAD;
		else if(abs(cross_range_angle) < (1*RAD))bank0_lim = 2*RAD;

		if(abs(cross_range_angle) < bank0_lim)min_bank_angle = 0;
		
	}

	

	//hud_range = tgt_range;
	//hud_crossrange_error = cross_range_dist;
	//hud_hacc_req = h_acc_req;
	//hud_hacc_current = current_h_acc;
	//hud_hacc_error = h_acc_error;
	int bankangle_int = (int)(bank_angle*1);
	if((bank_angle - bankangle_int) > 0.5)bank_angle = bankangle_int+1;
	else bank_angle = bankangle_int;
	bank_angle = 1*bank_angle;

	if(reentry_mode == 2 && tgt_dist > 2100000)
	{
	if (bank_angle < 5) bank_angle = 5;
	else if (bank_angle > 150) bank_angle = 150;
	}
	else if(tgt_dist < 5000)
	{
	bank_angle = 0;
	}
//	else if(tgt_dist < 15000)
//	{
//	if (bank_angle < 5) bank_angle = 5;
//	else if (bank_angle > 30) bank_angle = 30;
//	}
	else if(tgt_dist < 20000)
	{
	if (bank_angle < 5) bank_angle = 5;
	else if (bank_angle > 70) bank_angle = 70;
	}
	else
	{
	if (bank_angle < min_bank_angle) bank_angle = min_bank_angle;
	else if (bank_angle > 120) bank_angle = 120;
	}
	

	bank_angle = bank_angle*bank_dir_sw*RAD;

	if(bank_mode_sw == 2)
	{
		bank_angle = 0;
		if(reentry_mode == 2 && tgt_dist > 2100000)bank_angle = 180;
	}
	
	if(bank_timer == 0)		new_bank_angle = bank_angle;
	if(abs(v_acc_error) > 4)new_bank_angle = bank_angle;
	if(bank_mode_sw == 2)	new_bank_angle = bank_angle;

	bank_timer = bank_timer + steptime;
	if(bank_timer > 4)bank_timer = 0;
//	double ttt = (tgt_dist - (alt*0.5))/hSpeed;

//	sprintf(oapiDebugString(), "ttt %0.2f cur_bank %0.2f bank %0.2f h_acc_req %0.2f h_acc_curr %0.2f h_acc_error %0.2f v_acc_req %0.2f v_acc_error %0.2f v_acc %0.2f cra %0.2f",ttt, current_bank_angle,bank_angle*DEG, h_acc_req,current_h_acc,h_acc_error,v_acc_req, v_acc_error,current_v_acc,cross_range_angle*DEG);

	MATRIX3 rm = rotm(nml_unit, pitch_angle);
	flight_vector = mul(rm,vel_unit);
	rm = rotm(vel_unit,new_bank_angle);
	flight_vector = mul(rm, flight_vector);
	normalise(flight_vector);

	//sprintf(oapiDebugString(), "bank_mode_sw %d bank_dir_sw %d cra %0.2f", bank_mode_sw, bank_dir_sw, cross_range_angle*DEG);
}

void CrewDragonMFD::CalculateFlightVectorReentryNT()
{
	
	double alt = vme->GetAltitude(ALTMODE_MEANRAD);
	
	VECTOR3 vel_loc, pos_loc; 
	vme->GetAirspeedVector(FRAME_LOCAL, vel_loc);
	double speed = length(vel_loc);

	VECTOR3 vel_unit, pos_loc_unit, nml_unit;

	vel_unit = vel_loc;
	normalise(vel_unit);
	vel_unit = -vel_unit; ////////////////////////////

	vme->HorizonInvRot(_V(0,1,0),pos_loc);
	VECTOR3 vel_pos_nml = crossp(vel_unit, pos_loc);
	
	pos_loc_unit = pos_loc;
	normalise(pos_loc_unit);
	nml_unit = vel_pos_nml;
	normalise(nml_unit);

	double bank_angle = 0;
	double pitch_angle = 10*RAD;


	VECTOR3 vel_hrzn;
	vme->GetGroundspeedVector(FRAME_HORIZON, vel_hrzn);

	bank_dir_sw = 1;
	if(speed > 12000)bank_dir_sw = 1;
	else if(speed > 10000)bank_dir_sw = -1;
	else if(speed > 8000)bank_dir_sw = 1;
	else if(speed > 6000)bank_dir_sw = -1;
	else if(speed > 5000)bank_dir_sw = 1;
	else if(speed > 4000)bank_dir_sw = -1;
	else if(speed > 3000)bank_dir_sw = 1;
	else if(speed > 2000)bank_dir_sw = -1;

	if(bank_mode_sw == 0)bank_mode_sw = bank_dir_sw;
	else if (bank_mode_sw != bank_dir_sw) bank_mode_sw = 2;

	//sprintf(oapiDebugString(), "bank_mode_sw %d bank_dir_sw %d cra %0.2f", bank_mode_sw, bank_dir_sw, cross_range_angle*DEG);

	/////////////////////////////////////////////////////////////////////////////////////////////////////

	double hSpeed = length(_V(vel_hrzn.x,0,vel_hrzn.z));
	double vSpeed = vel_hrzn.y;

	double current_h_acc =  (hSpeed - old_hSpeed)/post_step;
	double current_v_acc = (vSpeed - old_vSpeed)/post_step;
	old_hSpeed = hSpeed;
	old_vSpeed = vSpeed;

	//sprintf(oapiDebugString(), "hSpeed %0.2f old_hSpeed %0.2f post_step %0.4f", hSpeed,old_hSpeed,post_step);
	

	double v_speed_req = (62000 - alt)*0.05;

	
	
	if(v_speed_req > 200)v_speed_req = 200;
	else if(v_speed_req < -300)v_speed_req = -300;
	

	double v_speed_error = v_speed_req - vSpeed;
//	vSpeed_adjust = (4*h_acc_error);
	double v_acc_req = v_speed_error/10;
	double v_acc_error = v_acc_req - current_v_acc;


	double bank_adjust = 20*abs(v_acc_error);
	if(bank_adjust > 30)bank_adjust = 30;

	double current_bank_angle = DEG*(acos(dotp(nml_unit, _V(-1,0,0))));

	if(v_acc_error > 0.1)bank_angle = current_bank_angle - bank_adjust;
	else if(v_acc_error < -0.1)bank_angle = current_bank_angle + bank_adjust;
	else bank_angle = current_bank_angle;
	
	double min_bank_angle = 5;
	double max_bank_angle = 150;

	int bankangle_int = (int)(bank_angle*0.25);
	if((bank_angle - bankangle_int) > 0.5)bank_angle = bankangle_int+1;
	else bank_angle = bankangle_int;
	bank_angle = 4*bank_angle;

	
	
	if (bank_angle < min_bank_angle) bank_angle = min_bank_angle;
	else if (bank_angle > max_bank_angle) bank_angle = max_bank_angle;
	
	bank_angle = bank_angle*bank_dir_sw*RAD;

	if(bank_mode_sw == 2)
	{
		bank_angle = 0;
		if(reentry_mode > 0 && speed > 7700)bank_angle = 180;
	}
	
	////////////////////////////////////// aoa reduction


//	sprintf(oapiDebugString(), "ttt %0.2f cur_bank %0.2f bank %0.2f h_acc_req %0.2f h_acc_curr %0.2f h_acc_error %0.2f v_acc_req %0.2f v_acc_error %0.2f v_acc %0.2f cra %0.2f",ttt, current_bank_angle,bank_angle*DEG, h_acc_req,current_h_acc,h_acc_error,v_acc_req, v_acc_error,current_v_acc,cross_range_angle*DEG);

	MATRIX3 rm = rotm(nml_unit, pitch_angle);
	flight_vector = mul(rm,vel_unit);
	rm = rotm(vel_unit,bank_angle);
	flight_vector = mul(rm, flight_vector);
	normalise(flight_vector);


}




