// ==============================================================
//
// dm2_capsule.cpp
// Control module for dm2_capsule vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include "gcAPI.h"
#include "OrbiterSoundSDK40.h"

// ==============================================================
// Some vessel parameters
// ==============================================================
///////// sound
#define SFTDK		1
#define HRDDK		2
#define HRDDKRAD	3
#define UNDKRAD		4

//////////////
#define AID_CDR1_BUTTONS	1
#define AID_CDR2_BUTTONS	2
#define AID_CDR3_BUTTONS	3
#define AID_CDR4_BUTTONS	4
#define AID_CDR5_BUTTONS	5
#define AID_CDR6_BUTTONS	6
#define AID_CDR7_BUTTONS	7
#define INPUT_BUTTONS		8
#define OV_SWITCH			9
//#define INFO_TAB			3
//#define AID_CDR4_BUTTONS	4
//#define HATCH_BUTTONS		5
//#define RCS_BUTTONS			6
//#define AUTO_BUTTONS		7
//#define TGT_BUTTONS			8
//#define NAV_BUTTONS			9
//#define PANEL_BUTTONS		10
//#define SEP_BUTTONS			14
//#define LIGHT_BUTTONS		15
//#define CHUTE_BUTTONS		16
//
#define AID_CDR1_PWR		11
#define AID_CDR2_PWR		12
#define AID_CDR3_PWR		13
#define AID_CDR4_PWR		14
#define AID_CDR5_PWR		15
#define AID_CDR6_PWR		16
#define AID_CDR7_PWR		17
//
#define GRP_CDR1_VC			64
#define GRP_CDR2_VC			66
#define GRP_CDR3_VC			65
#define GRP_CDR4_VC			60
#define GRP_CDR5_VC			61
#define GRP_CDR6_VC			62
#define GRP_CDR7_VC			63
//
typedef struct {
	HINSTANCE hDLL;
	SURFHANDLE tkbk_label;
	HFONT font[1];
	HPEN pen[1];
} GDIParams;

GDIParams g_Param;

const DWORD COMM_SET_BALLAST = 2305;
//const DWORD COMM_SET_TGT = 2306;
const DWORD COMM_SET_PAC = 2307;
const DWORD COMM_SEND_STS = 2306;
const DWORD COMM_SEND_STS2 = 2308;

const double ISP = 3137;
const double RCSTH = 400;
const double SUPERDRACO_TH = 70000;
const double FUELMASS = 2561;
const double DRYMASS = 7200;
const double TRUNKMASS = 2100;
const double DROGUEDOORMASS = 1;
const double PARADOORMASS = 2;
//const double NOSECONEMASS = 50;
const double HATCH_SPEED = 0.1;
const double DOCK_SPEED = 0.1;
const double NOSE_SPEED = 0.1;

const VECTOR3 DRG_DOOR_OFF = _V(0, -0.946, 2.771);
const VECTOR3 MAIN_DOOR_OFF = _V(0, -1.735, 0.557);
const VECTOR3 TRUNK_OFF = _V(0, -0.00, -2.23);//_V(0, -0.00, -2.33);
const VECTOR3 CHUTES_OFF = _V(0, -1.05, 2.39);
const VECTOR3 DROGUE_OFF = _V(0, -1.061, 2.32);
//const VECTOR3 CABIN_OFF = _V(0, 0, 0);
//const VECTOR3 DOCKCAM_OFF = _V(0.807,0.392,3.1);

//const double mu = 398600439968871.2;
//const double rt = 6371010;
//const VECTOR3 EARTH_AXIS = _V(0, 0.91748206283384, 0.39777715416821);
//const double earth_rot_rate = PI2/86164;
//
//#define GREEN RGB(125, 255, 125)

//VECTOR3 bspos = _V(1.08,0.00,3.00);
VECTOR3 bspos = _V(-0.58,0.86,2.98);
VECTOR3 bscol = _V(1,1,1);
VECTOR3 bsred = _V(1,0.1,0.1);
VECTOR3 bsgrn = _V(0,1,0);

//VECTOR3 beac_al_pos[12];
//beac_al_pos[01] = _V(-1.34, 1.34, 1.46)+TRUNK_OFF;
//beac_al_pos[12] = _V(-1.34, -1.34, 1.46)+TRUNK_OFF;
//VECTOR3 beac_port_3 = _V(-1.34, 1.34,-2.4)+TRUNK_OFF;
//VECTOR3 beac_port_4 = _V(-1.34, -1.34,-2.4)+TRUNK_OFF;
//VECTOR3 beac_port_5 = _V(-2.72, 0,-1.68)+TRUNK_OFF;
//VECTOR3 beac_stbd_1 = _V( 1.34, 1.34, 1.46)+TRUNK_OFF;
//VECTOR3 beac_stbd_2 = _V( 1.34, -1.34, 1.46)+TRUNK_OFF;
//VECTOR3 beac_stbd_3 = _V( 1.34, 1.34,-2.4)+TRUNK_OFF;
//VECTOR3 beac_stbd_4 = _V( 1.34, -1.34,-2.4)+TRUNK_OFF;
//VECTOR3 beac_stbd_5 = _V( 2.72, 0,-1.68)+TRUNK_OFF;
//VECTOR3 beac_upper = _V( 0, 2.72,-1.68)+TRUNK_OFF;
//VECTOR3 beac_lower = _V( 0,-2.72,-1.68)+TRUNK_OFF;
//double ballast = 0;
//char pldname[256];
//char tgt_name[256];
OBJHANDLE hGref;
//int pld_att_count_trans = -1;

int ntdvtx = 6;
//static TOUCHDOWNVTX tdvtx_geardown[4] = {
//	{_V(    0, -3, -0.5), 1e6, 1e6, 0.7, 0.7},
//	{_V(-2.6, 1.5,-0.5), 1e6, 1e6, 0.7, 0.7},
//	{_V( 2.6, 1.5,-0.5), 1e6, 1e6, 0.7, 0.7},
//	{_V( 0, 2 , 0), 1e6, 1e5, 1.5}
//};

static TOUCHDOWNVTX tdvtx_geardown[6] = {
	{_V( 0,-1.8, -0.4), 6e4, 5e3, 3.0, 0.2},
	{_V(-1.5, 1, -0.4), 6e4, 5e3, 3.0, 0.2},
	{_V( 1.5, 1, -0.4), 6e4, 5e3, 3.0, 0.2},
	{_V( 0,-5,  3),   8e4, 1e2, 3.0},
	{_V(-5, 4,  3),   8e4, 1e2, 3.0},
	{_V( 5, 4,  3),   8e4, 1e2, 3.0}
};

////// energy/range profile

//static const int ndp_5000 = 11;
//static const int step_5000 = 200000;
//static const double J_5000[ndp_5000] = 
//{92339210.2,92430780.0,92520463.4,92608138.4,92693648.9,92776785.8,
//92857207.4,92934287.6,93006842.4,93072493.1,93126521.0};
//
//static const int ndp_3000 = 21;
//static const int step_3000 = 100000;
//static const double J_3000[ndp_3000] = 
//{93126521.0,93146561.6,93159732.4,93164665.2,93159465.2,93139998.7,
//93102579.3,93043143.0,92956260.7,92835556.5,92673459.9,92461155.5,
//92188299.2,91843010.0,91411421.6,90880285.4,90234520.5,89459781.1,
//88544624.4,87478575.5,86252267.7};
//
//
//static const int ndp_1000 = 50;
//static const int step_1000 = 20000;
//static const double J_1000[ndp_1000] = 
//{86252267.7,85986985.0,85714845.3,85435759.9,85149650.1,84856747.8,
//84561530.6,84260309.3,83951955.3,83636401.6,83313539.1,82983270.6,
//82645456.7,82299997.5,81946746.6,81585555.6,81216283.1,80838748.5,
//80452828.4,80058295.8,79654971.7,79242688.5,78821208.7,78390333.8,
//77949860.4,77499551.7,77038123.8,76556099.4,76062317.6,75557672.5,
//75041777.2,74510544.4,73964092.4,73400649.9,72819288.5,72219468.8,
//71616831.5,70993342.2,70344756.2,69670095.2,68956500.8,68199856.3,
//67419181.3,66639357.7,65874666.4,65092463.2,64314181.8,63638865.5,
//63025163.0,62569315.8};
////////////////////////////



void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 2.5 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	*cl = 0;
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0};

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = 0.0;
	*cd = 2.5 + oapiGetInducedDrag (*cl, 0.16, 0.2);
	*cl = 0;
}

void VLiftCoeffNeg (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0};


	aoa += PI;
	int idx = max (0, min (15, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	///*cl = 3.5*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cl = 36*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0;
	*cd = 0.00002; //0.4*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
	
}

// 2. horizontal lift component (vertical stabiliser and body)

void HLiftCoeffNeg (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*22.5;
	static const double istep = 1.0/step;
	static const int nabsc = 17;
	static const double CL[nabsc] = {0, -0.1, -0.2, -0.1, 0, -0.1, -0.2, -0.1, 0, 0.1, 0.2, 0.1, 0, 0.1, 0.2, 0.1, 0};

	beta += PI;
	int idx = max (0, min (15, (int)(beta*istep)));
	double d = beta*istep - idx;
	///*cl = 3.5*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cl = 36*(CL[idx] + (CL[idx+1]-CL[idx])*d);
	*cm = 0;
	*cd = 0.00002; //0.4*(0.11 + oapiGetInducedDrag (*cl, 0.16, 0.2));
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class dm2_capsule: public VESSEL3 {
public:
	dm2_capsule (OBJHANDLE hVessel, int flightmodel);
	~dm2_capsule ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	///int clbkConsumeDirectKey(char *kstate);
	bool clbkLoadVC (int id);
	bool clbkLoadGenericCockpit ();
	void clbkVisualCreated (VISHANDLE vis, int refcount);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();
	bool clbkVCMouseEvent (int id, int event, VECTOR3 &p);
	bool clbkVCRedrawEvent (int id, int event, SURFHANDLE surf);
	void clbkMFDMode (int mfd, int mode);
	void clbkDockEvent (int dock, OBJHANDLE mate);
//	void clbkRCSMode (int mode);
//	void clbkNavMode (int mode, bool active);
//	void clbkFocusChanged (bool getfocus, OBJHANDLE hNewVessel, OBJHANDLE hOldVessel);
//	void GetConfigParam (DWORD paramtype);
	int MyID;


private:

	void DefineAnimations (void);
	void CalcMass(void);
	//void DoSolpRot(void);
	//void JettisonNC(void);
	//void JettisonFR(void);
	void JettisonTK(void);
	//void JettisonPD(void);
	void JettisonPayload(void);
	void DeleteTrunkPayload(void);
	void JettisonDrogue(void);
	void JettisonChutes(void);
	void ApplySkin();
	void SetChutes();
	void SetCamPos();
	//void DoNav();
	//void GetFlightTime(void);
	//VECTOR3 RotMat(VECTOR3 CURR_VECT,VECTOR3 ROT_AXIS,double rot_angle);
	void GoToFlightVector(void);
	//void RollToBankAngle(void);
	//void RollToBankAngleOrbital(void);
	//void CalculateFlightVector(void);
	//void CalculateBank(void);
	//void CalculateBankNoTarget(void);
	//void DoReentryBurn(void);
	//void GetTimeToBurn (void);
	void KillAllRCS(void);
	//void DoEnergyManagement(void);
	void Abort(void);

	void RegisterVC_CdrMFD ();
	void RedrawPanel_MFDButton (SURFHANDLE surf, int mfd);
	/*void RedrawHatchButton(SURFHANDLE surf);
	void RedrawRCSButton(SURFHANDLE surf);
	void RedrawAutoButton(SURFHANDLE surf);
	void RedrawInfoTab(SURFHANDLE surf);
	void RedrawTgtButton(SURFHANDLE surf);
	void RedrawNavButton(SURFHANDLE surf);
	void RedrawPanelButton(SURFHANDLE surf);
	void RedrawSepButton(SURFHANDLE surf);
	void RedrawLightButton(SURFHANDLE surf);
	void RedrawChuteButton(SURFHANDLE surf);*/

	THRUSTER_HANDLE th_rcs[28], th_group[8], th_sd[8];
	double th_rcs_level[12];
	PROPELLANT_HANDLE hProp;

	ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2, Section3, Section4, Section5;
	ATTACHMENTHANDLE hHoistattach, hF9attach;
	AIRFOILHANDLE capvert, caphrz, chutevert, chutehrz;
	SURFHANDLE bs1, bs2, bs3, bs4, sd1, sd2, sd3;
	SURFHANDLE hS_panel_d, hS_panel_n, hS_VC_d, hS_VC_n;
	MESHHANDLE hMcap, hMmainsdoor, hMdrgdoor, hMcrew, hMtrunk;
	BEACONLIGHTSPEC beac1, beac2, beac3;
	SpotLight *spotlight1;
	SpotLight *spotlight2;
	PSTREAM_HANDLE hDust;
	//ELEMENTS el;
	//ORBITPARAM prm;
	DOCKHANDLE hDock;
	//NOTEHANDLE hNote;

//private:

	double step_time;
	int pd_status,tk_status;
	int hatch_status, door_status, umb_status, nc_status, ring2_status;
	int chute_deploy_status, drogue_deploy_status, chute_arm_status;
	int abort_status, beac_status, light1_status, light2_status, day_night_status;

	MGROUP_TRANSFORM *umb1, *nc1, *nc2, *nc3;
	MGROUP_TRANSFORM *hatch1, *hatch2, *hatch3, *hatch4, *hatch5, *hatch6, *hatch7;
	MGROUP_TRANSFORM *drg1, *cht1, *cht2, *rotx, *roty;//, *col1;//, *abort1;
	MGROUP_TRANSFORM *door1, *door2, *door3;
	MGROUP_TRANSFORM *ring1, *ring2, *ring3, *ring4, *ring5, *ring6, *ring7, *ring8, *ring9;
	MGROUP_TRANSFORM *ring10, *ring11, *ring12, *ring13, *ring14;
	MGROUP_TRANSFORM *cloak1;
	MGROUP_TRANSFORM *sw_cover_1, *sw_cover_2, *sw_cover_3, *sw_cover_4, *sw_cover_5, *sw_cover_6;
	MGROUP_TRANSFORM *ov_sw;

	UINT anim_ring, anim_cloak, anim_ov_sw;
	UINT anim_ring2, anim_ring3;
	VECTOR3 ring_ref[1];
	UINT anim_hatch, anim_door, anim_umb, anim_nc;//, anim_panel;//, anim_abort;
	double hatch_proc, door_proc, umb_proc, nc_proc, ring_proc, ring2_proc, ring3_proc;//, panel_proc;
	//UINT anim_chute_deploy, anim_chute_collapse, anim_drogue_deploy;
	UINT anim_chute_deploy, anim_drogue_deploy;
	double chute_deploy_proc, drogue_deploy_proc;
	UINT anim_chute_rotx, anim_chute_roty;
	UINT anim_sw_cover_1, anim_sw_cover_2, anim_sw_cover_3, anim_sw_cover_4, anim_sw_cover_5, anim_sw_cover_6;
	int sw_cover_1_status, sw_cover_2_status, sw_cover_3_status, sw_cover_4_status,sw_cover_5_status, sw_cover_6_status;
	double sw_cover_1_proc, sw_cover_2_proc, sw_cover_3_proc, sw_cover_4_proc, sw_cover_5_proc, sw_cover_6_proc;
	double chute_rotx_proc, chute_roty_proc;
	double pld_mass;
	double drgdp, chtdp, vde, old_dnp;
	int skin_sw, landed_sw, rbi, rcs_cfg;
	int ov_sw_status;
	double afchord, afarea, afedit, dust_sw, dust_lvl, rpse_lvl;
//	double tgt_dist, tgt_arc, req_inc, flt_time, crossrange_error, tgt_hdg, req_hdg, curr_hdg;
	VECTOR3 flight_vector;
	VECTOR3 beac1_pos, beac2_pos, beac3_pos;
	int mfd_res;
	int dock_rad_sw;
//	double bank_angle;
//	int auto_reentry_sw, bank_phase_sw, auto_burn_sw, burn_sw;
//	double lambda_2, arc_to_burn, time_to_burn, v_2, v_2pb, ei_radius;
//	int data_sw;
//	double intv;
//	double range_0,post_step, tot_energy_old, dst_past_tld_old;
//	double bank_correction, digi_dar_old, digi_dar, r_3, r_3_fact;
//	int digi_dar_check_sw, ttb_check_sw;
	double post_step;
	double curr_para_angle;
	int cam_status,gc_cam_status, hud_dsp;
	double red_fact, red_fact_adjust;
	DEVMESHHANDLE dhMesh0, dhMesh1, dhMesh2, dhMesh3;
	DEVMESHHANDLE dHMeshCrew[4];
	void SetMeshMat(void);
	void SetDayMat(void);
	void SetNightMat(void);
//	void SetInstMat(void);
	double ballast;
	char tgt_name[256];
	int pld_att_count;
//	VECTOR3 crew_off[7];
	VECTOR3 crew_off[4];
	VECTOR3 campos;
	VECTOR3 dock_pos;
	int crew_count;
	ATTACHMENTHANDLE pld_attach;
	OBJHANDLE hMe;
	VECTOR3 mthr_vrot, mthr_rvel;
	int frame1_sw;
	double rcs_scale;
	//MATERIAL *mat1d, *mat1n, *mat15d, *mat15n, *mat18d, *mat18n, *mat22d, *mat22n;

//	oapi::Font *m_pHudNormalFont;
//	oapi::Pen *m_pHudNormalPen;
};

dm2_capsule::dm2_capsule (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    
//	m_pHudNormalFont = oapiCreateFont(30, false, "Lucida Console", FONT_BOLD);  // fixed-width (prop = false)
//	m_pHudNormalPen = oapiCreatePen(1,15,0x66FFFF);
}

dm2_capsule::~dm2_capsule ()
{
	delete ring1;
	delete ring2;
	delete ring3;
	delete ring4;
	delete ring5;
	delete ring6;
	delete ring7;
	delete ring8;
	delete ring9;
	delete ring10;
	delete ring11;
	delete ring12;
	delete ring13;
	delete ring14;
	delete umb1;
	delete nc1;
	delete nc2;
	delete nc3;
	delete hatch1;
	delete hatch2;
	delete hatch3;
	delete hatch4;
	delete hatch5;
	delete hatch6;
	delete hatch7;
	delete drg1;
	delete cht1;
	delete cht2;
	delete rotx;
	delete roty;
	//delete col1;
	delete door1;
	delete door2;
	delete door3;
	delete cloak1;
	delete sw_cover_1;
	delete sw_cover_2;
	delete sw_cover_3;
	delete sw_cover_4;
	delete sw_cover_5;
	delete sw_cover_6;
	delete ov_sw;

//	delete panel1;
//	delete abort1;
//	delete para_door;
	oapiReleaseTexture(bs1);
	oapiReleaseTexture(bs2);
	oapiReleaseTexture(sd1);
	oapiReleaseTexture(sd2);
	oapiReleaseTexture(sd3);
	oapiReleaseTexture(hS_panel_d);
	oapiReleaseTexture(hS_panel_n);
	oapiReleaseTexture(hS_VC_d);
	oapiReleaseTexture(hS_VC_n);

//	delete bs2;

//	oapiReleaseFont(m_pHudNormalFont);
//	oapiReleasePen(m_pHudNormalPen);
}




void dm2_capsule::DefineAnimations (void)
{
	// abort button
//	anim_abort = CreateAnimation (0);
//	static UINT abortGrp[1] = {35};
//	abort1 = new MGROUP_SCALE  (5, abortGrp, 1, _V(0,0.341,1.591), _V(0.001,0.001,0.001));
//	AddAnimationComponent (anim_abort, 0, 1, abort1);

// panel
//	anim_panel = CreateAnimation (0);
//	static UINT panel1Grp[19] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18};
//	panel1 = new MGROUP_ROTATE (6, panel1Grp, 19, _V(0,1.06,2.02), _V(-1,0,0), (float)(60*RAD));
//	AddAnimationComponent (anim_panel, 0, 1, panel1);
	/////////////////////////////////////////////////////////////

// riser channel cloak group
	anim_cloak = CreateAnimation (0);
	static UINT cloak1Grp[1] = {110};
	cloak1 = new MGROUP_SCALE  (0+crew_count, cloak1Grp, 1, _V(0,0,-0.1), _V(0.0001,0.0001,0.0001));
	AddAnimationComponent (anim_cloak, 0, 1, cloak1);

// docking ring // seq1

	anim_ring = CreateAnimation (0); // ring and petals

	static UINT ring1Grp[3] = {29, 30, 31};
	ring1 = new MGROUP_TRANSLATE (0+crew_count, ring1Grp, 3, _V(0, 0, 1.00));// 0.45
	AddAnimationComponent (anim_ring, 0, 1, ring1);

	anim_ring2 = CreateAnimation (0); // pivot arms

	static UINT ring2Grp[1] = {9};
	ring2 = new MGROUP_ROTATE (0+crew_count, ring2Grp, 1, _V(-0.525, 0.105, 3.124), _V(0.866019, -0.500011, 0), (float)(120*RAD));
	Section0 = AddAnimationComponent (anim_ring2, 0, 1, ring2);//0.02, 0.95

	anim_ring3 = CreateAnimation (0); // tubes and shafts

	static UINT ring3Grp[2] = {10,11};
	ring3 = new MGROUP_ROTATE (0+crew_count, ring3Grp, 2, _V(-0.582, 0.137, 2.943), _V(0.866019, -0.50001, 0), (float)(-100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring3, Section0);// 0.02, 0.95

	static UINT ring4Grp[1] = {12};
	ring4 = new MGROUP_ROTATE (0+crew_count, ring4Grp, 1, _V(-0.582, -0.137, 3.124), _V(-0.866019, -0.500011, 0), (float)(120*RAD));
	Section1 = AddAnimationComponent (anim_ring2, 0, 1, ring4);

	static UINT ring5Grp[2] = {13,14};
	ring5 = new MGROUP_ROTATE (0+crew_count, ring5Grp, 2, _V( -0.582, -0.137, 2.943), _V(-0.866027, -0.499997, 0), (float)(-100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring5, Section1);

	static UINT ring6Grp[1] = {15};
	ring6 = new MGROUP_ROTATE (0+crew_count, ring6Grp, 1, _V(0.172, -0.507, 3.124), _V(0, 1, 0), (float)(120*RAD));
	Section2 = AddAnimationComponent (anim_ring2, 0, 1, ring6);

	static UINT ring7Grp[2] = {16,17};
	ring7 = new MGROUP_ROTATE (0+crew_count, ring7Grp, 2, _V(0.172, -0.507, 2.943), _V(0, 1, 0), (float)(-100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring7, Section2);

	///// ring ref
	ring14 = new MGROUP_ROTATE (LOCALVERTEXLIST, MAKEGROUPARRAY(ring_ref), 1, _V(0.172, 0.54, 2.943), _V(0, 1, 0), (float)(-100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring14, Section2);

	static UINT ring8Grp[1] = {18};
	ring8 = new MGROUP_ROTATE (0+crew_count, ring8Grp, 1, _V(0.353, -0.403, 3.124), _V(-0.866026, 0.499999, 0), (float)(-120*RAD));
	Section3 = AddAnimationComponent (anim_ring2, 0, 1, ring8);

	static UINT ring9Grp[2] = {19,20};
	ring9 = new MGROUP_ROTATE (0+crew_count, ring9Grp, 2, _V(0.353, -0.403, 2.943), _V( -0.866019, 0.500011, 0), (float)(100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring9, Section3);

	static UINT ring10Grp[1] = {21};
	ring10 = new MGROUP_ROTATE (0+crew_count, ring10Grp, 1, _V(0.353, 0.403, 3.124), _V(0.866026, 0.499999, 0), (float)(-120*RAD));
	Section4 = AddAnimationComponent (anim_ring2, 0, 1, ring10);

	static UINT ring11Grp[2] = {22,23};
	ring11 = new MGROUP_ROTATE (0+crew_count, ring11Grp, 2, _V(0.353, 0.403, 2.943), _V(0.866026, 0.499999, 0), (float)(100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring11, Section4);

	static UINT ring12Grp[1] = {24};
	ring12 = new MGROUP_ROTATE (0+crew_count, ring12Grp, 1, _V(0.172, 0.573, 3.124), _V(0, -1, 0), (float)(-120*RAD));
	Section5 = AddAnimationComponent (anim_ring2, 0, 1, ring12);

	static UINT ring13Grp[2] = {25,26};
	ring13 = new MGROUP_ROTATE (0+crew_count, ring13Grp, 2, _V(0.172, 0.573, 2.943), _V(0, -1, 0), (float)(100*RAD));
	AddAnimationComponent (anim_ring3, 0, 1, ring13, Section5);


// nose cone // seq2

	anim_nc = CreateAnimation (0);
	static UINT nc1Grp[1] = {5};
	nc1 = new MGROUP_ROTATE (0+crew_count, nc1Grp, 1, _V(0, 0.821, 2.849), _V(1,0,0), (float)(-45*RAD));
	Section0 = AddAnimationComponent (anim_nc, 0, 1, nc1);

	static UINT nc2Grp[1] = {6};
	nc2 = new MGROUP_ROTATE (0+crew_count, nc2Grp, 1, _V(0, 0.818, 3.124), _V(1,0,0), (float)(45*RAD));
	Section1 = AddAnimationComponent (anim_nc, 0, 1, nc2, Section0);

	static UINT nc3Grp[2] = {7,8};
	nc3 = new MGROUP_ROTATE (0+crew_count, nc3Grp, 2, _V(0, 1.008, 2.85), _V(1,0,0), (float)(-110*RAD));
	AddAnimationComponent (anim_nc, 0, 1, nc3, Section1);

// top hatch // seq3

	anim_hatch = CreateAnimation (0);
	static UINT hatch1Grp[1] = {32};
	hatch1 = new MGROUP_ROTATE (0+crew_count, hatch1Grp, 1, _V(0, -0.396, 2.646), _V(1,0,0), (float)(-120*RAD));
	Section0 = AddAnimationComponent (anim_hatch, 0.5, 1, hatch1);

	static UINT hatch2Grp[1] = {33};
	hatch2 = new MGROUP_ROTATE (0+crew_count, hatch2Grp, 1, _V(0, -0.382, 2.646), _V(1,0,0), (float)(-120.75*RAD));
	AddAnimationComponent (anim_hatch, 0.5, 1, hatch2);

	static UINT hatch3Grp[1] = {34};
	hatch3 = new MGROUP_ROTATE (0+crew_count, hatch3Grp, 1, _V(0, 0.032, 2.645), _V(1,0,0), (float)(50*RAD));
	AddAnimationComponent (anim_hatch, 0.5, 1, hatch3, Section0);

	static UINT hatch4Grp[6] = {35, 36, 37, 41, 42, 125};
	hatch4 = new MGROUP_ROTATE (0+crew_count, hatch4Grp, 6, _V(0, 0.013, 2.627), _V(1,0,0), (float)(180*RAD));
	Section1 = AddAnimationComponent (anim_hatch, 0.5, 1, hatch4, Section0);

	static UINT hatch5Grp[2] = {40, 124};
	hatch5 = new MGROUP_ROTATE (0+crew_count, hatch5Grp, 2, _V(0.003, -0.213, 2.697), _V(0,0,1), (float)(180*RAD));
	Section2 = AddAnimationComponent (anim_hatch, 0.15, 0.4, hatch5, Section1);

	static UINT hatch6Grp[1] = {38};
	hatch6 = new MGROUP_ROTATE (0+crew_count, hatch6Grp, 1, _V(0.084, -0.267, 2.587), _V(-0.642936, -0.76592, 0), (float)(90*RAD));
	AddAnimationComponent (anim_hatch, 0, 0.1, hatch6, Section2);

	static UINT hatch7Grp[1] = {39};
	hatch7 = new MGROUP_ROTATE (0+crew_count, hatch7Grp, 1, _V(0.085, -0.268, 2.712), _V(0.642936, 0.76592, 0), (float)(90*RAD));
	AddAnimationComponent (anim_hatch, 0, 0.1, hatch7, Section2);


// door // seq0
	anim_door = CreateAnimation (0);
	static UINT door1Grp[1] = {0};
	door1 = new MGROUP_ROTATE (0+crew_count, door1Grp, 1, _V(0.356, -1.277, 1.898),_V(-1,0,0), (float)(150*RAD));
	Section1 = AddAnimationComponent (anim_door, 0, 1, door1);

	static UINT door2Grp[1] = {1};
	door2 = new MGROUP_ROTATE(0+crew_count, door2Grp, 1, _V(0.339, -1.183, 2.057),_V(-1,0,0), (float)(143*RAD));
	AddAnimationComponent (anim_door, 0, 1, door2);

	static UINT door3Grp[2] = {2,121};
	door3 = new MGROUP_ROTATE(0+crew_count, door3Grp, 2, _V(0.377, -1.431, 1.396), _V(-1,0,0), (float)(-140*RAD));
	AddAnimationComponent (anim_door, 0, 1, door3, Section1);

// 

// umb
	anim_umb = CreateAnimation (0);
	static UINT umb1Grp[1] = {0};
	//umb1 = new MGROUP_ROTATE  (5, umb1Grp, 1, _V(-1.004,-1.769,1.346), _V(-0.866,0.5,0), (float)(20*RAD));
	umb1 = new MGROUP_ROTATE  (5+crew_count, umb1Grp, 1, _V(-0.865,-1.629,1.559), _V(-0.866,0.5,0), (float)(20*RAD));
	AddAnimationComponent (anim_umb, 0, 1, umb1);

/// drogue deploy
	anim_drogue_deploy = CreateAnimation (0);
	static UINT drgGrp[2] = {0,1};
	drg1 = new MGROUP_SCALE  (2+crew_count, drgGrp, 2, _V(0,0,0.1), _V(0.001,0.001,0.001));
	AddAnimationComponent (anim_drogue_deploy, 0, 1, drg1);

/// chute deploy
	anim_chute_deploy = CreateAnimation (0);
	static UINT chtGrp[2] = {0,1};
	cht1 = new MGROUP_SCALE  (1+crew_count, chtGrp, 2, _V(0,0,-0.00), _V(0.001,0.001,1));
	AddAnimationComponent (anim_chute_deploy, 0, 1, cht1);
	cht2 = new MGROUP_SCALE  (1+crew_count, chtGrp, 2, _V(0,0,-0.00), _V(1,1,0.001));
	AddAnimationComponent (anim_chute_deploy, 0.5, 1, cht2);


/// chute rot
	anim_chute_rotx = CreateAnimation (0);
	rotx = new MGROUP_ROTATE  (1+crew_count, chtGrp, 0, _V(0,0,0.03), _V(-1,0,0), (float)(360*RAD));
	Section4 = AddAnimationComponent (anim_chute_rotx, 0, 1, rotx);

	anim_chute_roty = CreateAnimation (0.5);
	static UINT chtGrp2[2] = {0,1};
	roty = new MGROUP_ROTATE (1+crew_count, chtGrp2, 2, _V(0,0,0.03), _V(0,-1,0), (float)(180*RAD));
	AddAnimationComponent (anim_chute_roty, 0, 1, roty, Section4);

	////// vc switch covers
	anim_sw_cover_1 = CreateAnimation (0);
	static UINT cover1Grp[2] = {54,55};
	sw_cover_1 = new MGROUP_ROTATE (0+crew_count, cover1Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_1, 0, 1, sw_cover_1);

	anim_sw_cover_2 = CreateAnimation (0);
	static UINT cover2Grp[2] = {52,53};
	sw_cover_2 = new MGROUP_ROTATE (0+crew_count, cover2Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_2, 0, 1, sw_cover_2);

	anim_sw_cover_3 = CreateAnimation (0);
	static UINT cover3Grp[2] = {56,57};
	sw_cover_3 = new MGROUP_ROTATE (0+crew_count, cover3Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_3, 0, 1, sw_cover_3);

	anim_sw_cover_4 = CreateAnimation (0);
	static UINT cover4Grp[2] = {46,47};
	sw_cover_4 = new MGROUP_ROTATE (0+crew_count, cover4Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_4, 0, 1, sw_cover_4);

	anim_sw_cover_5 = CreateAnimation (0);
	static UINT cover5Grp[2] = {50,51};
	sw_cover_5 = new MGROUP_ROTATE (0+crew_count, cover5Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_5, 0, 1, sw_cover_5);

	anim_sw_cover_6 = CreateAnimation (0);
	static UINT cover6Grp[2] = {48,49};
	sw_cover_6 = new MGROUP_ROTATE (0+crew_count, cover6Grp, 2, _V(-0.503, 0.402, 2.268), _V(1,0,0), (float)(-184*RAD));
	AddAnimationComponent (anim_sw_cover_6, 0, 1, sw_cover_6);

	anim_ov_sw = CreateAnimation (0);
	static UINT swGrp[1] = {122};
	ov_sw = new MGROUP_ROTATE (0+crew_count, swGrp, 1, _V(0.304, 0.457, 2.286), _V(1,0,0), (float)(-55*RAD));
	AddAnimationComponent (anim_ov_sw, 0, 1, ov_sw);

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void dm2_capsule::clbkSetClassCaps (FILEHANDLE cfg)
{
	mthr_vrot = _V(0,0,0);
	mthr_rvel = _V(0,0,0);
	frame1_sw = 1;
	rcs_scale = 1;

	gcInitialize();
	mfd_res = 256;

	strcpy(tgt_name, "NULL");
	//strcpy(pldname, "NULL");
	dhMesh0 = NULL;
	dhMesh1 = NULL;
	dhMesh2 = NULL;
	dhMesh3 = NULL;
	dHMeshCrew[0] = NULL;
	dHMeshCrew[1] = NULL;
	dHMeshCrew[2] = NULL;
	dHMeshCrew[3] = NULL;

	step_time = 0;
	//panel_status = 0;
	//panel_proc = 0;
	ring2_status = 0;
	ring_proc = 0;
	ring2_proc = 0;
	ring3_proc = 0;
	ring_ref[0] = _V(-0.228, -0.54, 3.174);
	nc_status = 0;
	nc_proc = 0;
	pd_status = 0;
	tk_status = 0;
	hatch_status = 0;
	door_status = 0;
	umb_status = 2;
	hatch_proc = 0;
	door_proc = 0;
	umb_proc = 1;
	rbi = 0;
	rcs_cfg = 0;
	abort_status = 0;
	beac_status = 0;
	light1_status = 0;
	light2_status = 0;
	//laser_status = 0;

	chute_deploy_status = 2;
	drogue_deploy_status = 3;
	chute_deploy_proc = 1;
	//chute_collapse_proc = 0;
	drogue_deploy_proc = 1;

	chute_arm_status = 0;

	chute_rotx_proc = 0;
	chute_roty_proc = 0.5;

	sw_cover_1_status = 0;
	sw_cover_2_status = 0;
	sw_cover_3_status = 0;
	sw_cover_4_status = 0;
	sw_cover_5_status = 0;
	sw_cover_6_status = 0;

	sw_cover_1_proc = 0;
	sw_cover_2_proc = 0;
	sw_cover_3_proc = 0;
	sw_cover_4_proc = 0;
	sw_cover_5_proc = 0;
	sw_cover_6_proc = 0;
	ov_sw_status = 0;

	drgdp = 0;
	chtdp = 0;
	skin_sw = 0;
	vde = 1;
	landed_sw = 0;
	dock_rad_sw = 2;

	afchord = 180;
	afarea = 360;
	afedit = 0.001;
	dust_sw = 0;
	dust_lvl = 0;
	rpse_lvl = 0;

	cam_status = 0;
	gc_cam_status = 0;
	hud_dsp = 0;
	
	pld_mass = 0;
	old_dnp = 0;

	post_step = 0;

	flight_vector = _V(0,0,1);

	/*tgt_dist = 0;
	tgt_arc = 0;
	req_inc = 0;
	flt_time = 0;
	crossrange_error = 0;
	tgt_hdg = 0;
	req_hdg = 0;
	curr_hdg = 0;
	
	bank_angle = 0;
	auto_reentry_sw = 0;
	bank_phase_sw = 0;
	ttb_check_sw = 0;

	lambda_2 = 0;
	arc_to_burn = 0;
	auto_burn_sw = 0;
	burn_sw = 0;
	time_to_burn = 0;
	v_2 = 0;
	v_2pb = 0;
	ei_radius = 0;

	data_sw = 0;
	intv = 200000;
	range_0 = 5000000;
	post_step = 0;
	tot_energy_old = 0;
	dst_past_tld_old = 0;
	bank_correction = 0;
	digi_dar = 0;
	digi_dar_old = 0;
	digi_dar_check_sw = 0;
	r_3 = 6466000;
	r_3_fact = 0;*/
	curr_para_angle = 0;
	red_fact = 0;
	red_fact_adjust = 1;
	day_night_status = 0;

	ballast = 0;
	pld_att_count = -1;

	campos = _V(0,0,0);

	crew_count = 0;
	crew_off[0] = _V(-0.324,0.23,1.619);
	crew_off[1] = _V( 0.324,0.23,1.619);
	crew_off[2] = _V(-0.879,-0.024,1.416);
	crew_off[3] = _V( 0.929,-0.024,1.416);
	//crew_off[4] = _V(0,0.413,0.419);
	//crew_off[5] = _V(-0.675,0.11,0.419);
	//crew_off[6] = _V( 0.675,0.11,0.419);

	dock_pos = _V(0,0.0,3.228);

	for (int i = 0; i < 12; i++)
	{
		th_rcs_level[i] = 0;
	}

	oapiReadItem_int (cfg, "Crew", crew_count);
	if(crew_count > 4)crew_count = 4;
	else if(crew_count < 0)crew_count = 0;

	oapiReadItem_float (cfg, "Exhaust_Scale", rcs_scale);
	
	DefineAnimations ();

	// physical vessel parameters
	SetSize (4);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(1.9,1.9,1.72));//(_V(1.9,1.9,1.72));
	SetCrossSections (_V(8,8,8));
	SetCW(0.7,0.8,1,1);
	SetRotDrag (_V(0.05,0.05,0.025));
	SetTouchdownPoints ( tdvtx_geardown, ntdvtx);
	

	// camera parameters
	SetCameraOffset (campos);
	SetCameraRotationRange(0.9*PI,0.9*PI,0.45*PI,0.45*PI);

	////// check mfd size
	if (gcEnabled() == true)
	{
		FILEHANDLE hCfgFile = oapiOpenFile ("Orbiter_NG.cfg", FILE_IN_ZEROONFAIL, ROOT);
		if(hCfgFile != 0)
		{
			int cfg_mfd_res = 0;
			oapiReadItem_int (hCfgFile, "VCMfdSize", cfg_mfd_res);
			if (cfg_mfd_res == 256) mfd_res = 256;
			else if (cfg_mfd_res == 512) mfd_res = 512;
			else if (cfg_mfd_res == 1024) mfd_res = 1024;
		}
		oapiCloseFile (hCfgFile, FILE_IN_ZEROONFAIL);
		red_fact_adjust = 3;
	}
	else
	{
		FILEHANDLE hCfgFile = oapiOpenFile ("Orbiter.cfg", FILE_IN_ZEROONFAIL, ROOT);
		if(hCfgFile != 0)
		{
			int cfg_mfd_res = 0;
			oapiReadItem_int (hCfgFile, "VCMfdSize", cfg_mfd_res);
			if (cfg_mfd_res == 256) mfd_res = 256;
			else if (cfg_mfd_res == 512) mfd_res = 512;
			else if (cfg_mfd_res == 1024) mfd_res = 1024;
		}
		oapiCloseFile (hCfgFile, FILE_IN_ZEROONFAIL);
	}


	////////////// meshes

	hMcrew = oapiLoadMesh("DM2/DM2_Crew");
	for (int i = 0; i < crew_count; i++)
	{
	SetMeshVisibilityMode(AddMesh(hMcrew, &crew_off[i]),MESHVIS_EXTERNAL | MESHVIS_VC);
	}

	///////////////////////////
	hMcap = oapiLoadMesh("DM2/DM2_Capsule_Cupola"); //0
	SetMeshVisibilityMode(AddMesh(hMcap),MESHVIS_EXTERNAL | MESHVIS_VC);

	SetMeshVisibilityMode(AddMesh("DM2/DM2_Chutes", &CHUTES_OFF),MESHVIS_EXTERNAL | MESHVIS_VC);//1
	SetMeshVisibilityMode(AddMesh("DM2/DM2_Drogue", &DROGUE_OFF),MESHVIS_EXTERNAL | MESHVIS_VC);//2

	hMmainsdoor = oapiLoadMesh("DM2/DM2_Mains_Cover"); //3
	SetMeshVisibilityMode(AddMesh(hMmainsdoor, &MAIN_DOOR_OFF),MESHVIS_EXTERNAL | MESHVIS_VC);

	hMdrgdoor = oapiLoadMesh("DM2/DM2_Drogue_Cover"); //4
	SetMeshVisibilityMode(AddMesh(hMdrgdoor, &DRG_DOOR_OFF),MESHVIS_EXTERNAL | MESHVIS_VC);

	SetMeshVisibilityMode(AddMesh("DM2/DM2_Trunk", &TRUNK_OFF),MESHVIS_EXTERNAL | MESHVIS_VC); //5

	SetMeshVisibilityMode(AddMesh("DM2/DM2_Capsule_HUD"), MESHVIS_VC); //6
	
	///////// load sd charred trunk mesh
	oapiLoadMeshGlobal("DM2/DM2_Trunk_sd");

	
	////////// load charred textures
	bs1 = oapiLoadTexture ("DM2/Crewed_Dragon2_char.dds", true);
	bs2 = oapiLoadTexture ("DM2/Dragon2HS_char.dds", true);

	////////// load superdraco charred textures
	sd1 = oapiLoadTexture ("DM2/Crewed_Dragon2_sd.dds", true);
	sd2 = oapiLoadTexture ("DM2/Dragon2HS_sd.dds", true);
	sd3 = oapiLoadTexture ("DM2/dragon2trunk_sd.dds", true);

	////////// load day night textures
	hS_panel_d = oapiLoadTexture ("DM2/DM_1panel_d.dds", true);
	hS_panel_n = oapiLoadTexture ("DM2/DM_1panel_n.dds", true);
	hS_VC_d = oapiLoadTexture ("DM2/Dragon2VC_d.dds", true);
	hS_VC_n = oapiLoadTexture ("DM2/Dragon2VC_n.dds", true);

	// propellant resources
	
	hProp = CreatePropellantResource (FUELMASS,FUELMASS);

	// ***************** thruster definitions *******************
	
	th_rcs[0] = CreateThruster (_V( 0, 1, 0), _V( 0, 0, 1), RCSTH, hProp, ISP);
	th_rcs[1] = CreateThruster (_V( 0, 1, 0), _V( 0, 0,-1), RCSTH, hProp, ISP);

	th_rcs[2] = CreateThruster (_V( 0,-1, 0), _V( 0, 0, 1), RCSTH, hProp, ISP);
	th_rcs[3] = CreateThruster (_V( 0,-1, 0), _V( 0, 0,-1), RCSTH, hProp, ISP);

	th_rcs[4] = CreateThruster (_V( 0, 2, 0), _V( 1, 0, 0), 1*RCSTH, hProp, ISP);
	th_rcs[5] = CreateThruster (_V( 0, 2, 0), _V(-1, 0, 0), 1*RCSTH, hProp, ISP);

	th_rcs[6] = CreateThruster (_V( 0,-2, 0), _V( 1, 0, 0), 1*RCSTH, hProp, ISP);
	th_rcs[7] = CreateThruster (_V( 0,-2, 0), _V(-1, 0, 0), 1*RCSTH, hProp, ISP);

	th_rcs[8] = CreateThruster (_V( 1, 0, 0), _V( 0, 0, 1), RCSTH, hProp, ISP);
	th_rcs[9] = CreateThruster (_V( 1, 0, 0), _V( 0, 0,-1), RCSTH, hProp, ISP);

	th_rcs[10] = CreateThruster (_V( -1, 0, 0), _V( 0, 0, 1), RCSTH, hProp, ISP);
	th_rcs[11] = CreateThruster (_V( -1, 0, 0), _V( 0, 0,-1), RCSTH, hProp, ISP);

	th_rcs[12] = CreateThruster (_V( 1, 0, 0), _V( 0, 1, 0), RCSTH, hProp, ISP);
	th_rcs[13] = CreateThruster (_V( 1, 0, 0), _V( 0, -1, 0), RCSTH, hProp, ISP);

	th_rcs[14] = CreateThruster (_V( -1, 0, 0), _V( 0, 1, 0), RCSTH, hProp, ISP);
	th_rcs[15] = CreateThruster (_V( -1, 0, 0), _V( 0, -1, 0), RCSTH, hProp, ISP);

	th_rcs[16] = CreateThruster (_V( 1, 0, 0), _V( 0, 0, 1), 1*RCSTH, hProp, ISP);
	th_rcs[17] = CreateThruster (_V( 1, 0, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);

	th_rcs[18] = CreateThruster (_V( -1, 0, 0), _V( 0, 0, 1), 1*RCSTH, hProp, ISP);
	th_rcs[19] = CreateThruster (_V( -1, 0, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);

	th_rcs[20] = CreateThruster (_V( 0, 0, 1), _V( 1, 0, 0), RCSTH, hProp, ISP);
	th_rcs[21] = CreateThruster (_V( 0, 0, 1), _V(-1, 0, 0), RCSTH, hProp, ISP);

	th_rcs[22] = CreateThruster (_V( 0, 0, -1), _V( 1, 0, 0), RCSTH, hProp, ISP);
	th_rcs[23] = CreateThruster (_V( 0, 0, -1), _V(-1, 0, 0), RCSTH, hProp, ISP);

	th_rcs[24] = CreateThruster (_V( 1, 0, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);
	th_rcs[25] = CreateThruster (_V(-1, 0, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);
	th_rcs[26] = CreateThruster (_V( 0, 1, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);
	th_rcs[27] = CreateThruster (_V( 0,-1, 0), _V( 0, 0,-1), 1*RCSTH, hProp, ISP);



	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[5];
	th_group[1] = th_rcs[6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[16];
	th_group[1] = th_rcs[18];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_FORWARD);

	th_group[0] = th_rcs[17];
	th_group[1] = th_rcs[19];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BACK);

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[14];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

	th_group[0] = th_rcs[13];
	th_group[1] = th_rcs[15];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[20];
	th_group[1] = th_rcs[22];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	th_group[0] = th_rcs[21];
	th_group[1] = th_rcs[23];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[24];
	th_group[1] = th_rcs[25];
	th_group[2] = th_rcs[26];
	th_group[3] = th_rcs[27];
	CreateThrusterGroup (th_group, 4, THGROUP_MAIN);

	//th_group[0] = th_rcs[16];
	//th_group[1] = th_rcs[18];
	//CreateThrusterGroup (th_group, 2, THGROUP_MAIN);


//	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("DM2/Dragon_atrcs");
//	double ex_len = 4;
//	double ex_wid = 0.15;

	SURFHANDLE exhaust = oapiRegisterExhaustTexture ("DM2/exhaustrcs");
	double ex_len = 4.8 * rcs_scale;//4;
	double ex_wid = 0.27 * rcs_scale;//0.15;

	VECTOR3 top_pos = _V(0.9,1.37,1.096);
	VECTOR3 mid_pos = _V(0.85,1.47,0.85);
	VECTOR3 btm_pos = _V(1.086,1.37,0.641);

	VECTOR3 top_th = _V(0.59,0.5774,0.5);
	VECTOR3 mid_th = _V(1,0.2,0.75);
	VECTOR3 btm_th = _V(0.5,0.5,-1);
	normalise(top_th);
	normalise(mid_th);
	normalise(btm_th);

	// pup
	AddExhaust(th_rcs[2],ex_len,ex_wid,_V(top_pos.x,top_pos.y,top_pos.z),_V(-top_th.x,top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[2],ex_len,ex_wid,_V(-top_pos.x,top_pos.y,top_pos.z),_V(top_th.x,top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[1],ex_len,ex_wid,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[1],ex_len,ex_wid,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),exhaust);
	//AddExhaust(th_rcs[2],ex_len,ex_wid,_V(0.85,1.47,0.85),_V(mid_th.x,-mid_th.y,mid_th.z),exhaust);
	//AddExhaust(th_rcs[2],ex_len,ex_wid,_V(-0.85,1.47,0.85),_V(-mid_th.x,-mid_th.y,mid_th.z),exhaust);
	
	//pdown
	AddExhaust(th_rcs[0],ex_len,ex_wid,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[0],ex_len,ex_wid,_V(-top_pos.x,-top_pos.y,top_pos.z),_V(top_th.x,-top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[3],ex_len,ex_wid,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x, btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[3],ex_len,ex_wid,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x, btm_th.y,btm_th.z),exhaust);

	//yaw right
	AddExhaust(th_rcs[9],ex_len,ex_wid,_V(top_pos.x, top_pos.y,top_pos.z),_V(-top_th.x, top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[9],ex_len,ex_wid,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),exhaust);
	
	//yaw left
	AddExhaust(th_rcs[8],ex_len,ex_wid,_V(-top_pos.x, top_pos.y,top_pos.z),_V( top_th.x, top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[8],ex_len,ex_wid,_V(-top_pos.x,-top_pos.y,top_pos.z),_V( top_th.x,-top_th.y,top_th.z),exhaust);
	
	//roll right
	AddExhaust(th_rcs[7],ex_len,ex_wid,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[7],ex_len,ex_wid,_V(mid_pos.x,-mid_pos.y,mid_pos.z),_V( mid_th.x, mid_th.y,mid_th.z),exhaust);

	//roll left
	AddExhaust(th_rcs[6],ex_len,ex_wid,_V( mid_pos.x,mid_pos.y,mid_pos.z),_V( mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[6],ex_len,ex_wid,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),exhaust);

	//forward
	AddExhaust(th_rcs[16],ex_len,ex_wid,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[16],ex_len,ex_wid,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[18],ex_len,ex_wid,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x, btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[18],ex_len,ex_wid,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x, btm_th.y,btm_th.z),exhaust);

	//back
	AddExhaust(th_rcs[17],ex_len,ex_wid,_V( mid_pos.x,mid_pos.y,mid_pos.z),_V( mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[17],ex_len,ex_wid,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[19],ex_len,ex_wid,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[19],ex_len,ex_wid,_V( mid_pos.x,-mid_pos.y,mid_pos.z),_V( mid_th.x, mid_th.y,mid_th.z),exhaust);

	//up
	AddExhaust(th_rcs[12],ex_len,ex_wid,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[12],ex_len,ex_wid,_V(-top_pos.x,-top_pos.y,top_pos.z),_V( top_th.x,-top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[14],ex_len,ex_wid,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[14],ex_len,ex_wid,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),exhaust);

	//down
	AddExhaust(th_rcs[13],ex_len,ex_wid,_V(top_pos.x,top_pos.y,top_pos.z),_V(-top_th.x,top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[13],ex_len,ex_wid,_V(-top_pos.x,top_pos.y,top_pos.z),_V( top_th.x,top_th.y,top_th.z),exhaust);
	AddExhaust(th_rcs[15],ex_len,ex_wid,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x,btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[15],ex_len,ex_wid,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x,btm_th.y,btm_th.z),exhaust);
	
	//right
	AddExhaust(th_rcs[20],ex_len,ex_wid,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[20],ex_len,ex_wid,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[22],ex_len,ex_wid,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x,btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[22],ex_len,ex_wid,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),exhaust);
	
	//left
	AddExhaust(th_rcs[21],ex_len,ex_wid,_V(mid_pos.x,mid_pos.y,mid_pos.z),_V(mid_th.x,-mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[21],ex_len,ex_wid,_V(mid_pos.x,-mid_pos.y,mid_pos.z),_V(mid_th.x, mid_th.y,mid_th.z),exhaust);
	AddExhaust(th_rcs[23],ex_len,ex_wid,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V(btm_th.x,btm_th.y,btm_th.z),exhaust);
	AddExhaust(th_rcs[23],ex_len,ex_wid,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V(btm_th.x,-btm_th.y,btm_th.z),exhaust);

	// main
	AddExhaust(th_rcs[24],ex_len,ex_wid,_V(0.67,0.38,3.23),_V(0,0,1),exhaust);
	AddExhaust(th_rcs[25],ex_len,ex_wid,_V(-0.67,0.38,3.23),_V(0,0,1),exhaust);
	AddExhaust(th_rcs[26],ex_len,ex_wid,_V(0.67,-0.38,3.23),_V(0,0,1),exhaust);
	AddExhaust(th_rcs[27],ex_len,ex_wid,_V(-0.67,-0.38,3.23),_V(0,0,1),exhaust);

	////////////////// rcs pstream ///////////////////////////////////////

	SURFHANDLE exhaustrcs = oapiRegisterParticleTexture ("Contrail1");

//	double lifetime = 0.07;
//	if (gcEnabled() == true)lifetime = 0.11;

	double lifetime = 0.05;
	if (gcEnabled() == true)lifetime = 0.06;


//	PARTICLESTREAMSPEC exhaust_rcs2 = {
//		0, 0.02, 20, 10, 0.01, lifetime, 6, 2, PARTICLESTREAMSPEC::EMISSIVE,
//		PARTICLESTREAMSPEC::LVL_SQRT, 0.0, 1,
//		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
//		exhaustrcs
//	};

	PARTICLESTREAMSPEC exhaust_rcs2 = {
	0, 0.02, 40, 30, 0.1, lifetime, 6, 10, PARTICLESTREAMSPEC::EMISSIVE,
	PARTICLESTREAMSPEC::LVL_PSQRT, 0.8, 1,
	PARTICLESTREAMSPEC::ATM_FLAT, 0.5, 0.5
	};

	// pup
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x,top_pos.y,top_pos.z),_V(-top_th.x,top_th.y,top_th.z),&th_rcs_level[0]);
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x,top_pos.y,top_pos.z),_V(top_th.x,top_th.y,top_th.z),&th_rcs_level[0]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[0]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[0]);

	// pdown
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),&th_rcs_level[1]);
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x,-top_pos.y,top_pos.z),_V(top_th.x,-top_th.y,top_th.z),&th_rcs_level[1]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x, btm_th.y,btm_th.z),&th_rcs_level[1]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x, btm_th.y,btm_th.z),&th_rcs_level[1]);
	
	//yaw right
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x, top_pos.y,top_pos.z),_V(-top_th.x, top_th.y,top_th.z),&th_rcs_level[2]);
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),&th_rcs_level[2]);
	
	//yaw left
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x, top_pos.y,top_pos.z),_V( top_th.x, top_th.y,top_th.z),&th_rcs_level[3]);
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x,-top_pos.y,top_pos.z),_V( top_th.x,-top_th.y,top_th.z),&th_rcs_level[3]);
	
	//roll right
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[4]);
	AddParticleStream(&exhaust_rcs2,_V(mid_pos.x,-mid_pos.y,mid_pos.z),_V( mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[4]);

	//roll left
	AddParticleStream(&exhaust_rcs2,_V( mid_pos.x,mid_pos.y,mid_pos.z),_V( mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[5]);
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[5]);

	//forward
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[6]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[6]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x, btm_th.y,btm_th.z),&th_rcs_level[6]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x, btm_th.y,btm_th.z),&th_rcs_level[6]);

	//back
	AddParticleStream(&exhaust_rcs2,_V( mid_pos.x,mid_pos.y,mid_pos.z),_V( mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[7]);
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[7]);
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[7]);
	AddParticleStream(&exhaust_rcs2,_V( mid_pos.x,-mid_pos.y,mid_pos.z),_V( mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[7]);

	//up
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x,-top_pos.y,top_pos.z),_V(-top_th.x,-top_th.y,top_th.z),&th_rcs_level[8]);
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x,-top_pos.y,top_pos.z),_V( top_th.x,-top_th.y,top_th.z),&th_rcs_level[8]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V( btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[8]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[8]);

	//down
	AddParticleStream(&exhaust_rcs2,_V(top_pos.x,top_pos.y,top_pos.z),_V(-top_th.x,top_th.y,top_th.z),&th_rcs_level[9]);
	AddParticleStream(&exhaust_rcs2,_V(-top_pos.x,top_pos.y,top_pos.z),_V( top_th.x,top_th.y,top_th.z),&th_rcs_level[9]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V( btm_th.x,btm_th.y,btm_th.z),&th_rcs_level[9]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x,btm_th.y,btm_th.z),&th_rcs_level[9]);
	
	//right
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,mid_pos.y,mid_pos.z),_V(-mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[10]);
	AddParticleStream(&exhaust_rcs2,_V(-mid_pos.x,-mid_pos.y,mid_pos.z),_V(-mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[10]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,btm_pos.y,btm_pos.z),_V(-btm_th.x,btm_th.y,btm_th.z),&th_rcs_level[10]);
	AddParticleStream(&exhaust_rcs2,_V(-btm_pos.x,-btm_pos.y,btm_pos.z),_V(-btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[10]);
	
	//left
	AddParticleStream(&exhaust_rcs2,_V(mid_pos.x,mid_pos.y,mid_pos.z),_V(mid_th.x,-mid_th.y,mid_th.z),&th_rcs_level[11]);
	AddParticleStream(&exhaust_rcs2,_V(mid_pos.x,-mid_pos.y,mid_pos.z),_V(mid_th.x, mid_th.y,mid_th.z),&th_rcs_level[11]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,btm_pos.y,btm_pos.z),_V(btm_th.x,btm_th.y,btm_th.z),&th_rcs_level[11]);
	AddParticleStream(&exhaust_rcs2,_V(btm_pos.x,-btm_pos.y,btm_pos.z),_V(btm_th.x,-btm_th.y,btm_th.z),&th_rcs_level[11]);

	// main
	AddExhaustStream(th_rcs[24],_V(0.67,0.38,3.23),&exhaust_rcs2);
	AddExhaustStream(th_rcs[25],_V(-0.67,0.38,3.23),&exhaust_rcs2);
	AddExhaustStream(th_rcs[26],_V(0.67,-0.38,3.23),&exhaust_rcs2);
	AddExhaustStream(th_rcs[27],_V(-0.67,-0.38,3.23),&exhaust_rcs2);

	//// superdracos

	th_sd[0] = CreateThruster (_V( 1, 1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[1] = CreateThruster (_V( 1, -1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[2] = CreateThruster (_V( -1, 1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[3] = CreateThruster (_V( -1, -1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[4] = CreateThruster (_V( 1, 1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[5] = CreateThruster (_V( 1, -1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[6] = CreateThruster (_V( -1, 1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);
	th_sd[7] = CreateThruster (_V( -1, -1, 1), _V( 0, 0, 1), SUPERDRACO_TH, hProp, ISP);

	SURFHANDLE exhaust2 = oapiRegisterExhaustTexture ("exhaust2");

	double ex_len2 = 10;
	double ex_wid2 = 0.5;

	VECTOR3 dir1 = _V(0.439,0.296,-1.1);
	VECTOR3 dir2 = _V(0.484,0.215,-1.1);
	normalise(dir1);
	normalise(dir2);
	
	AddExhaust(th_sd[0],ex_len2,ex_wid2,_V(1.489,1.002,0.891),_V(dir1.x,dir1.y,dir1.z),exhaust2);
	AddExhaust(th_sd[1],ex_len2,ex_wid2,_V(1.489,-1.002,0.891),_V(dir1.x,-dir1.y,dir1.z),exhaust2);
	AddExhaust(th_sd[2],ex_len2,ex_wid2,_V(-1.489,1.002,0.891),_V(-dir1.x,dir1.y,dir1.z),exhaust2);
	AddExhaust(th_sd[3],ex_len2,ex_wid2,_V(-1.489,-1.002,0.891),_V(-dir1.x,-dir1.y,dir1.z),exhaust2);

	AddExhaust(th_sd[4],ex_len2,ex_wid2,_V(1.609,0.788,0.892),_V(dir2.x,dir2.y,dir2.z),exhaust2);
	AddExhaust(th_sd[5],ex_len2,ex_wid2,_V(1.609,-0.788,0.892),_V(dir2.x,-dir2.y,dir2.z),exhaust2);
	AddExhaust(th_sd[6],ex_len2,ex_wid2,_V(-1.609,0.788,0.892),_V(-dir2.x,dir2.y,dir2.z),exhaust2);
	AddExhaust(th_sd[7],ex_len2,ex_wid2,_V(-1.609,-0.788,0.892),_V(-dir2.x,-dir2.y,dir2.z),exhaust2);

	th_group[0] = th_sd[0];
	th_group[1] = th_sd[1];
	th_group[2] = th_sd[2];
	th_group[3] = th_sd[3];
	th_group[4] = th_sd[4];
	th_group[5] = th_sd[5];
	th_group[6] = th_sd[6];
	th_group[7] = th_sd[7];
	CreateThrusterGroup (th_group, 8, THGROUP_HOVER);

	///////////////////////////////////////////////////////
	/////AirFoils

	/////  neg lift airfoils
	//capvert = CreateAirfoil2 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeffNeg, 1.4, 3, 1); // 1 1 1
	//caphrz = CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeffNeg, 1.4, 3, 1); // 1 1 1
	
	//// hull drag
	CreateVariableDragElement (&vde, 2, _V(0,-0.044, 0.25)); // (&vde, 1.4, _V(0,-0.044, 0.25)
	///CreateVariableDragElement (&vde, 1.3, _V(0,0, 0)); 
	CreateVariableDragElement (&vde, 38, _V(0,0, 0)); 
	
	/// drogue drag
	CreateVariableDragElement (&drgdp, 3, _V(0,-0.5,10));

	hDock = CreateDock(dock_pos,_V(0,0,1),_V(0,1,0));

	/////////// attachments
	// f9 attach
	//hF9attach = CreateAttachment(true,_V(0,0,-3.115),_V(0,0,-1),_V(-0.1736, 0.9848, 0), "F9");
	hF9attach = CreateAttachment(true,_V(0,0,-4.08),_V(0,0,-1),_V(0, -1, 0), "F9");

	// recovery hoist
	hHoistattach = CreateAttachment(true,_V(0,0,3.0),_V(0,0,1),_V(0,1,0), "RV");
	
	/// grapple fixture
	//hRMSattach = CreateAttachment(true, _V(0,1.843,0.666), _V(0, 0.983965, 0.178361),_V(0, -0.178361, 0.983965), "XS");
	//hRMSattach = CreateAttachment(true, _V(0,1.643,0.666), _V(0, -0.983965, -0.178361),_V(0, 0.178361, -0.983965), "XS");

	// beacon
	beac1_pos = _V(-1.113, -0.155, 2.723); // red
	beac2_pos = _V(1.116, -0.003, 2.746); // grn
	beac3_pos = _V(-0.015, 0.018, 2.685); // wht

	beac1.shape = BEACONSHAPE_STAR;
	beac1.pos = &beac1_pos;
	beac1.col = &bsred;
	beac1.size = 0.12;
	beac1.falloff = 0.6;
	beac1.period = 0;
	beac1.duration = 0.1;
	beac1.tofs = 0;
	beac1.active = false;
	AddBeacon (&beac1);

	beac2.shape = BEACONSHAPE_STAR;
	beac2.pos = &beac2_pos;
	beac2.col = &bsgrn;
	beac2.size = 0.12;
	beac2.falloff = 0.6;
	beac2.period = 0;
	beac2.duration = 0.1;
	beac2.tofs = 0;
	beac2.active = false;
	AddBeacon (&beac2);

	beac3.shape = BEACONSHAPE_STAR;
	beac3.pos = &beac3_pos;
	beac3.col = &bscol;
	beac3.size = 0.15;
	beac3.falloff = 0.6;
	beac3.period = 1.5;
	beac3.duration = 0.03;
	beac3.tofs = 0;
	beac3.active = false;
	AddBeacon (&beac3);

   /* beac_laser_light.shape = 2;
	beac_laser_light.pos = &beac_laser_light_pos;
	beac_laser_light.col = &bscol;
	beac_laser_light.size = 0.2;
	beac_laser_light.falloff = 0.7;
	beac_laser_light.period = 0;
	beac_laser_light.duration = 2.9;
	beac_laser_light.tofs = 0;
	beac_laser_light.active = false;
	AddBeacon (&beac_laser_light);*/

	////// local light

	COLOUR4 col_d = {1,1,1,0};
	COLOUR4 col_s = {1,1,1,0};
	COLOUR4 col_a = {1,1,1,0};
	
	spotlight1 = (SpotLight*)AddSpotLight(_V(-0.015, 0.018, 2.875), _V(0,0,1),60, 0.061, 0, 0.05, (45*RAD), (120*RAD), col_d, col_s, col_a);
	spotlight1->Activate(false);

	COLOUR4 col_d1 = {0.5,0.5,0.5,0};
	COLOUR4 col_s1 = {1,1,1,0};
	COLOUR4 col_a1 = {0.7,0.7,0.7,0};

	spotlight2 = (SpotLight*)AddSpotLight(_V(-0.446, 1.237, 1.604), _V(0.176, -0.421, 0.89),0.5, 0.021, 0.05, 0.082, (15*RAD), (180*RAD), col_d1, col_s1, col_a1);
	spotlight2->Activate(false);


	//// laser

//	th_laser = CreateThruster (_V( 0, 0, 0), _V(0, 0, -1), 0, hProp, 10000000);
//	SURFHANDLE laser = oapiRegisterExhaustTexture ("dragon2/dragon2_laser");
//	AddExhaust(th_laser,100,0.05,laser_pos,_V(0,0,1),laser);




	/// touchdown dust

	PARTICLESTREAMSPEC dust_pstream = {
		0, 1, 400, 20, 3, 0.5, 3, 10, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0.5, 0.5,
		PARTICLESTREAMSPEC::ATM_PLOG, 0.7e-7, 1,
	};
	hDust = AddParticleStream(&dust_pstream, _V(0,0,-0.0),_V(0,0,-1), &dust_lvl);

	//// add high alt reentry pstream
	SURFHANDLE pstr3 = oapiRegisterParticleTexture ("DM2/reentry_contrail");
	PARTICLESTREAMSPEC rps2 = {
		0, 3, 40, 1, 0.008, 0.2, 2, 10, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_LIN, 0, 1, PARTICLESTREAMSPEC::ATM_FLAT, 0.7, 1,
		pstr3
	};
	AddParticleStream(&rps2, _V(0,0,0),_V(0,0,-1), &rpse_lvl);

	//double lightscale = *(double*)GetConfigParam (CFGPRM_SURFACELIGHTBRT);
}



// ==============================================================
// API callback interface
// ==============================================================
// --------------------------------------------------------------
// Module initialisation
// --------------------------------------------------------------
DLLCLBK void InitModule (HINSTANCE hModule)
{
	// allocate GDI resources
	g_Param.font[0] = CreateFont (-16, 0, 0, 0, 400, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");
}

DLLCLBK void ExitModule (HINSTANCE hModule)
{
	// deallocate GDI resources
	DeleteObject (g_Param.font[0]);
}

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new dm2_capsule (hvessel, flightmodel);
	
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (dm2_capsule*)vessel;
	
}

void dm2_capsule::clbkVisualCreated (VISHANDLE vis, int refcount)
{
dhMesh0 = GetDevMesh(vis, 3+crew_count); // mains cover
dhMesh1 = GetDevMesh(vis, 4+crew_count); // drogue cover
dhMesh2 = GetDevMesh(vis, 0+crew_count); // capsule
if(tk_status == 0) dhMesh3 = GetDevMesh(vis, 5+crew_count); // trunk

if(rpse_lvl > 0.5) SetMeshMat();
else if(red_fact > 0) SetMeshMat();

if(crew_count > 0)
{
for (int i = 0; i < crew_count; i++)
	{
		dHMeshCrew[i] = GetDevMesh(vis, i);
	}

}

///////// day night materials set



if(day_night_status == 1)SetNightMat();
else SetDayMat();

//// set size
if(landed_sw == 0 && drogue_deploy_status < 3) SetSize(32);


}

//void dm2_capsule::clbkFocusChanged (bool getfocus, OBJHANDLE hNewVessel, OBJHANDLE hOldVessel)
//{
//	if(getfocus == false)oapiDelAnnotation(hNote);
//}

void dm2_capsule::clbkDockEvent (int dock, OBJHANDLE mate)
{
	if(dock_rad_sw == 0)
	{
	if(mate != NULL && ring2_status == 0)
		{
			PlayVesselRadioExclusiveWave(MyID,HRDDK);
			dock_rad_sw = 1;
		}
	else if(mate != NULL && ring2_status == 2) PlayVesselRadioExclusiveWave(MyID,SFTDK);
	else if(mate == NULL ) PlayVesselRadioExclusiveWave(MyID,UNDKRAD);
	}
	

}

void dm2_capsule::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	for (int i = 0; i < (int)strlen(tgt_name); i++)
	{
		if (tgt_name[i] == ' ')
		{
			tgt_name[i] = '+';
			break;
		}
	}

	OBJHANDLE hMother = GetDockStatus(hDock);
	if(hMother != NULL)
	{
		VESSEL *v = oapiGetVesselInterface (hMother);
		VESSELSTATUS vs;
		v->GetStatus(vs);
		mthr_vrot = vs.vrot;
		mthr_rvel = vs.rvel;
//		v->GetAngularVel(mthr_vrot);
		
	}

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%d %0.4f %d %0.4f %d %0.4f %d %0.4f", hatch_status, hatch_proc, door_status, door_proc, umb_status, umb_proc, ring2_status, ring_proc);
	oapiWriteScenario_string (scn, "HTCH", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f %d %0.4f %0.4f %d", chute_deploy_status, chute_deploy_proc, chute_roty_proc, chute_rotx_proc, drogue_deploy_status, drogue_deploy_proc, curr_para_angle, chute_arm_status);
	//sprintf (cbuf, "%d %0.4f %0.4f %0.4f %0.4f %d %0.4f %0.4f %d", chute_deploy_status, chute_deploy_proc, chute_roty_proc, chute_rotx_proc, drogue_deploy_status, drogue_deploy_proc, curr_para_angle, chute_arm_status);
	oapiWriteScenario_string (scn, "PARA", cbuf);

	//sprintf (cbuf, "%s %d %d %d %d %d %d %d", tgt_name, auto_reentry_sw, rcs_cfg, auto_burn_sw, burn_sw, rbi, bank_phase_sw, abort_status);
	//oapiWriteScenario_string (scn, "TRGT", cbuf);

	sprintf (cbuf, "%d %d %d",  rcs_cfg,  rbi, abort_status);
	oapiWriteScenario_string (scn, "TRGT", cbuf);

	sprintf (cbuf, "%d %d %0.4f %d %d %0.4f %d %0.1f %0.4f %0.4f", nc_status, pd_status, nc_proc, tk_status, skin_sw, afedit, landed_sw, ballast, dust_sw, red_fact);
	oapiWriteScenario_string (scn, "PYLD", cbuf);

	sprintf (cbuf, "%d %d %d", beac_status, light1_status, light2_status);
	oapiWriteScenario_string (scn, "BEAC", cbuf);

	sprintf (cbuf, "%d %d %d %d %d %d %d", sw_cover_1_status, sw_cover_2_status, sw_cover_3_status, sw_cover_4_status, sw_cover_5_status, sw_cover_6_status, ov_sw_status);
	oapiWriteScenario_string (scn, "CVRS", cbuf);

	sprintf (cbuf, "%0.4f %0.4f %0.4f %0.4f %0.4f %0.4f", sw_cover_1_proc, sw_cover_2_proc, sw_cover_3_proc, sw_cover_4_proc, sw_cover_5_proc, sw_cover_6_proc);
	oapiWriteScenario_string (scn, "CVRP", cbuf);

	sprintf (cbuf, "%0.4f %0.4f", ring2_proc, ring3_proc);
	oapiWriteScenario_string (scn, "RING", cbuf);

	sprintf (cbuf, "%d %d %d %d", day_night_status, cam_status, gc_cam_status, hud_dsp);
	oapiWriteScenario_string (scn, "CABN", cbuf);

	sprintf (cbuf, "%0.5f %0.5f %0.5f", mthr_vrot.x, mthr_vrot.y, mthr_vrot.z);
	oapiWriteScenario_string (scn, "MVRT", cbuf);

	sprintf (cbuf, "%0.5f %0.5f %0.5f", mthr_rvel.x, mthr_rvel.y, mthr_rvel.z);
	oapiWriteScenario_string (scn, "MRVL", cbuf);
}


void dm2_capsule::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!_strnicmp (line, "HTCH", 4)) {
			sscanf (line+4, "%d%lf%d%lf%d%lf%d%lf", &hatch_status, &hatch_proc, &door_status, &door_proc, &umb_status, &umb_proc, &ring2_status, &ring_proc);
		} else if (!_strnicmp (line, "PARA", 4)) {
			sscanf (line+4, "%d%lf%lf%lf%d%lf%lf%d", &chute_deploy_status, &chute_deploy_proc, &chute_roty_proc, &chute_rotx_proc,  &drogue_deploy_status, &drogue_deploy_proc, &curr_para_angle, &chute_arm_status);
			//sscanf (line+4, "%d%lf%lf%lf%lf%d%lf%lf%d", &chute_deploy_status, &chute_deploy_proc, &chute_roty_proc, &chute_rotx_proc, &drogue_deploy_status, &drogue_deploy_proc, &curr_para_angle, &chute_arm_status);
		} else if (!_strnicmp (line, "TRGT", 4)) {
		//	sscanf (line+4, "%s%d%d%d%d%d%d%d", &tgt_name, &auto_reentry_sw, &rcs_cfg, &auto_burn_sw, &burn_sw, &rbi, &bank_phase_sw, &abort_status);
				sscanf (line+4, "%d%d%d", &rcs_cfg,  &rbi,  &abort_status);
		} else if (!_strnicmp (line, "PYLD", 4)) {
			sscanf (line+4, "%d%d%lf%d%d%lf%d%lf%lf%lf", &nc_status, &pd_status, &nc_proc, &tk_status, &skin_sw, &afedit, &landed_sw, &ballast, &dust_sw, &red_fact);
		} else if (!_strnicmp (line, "BEAC", 4)) {
			sscanf (line+4, "%d%d%d", &beac_status, &light1_status, &light2_status);
		} else if (!_strnicmp (line, "CVRS", 4)) {
			sscanf (line+4, "%d%d%d%d%d%d%d", &sw_cover_1_status, &sw_cover_2_status, &sw_cover_3_status, &sw_cover_4_status, &sw_cover_5_status, &sw_cover_6_status, &ov_sw_status);
		} else if (!_strnicmp (line, "CVRP", 4)) {
			sscanf (line+4, "%lf%lf%lf%lf%lf%lf", &sw_cover_1_proc, &sw_cover_2_proc, &sw_cover_3_proc, &sw_cover_4_proc, &sw_cover_5_proc, &sw_cover_6_proc);
		} else if (!_strnicmp (line, "RING", 4)) {
			sscanf (line+4, "%lf%lf", &ring2_proc, &ring3_proc);
		} else if (!_strnicmp (line, "CABN", 4)) {
			sscanf (line+4, "%d%d%d%d", &day_night_status, &cam_status, &gc_cam_status, &hud_dsp);
		} else if (!_strnicmp (line, "MVRT", 4)) {
			sscanf (line+4, "%lf%lf%lf", &mthr_vrot.x, &mthr_vrot.y, &mthr_vrot.z);
		} else if (!_strnicmp (line, "MRVL", 4)) {
			sscanf (line+4, "%lf%lf%lf", &mthr_rvel.x, &mthr_rvel.y, &mthr_rvel.z);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	SetAnimation(anim_sw_cover_1, sw_cover_1_proc);
	SetAnimation(anim_sw_cover_2, sw_cover_2_proc);
	SetAnimation(anim_sw_cover_3, sw_cover_3_proc);
	SetAnimation(anim_sw_cover_4, sw_cover_4_proc);
	SetAnimation(anim_sw_cover_5, sw_cover_5_proc);
	SetAnimation(anim_sw_cover_6, sw_cover_6_proc);
	SetAnimation(anim_ov_sw, ov_sw_status);
	
	if (tk_status == 0)
	{
	SetAnimation(anim_umb, umb_proc);
	}
	SetAnimation(anim_hatch, hatch_proc);
	SetAnimation(anim_door, door_proc);
	SetAnimation(anim_nc, nc_proc);
	SetAnimation(anim_ring, ring_proc);
	SetAnimation(anim_ring2, ring2_proc);
	SetAnimation(anim_ring3, ring3_proc);

	if (drogue_deploy_status > 1) SetAnimation(anim_drogue_deploy, drogue_deploy_proc);
	if (chute_deploy_status > 0) SetAnimation(anim_chute_deploy, chute_deploy_proc);
	if (chute_deploy_status == 0)
	{
	SetAnimation(anim_chute_roty, chute_roty_proc);
	SetAnimation(anim_chute_rotx, chute_rotx_proc);
	//SetAnimation(anim_chute_collapse, chute_collapse_proc);
	}

	if(landed_sw == 1)SetMeshVisibilityMode(1+crew_count, MESHVIS_NEVER);

	if(beac_status == 1)
	{
		beac1.active = true;
		beac2.active = true;
		beac3.active = true;
	}
	

	//if(laser_status == 1)SetThrusterLevel(th_laser, 1);
	
	if(light1_status == 1)
	{
			//beac_laser_light.active = true;
			spotlight1->Activate(true);
	}
	if(light2_status == 1)spotlight2->Activate(true);
	

	//if (hatch_status < 2)DelAttachment(hRMSattach);

	for (int i = 0; i < (int)strlen(tgt_name); i++)
	{
		if (tgt_name[i] == '+')
		{
			tgt_name[i] = ' ';
			break;
		}
	}

}


bool SetBallast(void *id, char *str, void *data)
{
	float new_ballast;
	sscanf(str, "%f", &new_ballast);
	if (new_ballast > 6000)return false;
	else if (new_ballast < 0)return false;
	else
	{
		//ballast = new_ballast;
		//pld_att_count_trans = -1;
		char abuf[256];
		sprintf (abuf, "%0.4f", new_ballast);
		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
		vme->SendBufferedKey(COMM_SET_BALLAST, false, abuf);
		vme->SendBufferedKey(COMM_SET_PAC, false, abuf);
		return true;
	}
}


bool AddPayload(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE hPld = oapiGetObjectByName(cbuf);
	if (hPld == NULL)return false;
	else
	{
	VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
	DWORD ad = vme->AttachmentCount(false);
	if (ad == 0) return false;

	ATTACHMENTHANDLE hPldAtt;

	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hPld);
	DWORD ac = v->AttachmentCount(true);
	if (ac == 0) return false;

	for (DWORD k = 0; k < ad; k++)
		{
			hPldAtt = vme->GetAttachmentHandle(false, k);
			OBJHANDLE hPld2 = vme->GetAttachmentStatus(hPldAtt);
			if (hPld2 == NULL) break;
			else if (k == (ad - 1)) return false;
		}

	for (DWORD j = 0; j < ac; j++)
		{
		ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(true, j);
		OBJHANDLE hOp = v->GetAttachmentStatus(hAtt);
		if (hOp == NULL)
			{
			
			if (vme->AttachChild(hPld, hPldAtt, hAtt) == true)
				{
					char abuf[256];
					sprintf (abuf, "%d", 1);
					vme->SendBufferedKey(COMM_SET_PAC, false, abuf);
					return true;
				}
			else return false;
			}
		else if (j == ac-1) return false;
		}
	
	}
	return false;
}


//bool SetTgt(void *id, char *str, void *data)
//{
//	char cbuf[256];
//	sscanf(str, "%s", &cbuf);
//
//	if (!_strnicmp (cbuf, "NULL", 4))
//	{
//		//strcpy(tgt_name,cbuf);
//		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
//		vme->SendBufferedKey(COMM_SET_TGT, false, cbuf);
//		oapiVCTriggerRedrawArea (-1, TGT_BUTTONS);
//		return true;
//	}
//
//	else
//	{
//	OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
//	if (bbtgt == NULL)bbtgt = oapiGetBaseByName(hGref, cbuf);
//	if (bbtgt == NULL)return false;
//	else 
//		{
//		//strcpy(tgt_name,cbuf);
//		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
//		vme->SendBufferedKey(COMM_SET_TGT, false, cbuf);
//		oapiVCTriggerRedrawArea (-1, TGT_BUTTONS);
//		return true;
//		}
//		return false;
//	}
//}


int dm2_capsule::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

	if (key == COMM_SET_BALLAST) {
			sscanf (kstate, "%lf", &ballast);
			return 1;
       }

	if (key == COMM_SEND_STS) {
			char sbuf[256];
			int status_int = 0;
			int pld_sts = 0;
			if(pld_mass > 0)pld_sts = 1;
			sprintf(sbuf, "%d%d%d%d%d%d", chute_arm_status+1, hatch_status+1, door_status+1, nc_status+1, pld_sts+1, tk_status+1);
			sscanf (sbuf, "%d", &status_int);
			return status_int;
       }

	if (key == COMM_SEND_STS2) {
			char tbuf[256];
			int status_int2 = 0;
			sprintf(tbuf, "%d%d%d%d%d%d",rcs_cfg+1, ring2_status+1,day_night_status+1, beac_status+1, light1_status+1, light2_status+1);
			sscanf (tbuf, "%d", &status_int2);
			return status_int2;
       }

	if (key == COMM_SET_PAC) {
			pld_att_count = -1;
			return 1;
       }

  if (!down) return 0; // only process keydown events

  OBJHANDLE hParent = GetAttachmentStatus(hF9attach);
  if (hParent == NULL) hParent = GetAttachmentStatus(hHoistattach);
 // if (hParent == NULL) hParent = GetAttachmentStatus(hRMSattach);
  if (hParent == NULL) hParent = GetDockStatus(hDock);
  bool h_p = true;
  if(hParent == NULL)h_p = false;

 if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }


 else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		case OAPI_KEY_U: // trunk
		if(h_p == false)
		{
		if (tk_status == 0) umb_status = 1;
		}
		return 1;

		case OAPI_KEY_N: // set cam pos
		if(oapiCockpitMode() == COCKPIT_VIRTUAL)
		{
		if (cam_status == 0) cam_status = 4;
		else if (cam_status == 1) cam_status = 0;
		else if (cam_status == 2) cam_status = 1;
		else if (cam_status == 3) cam_status = 2;
		else  cam_status = 3;
		SetCamPos();
		}
		else
		{
		if (gc_cam_status == 0) gc_cam_status = 1;
		else gc_cam_status = 0;
		SetCamPos();
		}
		return 1;

		//////// vc switch covers
		case OAPI_KEY_1: 
		if (sw_cover_1_status == 0) sw_cover_1_status = 1;
		else if (sw_cover_1_status == 1) sw_cover_1_status = 3;
		else if (sw_cover_1_status == 2)sw_cover_1_status = 3;
		else if (sw_cover_1_status == 3) sw_cover_1_status = 1;
		return 1;

		case OAPI_KEY_2: 
		if (sw_cover_2_status == 0) sw_cover_2_status = 1;
		else if (sw_cover_2_status == 1) sw_cover_2_status = 3;
		else if (sw_cover_2_status == 2)sw_cover_2_status = 3;
		else if (sw_cover_2_status == 3) sw_cover_2_status = 1;
		return 1;

		case OAPI_KEY_3: 
		if (sw_cover_3_status == 0) sw_cover_3_status = 1;
		else if (sw_cover_3_status == 1) sw_cover_3_status = 3;
		else if (sw_cover_3_status == 2)sw_cover_3_status = 3;
		else if (sw_cover_3_status == 3) sw_cover_3_status = 1;
		return 1;

		case OAPI_KEY_4: 
		if (sw_cover_4_status == 0) sw_cover_4_status = 1;
		else if (sw_cover_4_status == 1) sw_cover_4_status = 3;
		else if (sw_cover_4_status == 2)sw_cover_4_status = 3;
		else if (sw_cover_4_status == 3) sw_cover_4_status = 1;
		return 1;

		case OAPI_KEY_5: 
		if (sw_cover_5_status == 0) sw_cover_5_status = 1;
		else if (sw_cover_5_status == 1) sw_cover_5_status = 3;
		else if (sw_cover_5_status == 2)sw_cover_5_status = 3;
		else if (sw_cover_5_status == 3) sw_cover_5_status = 1;
		return 1;

		case OAPI_KEY_6: 
		if (sw_cover_6_status == 0) sw_cover_6_status = 1;
		else if (sw_cover_6_status == 1) sw_cover_6_status = 3;
		else if (sw_cover_6_status == 2)sw_cover_6_status = 3;
		else if (sw_cover_6_status == 3) sw_cover_6_status = 1;
		return 1;

		////////// override sw
		case OAPI_KEY_7: 
		if (ov_sw_status == 0) ov_sw_status = 1;
		else if (ov_sw_status == 1) ov_sw_status = 0;
		return 1;

		case OAPI_KEY_8: 
		if (day_night_status == 0)
		{
			day_night_status = 1;
			SetNightMat();
		}
		else if (day_night_status == 1)
		{
			day_night_status = 0;
			SetDayMat();
		}
		return 1;

		case OAPI_KEY_SPACE: 
		if (hud_dsp == 0)hud_dsp = 1;
		else hud_dsp = 0;
		return 1;

	} 
  }

else { // unmodified keys
    switch (key) {
		case OAPI_KEY_N: // set cam pos
		if(oapiCockpitMode() == COCKPIT_VIRTUAL)
		{
		if (cam_status == 0) cam_status = 1;
		else if (cam_status == 1) cam_status = 2;
		else if (cam_status == 2) cam_status = 3;
		else if (cam_status == 3) cam_status = 4;
		else  cam_status = 0;
		SetCamPos();
		}
		else
		{
		if (gc_cam_status == 0) gc_cam_status = 1;
		else gc_cam_status = 0;
		SetCamPos();
		}
		return 1;

		//case OAPI_KEY_M: // trunk
		//if(h_p == false)
		//{
		//if (tk_status == 0) umb_status = 1;
		//}
		//return 1;

		case OAPI_KEY_J: // jet pld
		JettisonPayload();
	//	oapiVCTriggerRedrawArea (-1, SEP_BUTTONS);
		return 1;

		case OAPI_KEY_K: // nose hatch
		if (hatch_status == 0) hatch_status = 1;
		else if (hatch_status == 1) hatch_status = 3;
		else if (hatch_status == 2)hatch_status = 3;
		else if (hatch_status == 3) hatch_status = 1;
	//	oapiVCTriggerRedrawArea (-1, HATCH_BUTTONS);
		return 1;

		case OAPI_KEY_G: // side door
		if (door_status == 0) door_status = 1;
		else if (door_status == 1) door_status = 3;
		else if (door_status == 2) door_status = 3;
		else if (door_status == 3) door_status = 1;
	//	oapiVCTriggerRedrawArea (-1, HATCH_BUTTONS);
		return 1;

	//	case OAPI_KEY_V: // dockring extend
	//	if (ring2_status == 0 && nc_status == 1) ring2_status = 1;
	//	else if (ring2_status == 0 && nc_status == 2) ring2_status = 1;
	//	else if (ring2_status == 1) ring2_status = 3;
	//	else if (ring2_status == 2) ring2_status = 3;
	//	else if (ring2_status == 3 && nc_status == 1) ring2_status = 1;
	//	else if (ring2_status == 3 && nc_status == 2) ring2_status = 1;
	////	oapiVCTriggerRedrawArea (-1, NAV_BUTTONS);
	//	return 1;

		case OAPI_KEY_B: // nose cone open/close
		if (nc_status == 0) nc_status = 1;
		else if (nc_status == 1)
		{
			if (ring2_status == 1) ring2_status = 3;
			else if (ring2_status == 2) ring2_status = 3;
			nc_status = 3;
		}
		else if (nc_status == 2)
		{
			if (ring2_status == 1) ring2_status = 3;
			else if (ring2_status == 2) ring2_status = 3;
			nc_status = 3;
			DelThrusterGroup(THGROUP_MAIN, false);
			SetThrusterLevel(th_rcs[24], 0);
			SetThrusterLevel(th_rcs[25], 0);
			SetThrusterLevel(th_rcs[26], 0);
			SetThrusterLevel(th_rcs[27], 0);
		}
		else if (nc_status == 3) nc_status = 1;
	//	oapiVCTriggerRedrawArea (-1, HATCH_BUTTONS);
		return 1;

		case OAPI_KEY_W:
		pld_att_count = -1;
		oapiOpenInputBox("Enter Ballast Mass (kg, 0-6000)",SetBallast, 0, 50, 0);
		CalcMass();
		return 1;

		/*case OAPI_KEY_S:
		hGref = GetGravityRef();
		oapiOpenInputBox("Enter Landing Target Name",SetTgt, 0, 50, 0);
		CalcMass();
		oapiVCTriggerRedrawArea (-1, TGT_BUTTONS);
		return 1;*/

		case OAPI_KEY_P:
		if (pld_att_count > 0)
		{
		pld_att_count = -1;
		//hMe = GetHandle();
		oapiOpenInputBox("Enter Payload by Name",AddPayload, 0, 50, 0);
		}
		return 1;

		/*case OAPI_KEY_E: /// auto reentry
		if(h_p == false)
		{
		if(auto_reentry_sw == 0)
		{
			auto_reentry_sw = 1;
			if(auto_burn_sw == 1)auto_burn_sw = 0;
		}
		else
			{
			auto_reentry_sw = 0;
			KillAllRCS ();
			}
		}
		oapiVCTriggerRedrawArea (-1, AUTO_BUTTONS);
		return 1;*/

		case OAPI_KEY_D: /// cfg rcs for reentry
		if(rcs_cfg == 0)rcs_cfg = 1;
		else 
		{
		rcs_cfg = 0;
		KillAllRCS ();
		}
	//	oapiVCTriggerRedrawArea (-1, RCS_BUTTONS);
		return 1;

		case OAPI_KEY_C: // toggle hud info
		if(rbi == 0)rbi = 1;
		else rbi = 0;
		return 1;

		//case OAPI_KEY_U: // auto_burn
		//RESETKEY(kstate,OAPI_KEY_U);
		//if(h_p == false)
		//{
		//if (auto_burn_sw == 0)
		//	{
		//	auto_reentry_sw = 0;
		//	auto_burn_sw = 1;
		//	}
		//else if (auto_burn_sw == 1) 
		//	{
		//	auto_burn_sw = 0;
		//	burn_sw = 0;
		//	KillAllRCS ();
		//	}
		//}
		//oapiVCTriggerRedrawArea (-1, AUTO_BUTTONS);
		//return 1;

		case OAPI_KEY_F1:
		SetCamPos();
		return 1;


	//	case OAPI_KEY_F8: // set inst mesh mat
	//	SetInstMat();
	//	return 1;

		case OAPI_KEY_A:
		RESETKEY(kstate,OAPI_KEY_A);
		if(abort_status == 0)abort_status = 1;
		return 1;

		case OAPI_KEY_L:
		RESETKEY(kstate,OAPI_KEY_L);
		if(chute_arm_status == 0)chute_arm_status = 1;
		else if(chute_arm_status == 1)chute_arm_status = 0;
//		oapiVCTriggerRedrawArea (-1, CHUTE_BUTTONS);
		return 1;

		case OAPI_KEY_1:
		//RESETKEY(kstate,OAPI_KEY_L);
		if(beac_status == 0)
		{
		/*if(tk_status == 0)
			{
			for (int i = 0; i < 4; i++)
				{
				beac_al[i].active = true;
				}
			}*/
		beac1.active = true;
		beac2.active = true;
		beac3.active = true;
		beac_status = 1;
		}
		else if(beac_status == 1)
		{
		/*for (int i = 0; i < 4; i++)
			{
			beac_al[i].active = false;
			}*/
		beac1.active = false;
		beac2.active = false;
		beac3.active = false;
		beac_status = 0;
		}
//		oapiVCTriggerRedrawArea (-1, LIGHT_BUTTONS);
		return 1;

		case OAPI_KEY_2:
		if(light1_status == 0)// && nc_status == 2)
		{
		light1_status = 1;
		//beac_laser_light.active = true;
		spotlight1->Activate(true);
		}
		else if(light1_status == 1)
		{
		light1_status = 0;
	//	beac_laser_light.active = false;
		spotlight1->Activate(false);
		}
	//	oapiVCTriggerRedrawArea (-1, LIGHT_BUTTONS);
		return 1;

		case OAPI_KEY_3:
		if(light2_status == 0)// && nc_status == 2)
		{
		light2_status = 1;
		spotlight2->Activate(true);
		}
		else if(light2_status == 1)
		{
		light2_status = 0;
		spotlight2->Activate(false);
		}
//		oapiVCTriggerRedrawArea (-1, LIGHT_BUTTONS);
		return 1;

	}
  }
  return 0;
}



void dm2_capsule::clbkPostCreation()
{
	////////////// sound
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\Vessel\\attsustain.wav",	REPLACE_MAIN_THRUST);
	RequestLoadVesselWave(MyID,HRDDK,"Sound\\Vessel\\docking.wav",DEFAULT);
	RequestLoadVesselWave(MyID,HRDDKRAD,"Sound\\_CustomVesselsSounds\\DM2\\drg_dock_cplt.wav",DEFAULT);
	RequestLoadVesselWave(MyID,SFTDK,"Sound\\_CustomVesselsSounds\\DM2\\drg_sft_cap.wav",DEFAULT);
	RequestLoadVesselWave(MyID,UNDKRAD,"Sound\\Vessel\\undocking_radio.wav",DEFAULT);
	SoundOptionOnOff(MyID,PLAYDOCKINGSOUND, FALSE);
	//////////////////////

	hMe = GetHandle();

//// set dock
OBJHANDLE hMother = GetDockStatus(hDock);
if(hMother == NULL)
{
VECTOR3 new_dock_pos = dock_pos;
new_dock_pos.z = new_dock_pos.z + (ring_proc * 1.00);//0.45
SetDockParams(hDock, new_dock_pos, _V(0,0,1), _V(0,1,0));
}
else if(ring2_status > 0)
{
VESSEL *v = oapiGetVesselInterface (hMother);
VESSELSTATUS vs;
v->GetStatus(vs);
int nDock;	
nDock = v->DockCount();
for (DWORD j = 0; j < nDock; j++)
	{
	DOCKHANDLE hMotherDock = v->GetDockHandle(j);
	OBJHANDLE hChild = v->GetDockStatus(hMotherDock);
	if(hChild == GetHandle())
		{
		//VECTOR3 mtr_am;
		//v->GetAngularVel(mtr_am);
		Undock(0);
		VECTOR3 new_dock_pos = dock_pos;
		new_dock_pos.z = new_dock_pos.z + (ring_proc * 1.00);//0.45
		SetDockParams(hDock, new_dock_pos, _V(0,0,1), _V(0,1,0));
		Dock(hMother,0,j,2);
		//v->SetAngularVel(mthr_vrot);
		vs.vrot = mthr_vrot;
		vs.rvel = mthr_rvel;
		v->DefSetState(&vs);
		break;
		}
	}

}
	
	/// mains safety
	if(nc_status != 2)
	{
		/*th_group[0] = th_rcs[24];
		th_group[1] = th_rcs[25];
		th_group[2] = th_rcs[26];
		th_group[3] = th_rcs[27];
		CreateThrusterGroup (th_group, 4, THGROUP_MAIN);*/
		DelThrusterGroup (THGROUP_MAIN);
		SetThrusterLevel(th_rcs[24], 0);
		SetThrusterLevel(th_rcs[25], 0);
		SetThrusterLevel(th_rcs[26], 0);
		SetThrusterLevel(th_rcs[27], 0);
	}

	/// mesh management
	if (drogue_deploy_status == 0) 
	{		
		DelMesh(4+crew_count);
		DelMesh(2+crew_count);
	}
	else if(drogue_deploy_status < 3)DelMesh(4+crew_count);
	
	if (tk_status == 1)
		{
		DelMesh(5+crew_count);
		}
	
	if (pd_status == 1) DelMesh(4+crew_count);

	/////////// interior material adjust for D3D7

	/*if (gcEnabled() != true)
	{
	MATERIAL *mat1 = oapiMeshMaterial (hMcap, 9);

	mat1->diffuse.r = 0.1;
	mat1->diffuse.g = 0.1;
	mat1->diffuse.b = 0.1;
	mat1->diffuse.a = 1;
	mat1->ambient.r = 0.1;
	mat1->ambient.g = 0.1;
	mat1->ambient.b = 0.1;
	mat1->ambient.a = 1;
	mat1->emissive.r = 0.9;
	mat1->emissive.g = 0.9;
	mat1->emissive.b = 0.9;
	mat1->emissive.a = 1;
	mat1->specular.r = 0.1;
	mat1->specular.g = 0.1;
	mat1->specular.b = 0.1;
	mat1->specular.a = 1;
	
	if(dhMesh2)oapiSetMaterial(dhMesh2,9, mat1);
	}*/
	/// mass management
	CalcMass();

	//if (hatch_status < 2)DelAttachment(hRMSattach);

	// aerodynamics
	if(chute_deploy_status < 2)
	{
	SetCrossSections (_V(200,200,350));
	chutevert = CreateAirfoil2 (LIFT_VERTICAL, _V(0,-1,82), VLiftCoeff, afchord, afarea, 1);
    chutehrz = CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,-1,82), HLiftCoeff, afchord, afarea, 1);
	SetCrossSections (_V(((150*afedit)+50),((150*afedit)+50),((300*afedit)+50)));
	SetRotDrag (_V(3,3,3));
	SetPMI(_V(16,16,16));
	//SetSize(20);
	}

	else if(drogue_deploy_status == 1)
	{
		SetRotDrag (_V(1,1,1));
		//SetSize(20);
		SetCrossSections (_V(20,20,20));
	}

	// skin
	if (skin_sw == 1)
	{
	//bs3 = oapiGetTextureHandle (hMcap, 1);
	//oapiBlt(bs3, bs1, 0, 0, 0, 0, 2048, 2048);
	oapiSetTexture(hMcap,1,bs1);
	oapiSetTexture(hMcap,8,bs1);
	}
	else if (skin_sw == 2)
	{
	//bs3 = oapiGetTextureHandle (hMcap, 1);
	//oapiBlt(bs3, bs1, 0, 0, 0, 0, 2048, 2048);
	oapiSetTexture(hMcap,1,bs1);
	oapiSetTexture(hMcap,8,bs1);
	//bs4 = oapiGetTextureHandle (hMcap, 2);
	//oapiBlt(bs4, bs2, 0, 0, 0, 0, 1024, 1024);
	oapiSetTexture(hMcap,4,bs2);
	oapiSetTexture(hMmainsdoor,1,bs1);
	oapiSetTexture(hMdrgdoor,1,bs1);
	}

	else if (skin_sw == -2)
	{
	oapiSetTexture(hMcap,8,sd1);
	if(tk_status == 0)oapiSetTexture(hMtrunk,1,sd3);
	}
	else if (skin_sw == -1)
	{
	oapiSetTexture(hMcap,8,sd1);
	if(tk_status == 0)oapiSetTexture(hMtrunk,1,sd3);
	oapiSetTexture(hMcap,4,sd2);
	//oapiSetTexture(hMmainsdoor,1,sd1);
	//oapiSetTexture(hMdrgdoor,1,sd1);
	}

	if (tk_status == 1)
	{
	capvert = CreateAirfoil2 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeffNeg, 1.4, 3, 1); // 1 1 1
	caphrz = CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeffNeg, 1.4, 3, 1); // 1 1 1

	//// reentry stream
	//SURFHANDLE pstr4 = oapiRegisterExhaustTexture ("contrail3");
	//PARTICLESTREAMSPEC rps2 = {
	//	0, 3, 30, 0, 0.004, 0.2, 2, 1, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_FLAT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLIN, 4.5e7, 6e7,
	//	pstr4
	//};
	//AddReentryStream (&rps2);

	//PARTICLESTREAMSPEC rps = {
	//	0, 3, 20, 0, 0.008, 0.4, 20, 2, PARTICLESTREAMSPEC::DIFFUSE,
	//	PARTICLESTREAMSPEC::LVL_FLAT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLIN, 6e8, 3e9
	//};
	//AddReentryStream (&rps);
	}
	if(landed_sw == 1)
	{
	DelAirfoil(chutevert);
	DelAirfoil(chutehrz);
	SetCrossSections (_V(8,8,8));
	//SetSize(3);
	}
	//else if(drogue_deploy_status < 3) SetSize(20);
}



void dm2_capsule::clbkPreStep(double SimT, double SimDT, double MJD)
{
	OBJHANDLE hParent = GetAttachmentStatus(hF9attach);
  if (hParent == NULL) hParent = GetAttachmentStatus(hHoistattach);
 // if (hParent == NULL) hParent = GetAttachmentStatus(hRMSattach);
  bool has_parent = true;
  if(hParent == NULL)has_parent = false;
	  
//	if(has_parent == false)DoNav();
	

	double altd = GetAltitude(ALTMODE_GROUND);
	////////////////////////////////////// reporting 

//	double drg = GetDrag();
	double density = 0;
	if(altd < 76000)density = GetAtmDensity();

//	double velcty = GetAirspeed();
//	double lft = GetLift();
//	double ld = lft/drg;

//	double coed = drg/((density*0.5)*(velcty*velcty)*(10.17));
//	double coel = ld*coed;

	////////////////////////////////////// ref traj
//int rep_sw = 1;
//if(rep_sw == 0)
//{
//	if (tgt_dist < 3000000 && range_0 > 3000000)
//	{		
//	range_0 = 3000000;
//	intv = 100000;
//	data_sw = 0;
//	}
//	if (tgt_dist < 1000000  && range_0 > 1000000)
//	{		
//	range_0 = 1000000;
//	intv = 20000;
//	data_sw = 0;
//	}


//	if (tgt_dist < (range_0-(data_sw*intv)))
//	{
//		VECTOR3 rvelg;
//		GetRelativeVel(hGref,rvelg);
//		double v2 = length(rvelg);
//		v2 = v2*v2;
//
//		double gcr = mu/((rt+altd)*(rt+altd));
//
//		double ke = 0.5*v2;
//		double pe = gcr*(rt+altd);
//
//		double tot_energy = ke+pe;
//
//	
//
//	char resultstr[256];
//	sprintf(resultstr, "%0.1f , %0.1f",tgt_dist, tot_energy);
//
//	char filename[256] = "Dragon_Data.txt";
//
//	FILEHANDLE hFile = oapiOpenFile(filename, FILE_APP, ROOT);
//			if (!hFile)
//			{
//			hFile = oapiOpenFile(filename, FILE_OUT, ROOT);
//			oapiWriteLine(hFile, resultstr);
//			oapiCloseFile(hFile, FILE_OUT);
//			}		
//
//			else
//			{
//			oapiWriteLine(hFile, resultstr);
//			oapiCloseFile(hFile, FILE_APP);
//			}
//			data_sw = data_sw+1;
//	}
//}
//sprintf(oapiDebugString(), "coed %0.3f coel %0.3f ld %0.3f", coed,coel,ld);

////// configure rcs
if (rcs_cfg == 1)
{
double roll_th = GetThrusterGroupLevel(THGROUP_ATT_BANKLEFT) - GetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT);

double bank_rcs_fact = 0.3836;
if(roll_th > 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, (roll_th*bank_rcs_fact));
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	}
else if(roll_th < 0)
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, (-roll_th*bank_rcs_fact));
	}
else
	{
	SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
	SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
	}
}


//////////////// do auto reentry
//if(auto_burn_sw == 0 && has_parent == false)DoReentryBurn();
//if(auto_burn_sw == 1)
//{
//	if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);
//	DoReentryBurn();
//	GoToFlightVector();
//	RollToBankAngle();
//	if(auto_burn_sw == 2)KillAllRCS ();
//}
//
//else if(auto_reentry_sw > 0)
//{
//if (oapiGetTimeAcceleration() > 10) oapiSetTimeAcceleration(10);
//if (auto_reentry_sw == 1)CalculateBank();
//if (auto_reentry_sw == 2)CalculateBankNoTarget();
//CalculateFlightVector();
//GoToFlightVector();
//RollToBankAngle();
//if(GetAltitude() < 15000) 
//	{
//	KillAllRCS ();
//	auto_reentry_sw = 0;
//	oapiVCTriggerRedrawArea (-1, AUTO_BUTTONS);
//	}
//}





/////// set coeff drag and lift re:apollo

if (tk_status == 1)
{
vde = 0.65+(0.35*(altd-50000)/150000);
if (vde > 1)vde = 1;
else if (vde < 0.65)vde = 0.65;

EditAirfoil(capvert,11101,_V(0,0,0),VLiftCoeff,1.4,((1/vde)*3),1);
EditAirfoil(caphrz,11101,_V(0,0,0),VLiftCoeff,1.4,((1/vde)*3),1);

double mch_nmbr = GetMachNumber();
if (altd > 76000) mch_nmbr = 20;
mch_nmbr = (mch_nmbr-10)/4;
if (mch_nmbr < 0)mch_nmbr = 0;
else if (mch_nmbr > 1)mch_nmbr = 1;

double rho_nmbr = ((100000*density) - 3)/10;
if (rho_nmbr < 0)rho_nmbr = 0;
else if (rho_nmbr > 1)rho_nmbr = 1;

rpse_lvl  = rho_nmbr*mch_nmbr;
if (rpse_lvl < 0)rpse_lvl = 0;
else if (rpse_lvl > 1)rpse_lvl = 1;
//sprintf(oapiDebugString(), "rpse_lvl %0.3f", rpse_lvl);
}

//////// payload attachment management
if (pld_att_count == -1)
{
pld_att_count = 0;
DWORD ac = AttachmentCount(false);
if (ac > 0)
	{
	for (DWORD j = 0; j < ac; j++)
		{
		int attind = (ac-(j+1));
		ATTACHMENTHANDLE hAtt = GetAttachmentHandle(false, attind);
		OBJHANDLE hOp = GetAttachmentStatus(hAtt);
		if (hOp == NULL)
			{
			pld_att_count = pld_att_count+1;
			pld_attach = hAtt;
			}
		}
	}
CalcMass();
}
/////////////////////////////// beacons


step_time = SimDT;

//if((light_status+laser_status) > 0)
//{
//if(nc_status != 2)
//	{
//	beac_laser_light.active = false;
//	spotlight1->Activate(false);
//	SetThrusterLevel(th_laser,0);
//	light_status = 0;
//	laser_status = 0;
//	oapiVCTriggerRedrawArea (-1, LIGHT_BUTTONS);
//	}
//}
		
		
/*if (tk_status != 0)
	{
		for (int i = 0; i < 4; i++)
		{
		beac_al[i].active = false;
		}
	}*/	
	



/////////////////////////////////////////////////


if(abort_status == 0)
{
OBJHANDLE hLauncher = GetAttachmentStatus(hF9attach);
if(hLauncher == NULL)abort_status = -1;
}
else if(abort_status > 0)Abort();

//double cs_dynp = GetDynPressure();
//if (cs_dynp > 10000)
//{
//	double shake_factor = 0.01*((cs_dynp-10000)*0.0001)*(oapiRand()-0.5);
//	VECTOR3 shakey_cam_pos = campos;
//	shakey_cam_pos.y = shakey_cam_pos.y+shake_factor;
//	SetCameraOffset(shakey_cam_pos);
//}
////////////////// ANIMATIONS ///////////////

//6 vc switch covers ///

if (sw_cover_1_status == 0)sw_cover_1_proc = 0;
else if (sw_cover_1_status == 1)
{
	double da = SimDT * 1;
	sw_cover_1_proc = sw_cover_1_proc + da;
	if(sw_cover_1_proc > 1)
	{
		sw_cover_1_proc = 1;
		sw_cover_1_status = 2;
	}

}
else if (sw_cover_1_status == 2)sw_cover_1_proc = 1;
else if (sw_cover_1_status == 3)
{
	double da = SimDT * 1;
	sw_cover_1_proc = sw_cover_1_proc - da;
	if(sw_cover_1_proc < 0)
	{
		sw_cover_1_proc = 0;
		sw_cover_1_status = 0;
	}
}
SetAnimation(anim_sw_cover_1, sw_cover_1_proc);
//////

if (sw_cover_2_status == 0)sw_cover_2_proc = 0;
else if (sw_cover_2_status == 1)
{
	double da = SimDT * 1;
	sw_cover_2_proc = sw_cover_2_proc + da;
	if(sw_cover_2_proc > 1)
	{
		sw_cover_2_proc = 1;
		sw_cover_2_status = 2;
	}

}
else if (sw_cover_2_status == 2)sw_cover_2_proc = 1;
else if (sw_cover_2_status == 3)
{
	double da = SimDT * 1;
	sw_cover_2_proc = sw_cover_2_proc - da;
	if(sw_cover_2_proc < 0)
	{
		sw_cover_2_proc = 0;
		sw_cover_2_status = 0;
	}
}
SetAnimation(anim_sw_cover_2, sw_cover_2_proc);
//////

if (sw_cover_3_status == 0)sw_cover_3_proc = 0;
else if (sw_cover_3_status == 1)
{
	double da = SimDT * 1;
	sw_cover_3_proc = sw_cover_3_proc + da;
	if(sw_cover_3_proc > 1)
	{
		sw_cover_3_proc = 1;
		sw_cover_3_status = 2;
	}

}
else if (sw_cover_3_status == 2)sw_cover_3_proc = 1;
else if (sw_cover_3_status == 3)
{
	double da = SimDT * 1;
	sw_cover_3_proc = sw_cover_3_proc - da;
	if(sw_cover_3_proc < 0)
	{
		sw_cover_3_proc = 0;
		sw_cover_3_status = 0;
	}
}
SetAnimation(anim_sw_cover_3, sw_cover_3_proc);
//////

if (sw_cover_4_status == 0)sw_cover_4_proc = 0;
else if (sw_cover_4_status == 1)
{
	double da = SimDT * 1;
	sw_cover_4_proc = sw_cover_4_proc + da;
	if(sw_cover_4_proc > 1)
	{
		sw_cover_4_proc = 1;
		sw_cover_4_status = 2;
	}

}
else if (sw_cover_4_status == 2)sw_cover_4_proc = 1;
else if (sw_cover_4_status == 3)
{
	double da = SimDT * 1;
	sw_cover_4_proc = sw_cover_4_proc - da;
	if(sw_cover_4_proc < 0)
	{
		sw_cover_4_proc = 0;
		sw_cover_4_status = 0;
	}
}
SetAnimation(anim_sw_cover_4, sw_cover_4_proc);
//////

if (sw_cover_5_status == 0)sw_cover_5_proc = 0;
else if (sw_cover_5_status == 1)
{
	double da = SimDT * 1;
	sw_cover_5_proc = sw_cover_5_proc + da;
	if(sw_cover_5_proc > 1)
	{
		sw_cover_5_proc = 1;
		sw_cover_5_status = 2;
	}

}
else if (sw_cover_5_status == 2)sw_cover_5_proc = 1;
else if (sw_cover_5_status == 3)
{
	double da = SimDT * 1;
	sw_cover_5_proc = sw_cover_5_proc - da;
	if(sw_cover_5_proc < 0)
	{
		sw_cover_5_proc = 0;
		sw_cover_5_status = 0;
	}
}
SetAnimation(anim_sw_cover_5, sw_cover_5_proc);
//////

if (sw_cover_6_status == 0)sw_cover_6_proc = 0;
else if (sw_cover_6_status == 1)
{
	double da = SimDT * 1;
	sw_cover_6_proc = sw_cover_6_proc + da;
	if(sw_cover_6_proc > 1)
	{
		sw_cover_6_proc = 1;
		sw_cover_6_status = 2;
	}

}
else if (sw_cover_6_status == 2)sw_cover_6_proc = 1;
else if (sw_cover_6_status == 3)
{
	double da = SimDT * 1;
	sw_cover_6_proc = sw_cover_6_proc - da;
	if(sw_cover_6_proc < 0)
	{
		sw_cover_6_proc = 0;
		sw_cover_6_status = 0;
	}
}
SetAnimation(anim_sw_cover_6, sw_cover_6_proc);
//////

/// overide sw
SetAnimation(anim_ov_sw, ov_sw_status);


// umb if trunk
if (tk_status == 0)
{

/// umb
if (umb_status == 0)
	{
	umb_proc = 0;
	SetAnimation(anim_umb, umb_proc);
	}
else if (umb_status == 1)
	{
		double da = SimDT * 1;
		umb_proc = umb_proc-da;
		if (umb_proc < 0)
		{
		umb_proc = 0;
		umb_status = 0;
		}
		SetAnimation(anim_umb, umb_proc);
		if (umb_status == 0)JettisonTK();
	}
else if (umb_status == 2) SetAnimation(anim_umb, 1.0);
}
//else
//{

//////////////// skin
	if(skin_sw == 0)
	{
	if (GetThrusterGroupLevel(THGROUP_HOVER) > 0.2) skin_sw = -3;
	}

	if (skin_sw < 1)
	{
	/*double new_dnp = GetDynPressure();
	if (new_dnp > 9000)
		{
			skin_sw = 0;
			ApplySkin();
		}
	else if (new_dnp > 7000)
		{
		if (new_dnp < old_dnp)
			{
			skin_sw = 0;
			ApplySkin();
			}
		else old_dnp = new_dnp;
		}*/
		if(rpse_lvl > 0.5)
		{
		skin_sw = 0;
		ApplySkin();
		}
	}
	else if (skin_sw == 1)ApplySkin();
	if (skin_sw == -3)ApplySkin();
	else if (skin_sw == -2)ApplySkin();
	
	////////////// drogue
	if (altd < 3000 && drogue_deploy_status == 1)JettisonDrogue();
	else if (altd < 13700 && drogue_deploy_status == 3)
	{
		if(chute_arm_status == 1)
		{
		drogue_deploy_status = 2;
		SetMeshVisibilityMode(4+crew_count,MESHVIS_NEVER);
		}
	}

//}



/// nose hatch
if (hatch_status == 1)
	{
		double da = SimDT * HATCH_SPEED;
		hatch_proc = hatch_proc+da;
		if (hatch_proc > 1)
		{
		hatch_proc = 1;
		hatch_status = 2;
	//	oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
	//	SetAttachmentParams(hRMSattach, _V(0,1.843,0.666), _V(0, 0.983965, 0.178361),_V(0, -0.178361, 0.983965));
		}
		SetAnimation(anim_hatch, hatch_proc);
	}
else if (hatch_status == 2) 
{
	SetAnimation(anim_hatch, 1.0);
//	SetAttachmentParams(hRMSattach, _V(0,1.843,0.666), _V(0, 0.983965, 0.178361),_V(0, -0.178361, 0.983965));
}
else if (hatch_status == 3)
	{
		double da = SimDT * HATCH_SPEED;
		hatch_proc = hatch_proc-da;
		if (hatch_proc < 0)
		{
		hatch_proc = 0;
		hatch_status = 0;
//		oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
		//DelAttachment(hRMSattach);
		}
		SetAnimation(anim_hatch, hatch_proc);
//		SetAttachmentParams(hRMSattach, _V(0,1.643,0.666), _V(0, -0.983965, -0.178361),_V(0, 0.178361, -0.983965));
	}
if(beac_status == 1 && hatch_status == 0)beac3.active = true;
else beac3.active = false;

/// side door
if (door_status == 0) SetAnimation(anim_door, 0.0);
else if (door_status == 1)
	{
		double da = SimDT * HATCH_SPEED;
		door_proc = door_proc+da;
		if (door_proc > 1)
		{
		door_proc = 1;
		door_status = 2;
//		oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
		}
		SetAnimation(anim_door, door_proc);
	}
else if (door_status == 2) SetAnimation(anim_door, 1.0);
else if (door_status == 3)
	{
		double da = SimDT * HATCH_SPEED;
		door_proc = door_proc-da;
		if (door_proc < 0)
		{
		door_proc = 0;
		door_status = 0;
//		oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
		}
		SetAnimation(anim_door, door_proc);
	}

/// dock ring pivots
if (ring2_status == 0) SetAnimation(anim_ring2, 0.0);
else if (ring2_status == 1)
	{
		//dock_rad_sw = 2;
		double da = SimDT * HATCH_SPEED;
		ring2_proc = ring2_proc+da;
		if (ring2_proc > 1)
		{
		ring2_proc = 1;
		ring2_status = 2;
		}
		SetAnimation(anim_ring2, ring2_proc);
	}
else if (ring2_status == 2) SetAnimation(anim_ring2, 1.0);
else if (ring2_status == 3)
	{
		//dock_rad_sw = 2;
		double da = SimDT * HATCH_SPEED;
		ring2_proc = ring2_proc-da;
		if (ring2_proc < 0)
		{
		ring2_proc = 0;
		ring2_status = 0;
		OBJHANDLE hMother = GetDockStatus(hDock);
		if(hMother != NULL) 
			{
			PlayVesselWave(MyID,HRDDK);
			dock_rad_sw = 1;
			}
		}
		SetAnimation(anim_ring2, ring2_proc);
	}

if(dock_rad_sw == 1)
{
	if(IsPlaying(MyID,HRDDK) == FALSE)
	{
	dock_rad_sw = 0;
	PlayVesselRadioExclusiveWave(MyID,HRDDKRAD);
	}
}

/////////////// vectors for pivots and shafts/tubes/ring

VECTOR3 a = _V(0.172,0,3.124);
VECTOR3 b = _V(0.172,0,2.943);
VECTOR3 c = _V(-0.228,0,3.174);

double tot_dist = -0.4;

VECTOR3 pvt = b - a;
VECTOR3 arm = c - b;
double pvt_l = length(pvt);
double arm_l = length(arm);

double init_theta = asin(tot_dist/arm_l);
double pvt_rot_angle = ring2_proc*(-120*RAD);

double pvt_dx = sin(pvt_rot_angle)*pvt_l;
double req_arm_dx = tot_dist - pvt_dx;
double req_arm_angle = asin(req_arm_dx/arm_l);
double arm_pvt_angle = req_arm_angle + pvt_rot_angle;
double req_arm_d_angle = arm_pvt_angle - init_theta;

//sprintf(oapiDebugString(), "req_arm_d_angle %0.3f arm_pvt_angle %0.3f init_theta %0.3f", req_arm_d_angle*DEG, arm_pvt_angle*DEG, init_theta*DEG);

ring3_proc = req_arm_d_angle/(-100*RAD);
SetAnimation(anim_ring3, ring3_proc);

//double init_ring_y = 3.174;
//double pvt_end_y = 3.124 - (cos(pvt_rot_angle)*pvt_l);
//double arm_dy = cos(req_arm_angle)*arm_l;
//double arm_end_y = pvt_end_y + arm_dy;
//double ring_dy = arm_end_y - init_ring_y;
//ring_proc = ring_dy*0.92;


ring_proc = (ring_ref[0].z - 3.174);//*0.92;
SetAnimation(anim_ring, ring_proc);

//sprintf(oapiDebugString(), "ring_proc %0.3f", ring_proc);

/////////////// dock port

OBJHANDLE hMother = GetDockStatus(hDock);
if(hMother == NULL)
{
VECTOR3 new_dock_pos = dock_pos;
new_dock_pos.z = new_dock_pos.z + (ring_proc * 1.00);//0.45
SetDockParams(hDock, new_dock_pos, _V(0,0,1), _V(0,1,0));
}
else if (ring2_status == 1 && frame1_sw == 0)
{
dock_rad_sw = 2;
VESSEL *v = oapiGetVesselInterface (hMother);
int nDock;	
nDock = v->DockCount();
for (DWORD j = 0; j < nDock; j++)
	{
	DOCKHANDLE hMotherDock = v->GetDockHandle(j);
	OBJHANDLE hChild = v->GetDockStatus(hMotherDock);
	if(hChild == GetHandle())
		{
//		VECTOR3 mtr_am;
//		v->GetAngularVel(mtr_am);
		VESSELSTATUS vs;
		v->GetStatus(vs);
		Undock(0);
		VECTOR3 new_dock_pos = dock_pos;
		new_dock_pos.z = new_dock_pos.z + (ring_proc * 1.00);//0.45
		SetDockParams(hDock, new_dock_pos, _V(0,0,1), _V(0,1,0));
		Dock(hMother,0,j,2);
//		v->SetAngularVel(mtr_am);
		v->DefSetState(&vs);
		}
	}

}
else if (ring2_status == 3  && frame1_sw == 0)
{
dock_rad_sw = 2;
VESSEL *v = oapiGetVesselInterface (hMother);
int nDock;	
nDock = v->DockCount();
for (DWORD j = 0; j < nDock; j++)
	{
	DOCKHANDLE hMotherDock = v->GetDockHandle(j);
	OBJHANDLE hChild = v->GetDockStatus(hMotherDock);
	if(hChild == GetHandle())
		{
	//	VECTOR3 mtr_am;
	//	v->GetAngularVel(mtr_am);
		VESSELSTATUS vs;
		v->GetStatus(vs);
		Undock(0);
		VECTOR3 new_dock_pos = dock_pos;
		new_dock_pos.z = new_dock_pos.z + (ring_proc * 1.00);//0.45
		SetDockParams(hDock, new_dock_pos, _V(0,0,1), _V(0,1,0));
		Dock(hMother,0,j,2);
	//	v->SetAngularVel(mtr_am);
		v->DefSetState(&vs);
		}
	}

}

/// nose cone
if (nc_status == 0)SetAnimation(anim_nc, 0);
else if (nc_status == 1)
	{
		double da = SimDT * HATCH_SPEED;
		nc_proc = nc_proc+da;
		if (nc_proc > 1)
		{
		nc_proc = 1;
		nc_status = 2;
		th_group[0] = th_rcs[24];
		th_group[1] = th_rcs[25];
		th_group[2] = th_rcs[26];
		th_group[3] = th_rcs[27];
		CreateThrusterGroup (th_group, 4, THGROUP_MAIN);
	//	oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
		//beac.active = true;
		//beac2.active = true;
		//beac3.active = true;
		}
		SetAnimation(anim_nc, nc_proc);
	}
else if (nc_status == 2) SetAnimation(anim_nc, 1.0);
else if (nc_status == 3)
	{
		double da = SimDT * HATCH_SPEED;
		nc_proc = nc_proc-da;
		if (nc_proc < 0)
		{
		nc_proc = 0;
		nc_status = 0;
//		oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
		//beac.active = false;
		//beac2.active = false;
		//beac3.active = false;
		}
		SetAnimation(anim_nc, nc_proc);
	}

/////// nc retro safety // doesnt work!

if (nc_status != 2)
{
	SetThrusterGroupLevel(THGROUP_MAIN,0);
	SetThrusterLevel(th_rcs[24], 0);
	SetThrusterLevel(th_rcs[25], 0);
	SetThrusterLevel(th_rcs[26], 0);
	SetThrusterLevel(th_rcs[27], 0);
}

///////// chutes
//////drogue

if (drogue_deploy_status > 1)
{
	if (drogue_deploy_status == 3)SetAnimation(anim_drogue_deploy, 1.0);
	else if (drogue_deploy_status == 2)
	{
	if (GetSize() < 10)SetSize(32);

	double da = SimDT * 1;
	if(altd < 2500)da = SimDT * 0.25;
	drogue_deploy_proc = drogue_deploy_proc - da;
	SetCrossSections (_V(((16*drgdp)+4),((16*drgdp)+4),((16*drgdp)+4)));

	if (drogue_deploy_proc < 0)
		{
		drogue_deploy_proc = 0;
		drogue_deploy_status = 1;
		SetRotDrag (_V(1,1,1));
		//if (GetSize() < 10)SetSize(10);
		}
	SetAnimation(anim_drogue_deploy, drogue_deploy_proc);
	}
}


//// chutes deploy
if (chute_deploy_status > 0)
{
	if (chute_deploy_status == 2)SetAnimation(anim_chute_deploy, 1.0);
	else if (chute_deploy_status == 1)
	{
	double da = SimDT * 0.25;
	if(altd < 2500)da = SimDT * 0.125;
	//if (GetSize() < 20)SetSize(20);
	chute_deploy_proc = chute_deploy_proc - da;
	
	SetRotDrag (_V(3,3,3));
	SetPMI(_V(16,16,16));

	EditAirfoil(chutevert,11101,_V(0,1,(2+(chtdp*80))),VLiftCoeff,(afchord*afedit),(afarea*afedit),1);
	EditAirfoil(chutehrz,11101,_V(0,1,(2+(chtdp*80))),VLiftCoeff,(afchord*afedit),(afarea*afedit),1);
	SetCrossSections (_V(((192*afedit)+8),((192*afedit)+8),((342*afedit)+8)));

	if (chute_deploy_proc < 0)
		{
		chute_deploy_proc = 0;
		chute_deploy_status = 0;
		drogue_deploy_proc = 1;
		//SetPMI(_V(200,200,10));
		SetCrossSections (_V(((192*afedit)+8),((192*afedit)+8),((342*afedit)+8)));
		SetRotDrag (_V(1,1,1));
		//SetSize(20);
		}
	SetAnimation(anim_chute_deploy, chute_deploy_proc);
	}
	
}
//if (chute_deploy_status < 2)SetAnimation(anim_para_door, 1);
if (chute_deploy_status < 2)
{
	SetMeshVisibilityMode(3+crew_count,MESHVIS_NEVER);
	SetAnimation(anim_cloak, 0);
}
else SetAnimation(anim_cloak, 1);


drgdp = 1-drogue_deploy_proc;
chtdp = 1-chute_deploy_proc;
if (chute_deploy_status < 2) 
{
	double da = SimDT * 0.05;
	afedit = afedit+da;
	if(afedit > 1)afedit = 1;
}
if (chute_deploy_status == 0) SetChutes();

///////////////// rcs pstreams level ////////////////////////////

th_rcs_level[0] = GetThrusterGroupLevel(THGROUP_ATT_PITCHUP);
th_rcs_level[1] = GetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN);
th_rcs_level[2] = GetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT);
th_rcs_level[3] = GetThrusterGroupLevel(THGROUP_ATT_YAWLEFT);
th_rcs_level[4] = GetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT);
th_rcs_level[5] = GetThrusterGroupLevel(THGROUP_ATT_BANKLEFT);

th_rcs_level[6] = GetThrusterGroupLevel(THGROUP_ATT_FORWARD);
th_rcs_level[7] = GetThrusterGroupLevel(THGROUP_ATT_BACK);
th_rcs_level[8] = GetThrusterGroupLevel(THGROUP_ATT_UP);
th_rcs_level[9] = GetThrusterGroupLevel(THGROUP_ATT_DOWN);
th_rcs_level[10] = GetThrusterGroupLevel(THGROUP_ATT_RIGHT);
th_rcs_level[11] = GetThrusterGroupLevel(THGROUP_ATT_LEFT);

//////// abort button
//if(abort_status == 2)SetAnimation(anim_abort, 1);
//else SetAnimation(anim_abort, 0);
}

void dm2_capsule::clbkPostStep (double simt, double simdt, double mjd)
{
	post_step = simdt;
	if(rpse_lvl > 0.5) SetMeshMat();
	else if(red_fact > 0) SetMeshMat();
	if(dock_rad_sw == 2)dock_rad_sw = 0;
	frame1_sw = 0;
}


bool dm2_capsule::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	//VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);

	//if(oapiCockpitMode() == COCKPIT_VIRTUAL) d = d * 2;
	//else VESSEL3::clbkDrawHUD (mode, hps, skp);
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}

	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);

if (oapiCockpitMode() != COCKPIT_VIRTUAL  && hud_dsp == 0)
{
	if(abort_status == 0)
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"A = ABORT");
	skp->Text ((10+roxl), (wd+royl), abuf,strlen(abuf));
	}


	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double test = ((m1+m2)/m1);
	double dv = (ISP*(log(test)));

	sprintf(abuf,"Remaining dV = %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}


//if(rbi == 0)
//{
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Payload Mass %0.1fkg", pld_mass);
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"W = Set Ballast %0.1fkg", ballast);
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"P = Add Payload");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}
	

	if (hatch_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"K = Open Nose Hatch");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else if (hatch_status == 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"K = Open Nose Hatch");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"K = Close Nose Hatch");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}


	if (door_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Open Side Hatch");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}
	else if (door_status == 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Open Side Hatch");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Close Side Hatch");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Jettison Payload");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}


if (tk_status == 0)
{
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Ctrl+U = Jettison Trunk");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}
}

if (nc_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"B = Open Nose Cone");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}
	else if (nc_status == 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"B = Open Nose Cone");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"B = Close Nose Cone");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}

	/*if (ring2_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"V = Extend Docking Ring");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}
	else if (ring2_status == 1)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"V = Retract Docking Ring");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}
	else if (ring2_status == 2)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"V = Retract Docking Ring");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}
	else if (ring2_status == 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"V = Extend Docking Ring");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}*/

	

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"N/Ctrl+N = Toggle Cockpit View");
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}

	/*{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"C = Toggle HUD Info: Main/Reentry");
	skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));
	}*/

	if (rcs_cfg == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"D = Roll/Yaw RCS: Normal");
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"D = Roll/Yaw RCS: Reentry");
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"L = Arm/Safe Chutes:");
	skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));

	if (drogue_deploy_status != 3)sprintf(abuf,"                    Deployed");

	else if (chute_arm_status == 0)sprintf(abuf,"                    Safe");

	else if (chute_arm_status == 1)
		{
			skp->SetTextColor(0x006666FF);
			sprintf(abuf,"                    Armed");
		}
	skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Ctrl+8 = Cabin Lighting");
	skp->Text ((10+roxl), (w12+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"1/2/3 = Nav.Beacons/Fwd.Light_1/Fwd.Light_2");
	skp->Text ((10+roxl), (w13+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Ctrl+1-7 = Panel Btn.Covers/Switch");
	skp->Text ((10+roxl), (w14+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Ctrl+Space = Show/Hide HUD Data");
	skp->Text ((10+roxl), (w15+royl), abuf,strlen(abuf));
	}
//}
//else
//{
	/*if(auto_burn_sw == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"U = Auto Deorbit Burn: OFF");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}
	else if(auto_burn_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"U = Auto Deorbit Burn: ON");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x00666666);
	char abuf[256];
	sprintf(abuf,"Auto Deorbit Finished");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	if(auto_reentry_sw == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"E = Auto Reentry Attitude: OFF");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"E = Auto Reentry Attitude: ON");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}*/

	/*if (rcs_cfg == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"D = Roll/Yaw RCS: Normal");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"D = Roll/Yaw RCS: Reentry");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}*/

	/*{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"S = Set Reentry Target: %s", tgt_name);
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}*/

	/*if (auto_burn_sw == 0)
	{
		{
		skp->SetTextColor(0x0066FF66);
		char abuf[256];
		sprintf(abuf,"Time to Burn %0.1fs", time_to_burn);
		skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
		}
	}

	else if (auto_burn_sw == 1)
	{
	if(burn_sw == 0)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Time to Burn %0.1fs", time_to_burn);
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"DEORBIT BURN ACTIVE");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}
	}
	else
	{
	skp->SetTextColor(0x00666666);
	char abuf[256];
	sprintf(abuf,"Deorbit Burn Finished");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}



	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Range %0.1fkm", tgt_dist*0.001);
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Cross Range Error %0.1fkm", crossrange_error*0.001);
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Vel.Hdg. %0.2f", curr_hdg*DEG);
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Req.Hdg. %0.2f", req_hdg*DEG);
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Req.Inc. %0.2f", req_inc*DEG );
	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
	}*/

	/*{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"L = Arm/Safe Chutes:");
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));

	if (drogue_deploy_status != 3)sprintf(abuf,"                    Deployed");

	else if (chute_arm_status == 0)sprintf(abuf,"                    Safe");

	else if (chute_arm_status == 1)
		{
			skp->SetTextColor(0x006666FF);
			sprintf(abuf,"                    Armed");
		}
	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"C = Toggle HUD Info: Main/Reentry");
	skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));
	}*/
//}
	
}
//else
//{
//oapi::Font *pOrgHUDFont = skp->SetFont(m_pHudNormalFont);
//
//if(rbi == 0)VESSEL3::clbkDrawHUD (mode, hps, skp);
//
//else
//{
//	royl = -220;
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"W = Set Ballast %0.1fkg", ballast);
//	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"P = Attach Payload");
//	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"N = Set View:Pilot/Crew/Docking");
//	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
//	}
//
//
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"Payload Mass %0.1fkg", pld_mass);
//	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	double m1 = GetEmptyMass();
//	double m2 = GetPropellantMass(hProp);
//	double test = ((m1+m2)/m1);
//	double dv = (ISP*(log(test)));
//
//	sprintf(abuf,"Total Mass %0.1fkg",(m1+m2));
//	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
//	sprintf(abuf,"Propellant Mass %0.1fkg", m2);
//	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
//	sprintf(abuf,"Remaining dV %0.2f m/s", dv);
//	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
//	}
//
//
//	oapi::Pen *pOrgHUDPen = skp->SetPen(m_pHudNormalPen);
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"Draco RCS Level");
//	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
//
//	sprintf(abuf,"SuperDraco Level");
//	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
//
//	double rcs_lev = GetThrusterGroupLevel(THGROUP_ATT_PITCHUP);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_YAWLEFT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_BANKLEFT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_UP);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_DOWN);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_LEFT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_RIGHT);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_FORWARD);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_ATT_BACK);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_MAIN);
//	rcs_lev = rcs_lev + GetThrusterGroupLevel(THGROUP_RETRO);
//
//	int rcs_lev_int = 10*((int)(rcs_lev*4));
//
//	if(rcs_lev > 0)skp->Line((10+roxl+270), (w7+royl+15),(10+roxl+270+rcs_lev_int), (w7+royl+15));
//
//	rcs_lev = 0;
//
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[0]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[1]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[2]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[3]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[4]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[5]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[6]);
//	rcs_lev = rcs_lev + GetThrusterLevel(th_sd[7]);
//
//	rcs_lev_int = 10*((int)(rcs_lev*3));
//	if(rcs_lev > 0)skp->Line((10+roxl+270), (w8+royl+15),(10+roxl+270+rcs_lev_int), (w8+royl+15));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"Drogue Chute");
//	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
//
//	if (drogue_deploy_status != 3)sprintf(abuf,"             Deployed");
//
//	else if (chute_arm_status == 0)sprintf(abuf,"              Safe");
//
//	else if (chute_arm_status == 1)
//		{
//		skp->SetTextColor(0x006666FF);
//		sprintf(abuf,"             Armed");
//		}
//	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"Main Chute");
//	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
//
//	if (chute_deploy_status != 2)sprintf(abuf,"           Deployed");
//
//	else if (chute_arm_status == 0)sprintf(abuf,"            Safe");
//
//	else if (chute_arm_status == 1)
//		{
//		skp->SetTextColor(0x006666FF);
//		sprintf(abuf,"           Armed");
//		}
//	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FF66);
//	char abuf[256];
//	sprintf(abuf,"G-Load");
//	skp->Text ((10+roxl), (w11+royl), abuf,strlen(abuf));
//
//	OBJHANDLE hVes = GetAttachmentStatus(hF9attach);
//	//if(hVes == NULL) GetDockStatus(hDock);
//	VESSEL2* v;
//	if(hVes == NULL) v = (VESSEL2*)oapiGetVesselInterface(GetHandle());
//	else v = (VESSEL2*)oapiGetVesselInterface(hVes);
//		{
//		VECTOR3 total_force;
//		v->GetForceVector(total_force);
//		VECTOR3 weight;
//		v->GetWeightVector(weight);
//		double mass = v->GetMass();
//		VECTOR3 net_force = total_force - weight;
//		double net_acc = length(net_force)/mass;
//		double g_load = net_acc/9.81;
//		DWORD fs = v->GetFlightStatus();
//		//if (fs == 01 && hVes == NULL) g_load = 1;
//		if(fs == 01)g_load = 1;
//		else if(GetDockStatus(hDock) != NULL)g_load = 0;
//
//		int g_load_int = (int)(g_load*0.1*240);
//		skp->Line((10+roxl+170), (w11+royl+15),(10+roxl+170+g_load_int), (w11+royl+15));
//
//		skp->SetTextColor(0x0066FFFF);
//		sprintf(abuf,"%0.1fG", g_load);
//		skp->Text ((10+roxl+170+g_load_int+10), (w11+royl), abuf,strlen(abuf));
//		}
//	}
//	skp->SetPen(pOrgHUDPen);
//
//	////////////////////// lower half
//	royl = 20;
//
//
//	if (auto_burn_sw == 0)
//	{
//		skp->SetTextColor(0x0066FF66);
//		char abuf[256];
//		sprintf(abuf,"Time to Burn %0.1fs", time_to_burn);
//		skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
//	}
//
//	else if (auto_burn_sw == 1)
//	{
//	if(burn_sw == 0)
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Time to Burn %0.1fs", time_to_burn);
//	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
//	}
//	else
//	{
//	skp->SetTextColor(0x006666FF);
//	char abuf[256];
//	sprintf(abuf,"DEORBIT BURN ACTIVE");
//	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
//	}
//	}
//	else
//	{
//	skp->SetTextColor(0x00666666);
//	char abuf[256];
//	sprintf(abuf,"Deorbit Burn Finished");
//	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
//	}
//
//
//
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Range %0.1fkm", tgt_dist*0.001);
//	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Cross Range Error %0.1fkm", crossrange_error*0.001);
//	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Vel.Hdg. %0.2f", curr_hdg*DEG);
//	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Req.Hdg. %0.2f", req_hdg*DEG);
//	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
//	}
//
//	{
//	skp->SetTextColor(0x0066FFFF);
//	char abuf[256];
//	sprintf(abuf,"Req.Inc. %0.2f", req_inc*DEG );
//	skp->Text ((10+roxl), (w9+royl), abuf,strlen(abuf));
//	}
//
//	if(abort_status == 0)
//	{
//	skp->SetTextColor(0x006666FF);
//	char abuf[256];
//	sprintf(abuf,"A = ABORT");
//	skp->Text ((10+roxl), (w10+royl), abuf,strlen(abuf));
//	}
//	
//}
//
//skp->SetFont(pOrgHUDFont);
//
//}
	
	return true;
}

bool dm2_capsule::clbkLoadVC (int id)
{
	//static VCHUDSPEC huds = {  // common HUD specs
	//	mesh_vc,            // nmesh
	//	GRP_VirtualHUD_VC,  // ngroup
	//	{0,0,0},            // hudcnt (to be filled)
	//	0.176558            // size
	//};

	DWORD nmesh = 0+crew_count;
	DWORD ngroupl = 3;
	DWORD ngroupr = 1;
	DWORD ngrouph = 114;

	//static VCMFDSPEC mfd_left = {nmesh,ngroupl};
	//static VCMFDSPEC mfd_right = {nmesh,ngroupr};

	static VCHUDSPEC huds = {  // common HUD specs
		6+crew_count,            // nmesh
		0,  // ngroup
		{-0.275, 0.88, 2.296},            // hudcnt (to be filled)
		0.8            // size
		};

	//static VCMFDSPEC mfds = {
	//	mesh_vc, 0
	//};

	static EXTMFDSPEC mfds = { // common MFD specs
		{0,0,0,0},          // pos
		nmesh,            // nmesh
		0,                  // ngroup (to be filled)
		MFD_SHOWMODELABELS, // flag
		6, 6,               // nbt1, nbt2
		(int)(mfd_res/6), (int)(mfd_res/7)        // bt_yofs, bt_ydist
	};

	static const int mfdgrp[7] = {
		GRP_CDR1_VC,GRP_CDR2_VC, GRP_CDR3_VC,
		GRP_CDR4_VC,GRP_CDR5_VC, GRP_CDR6_VC, GRP_CDR7_VC
	};

	bool ok = false;

	// register MFD function buttons
	// this needs to be done globally, so that the labels are correctly updated from all VC positions
	SURFHANDLE tex1 = oapiGetTextureHandle (hMcap, 7);
	SURFHANDLE texl = oapiGetTextureHandle (hMcap, 9);
	SURFHANDLE texr = oapiGetTextureHandle (hMcap,10);

	// commander MFD function buttons
	oapiVCRegisterArea (AID_CDR3_BUTTONS, _R(22,0,164,472), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	oapiVCRegisterArea (AID_CDR1_BUTTONS, _R(22,0,164,472), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, texl);
	oapiVCRegisterArea (AID_CDR2_BUTTONS, _R(22,0,164,472), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, texr);
	//oapiVCRegisterArea (AID_CDR3_BUTTONS, _R(0,0,255,511), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	
	//oapiVCRegisterArea (AID_CDR2_BUTTONS, _R(360,324,430,340), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (AID_CDR1_BUTTONS, _R(0,1,255,28), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (AID_CDR2_BUTTONS, _R(0,29,255,56), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_LBUP|PANEL_MOUSE_LBPRESSED|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (HATCH_BUTTONS, _R(0,57,255,70), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (RCS_BUTTONS, _R(0,71,127,99), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (AUTO_BUTTONS, _R(128,71,255,99), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (INFO_TAB, _R(0,100,127,127), PANEL_REDRAW_ALWAYS, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (TGT_BUTTONS, _R(128,100,255,127), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (NAV_BUTTONS, _R(128,156,255,169), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (PANEL_BUTTONS, _R(128,170,171,183), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (SEP_BUTTONS, _R(0,156,127,183), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (LIGHT_BUTTONS, _R(128,184,255,197), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	//oapiVCRegisterArea (CHUTE_BUTTONS, _R(0,184,127,211), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY, PANEL_MAP_BACKGROUND, tex1);
	switch (id) {
	case 0: // commander position
		//SetCameraOffset (_V(-0.67,2.55,14.4));
		//SetCameraDefaultDirection (_V(0,0,1));
		//SetCameraMovement (_V(0,0,0.3), 0, 0, _V(-0.3,0,0), 75*RAD, -5*RAD, _V(0.3,0,0), -20*RAD, -27*RAD);
		//huds.hudcnt = _V(-0.10,1.05,1.9);

		RegisterVC_CdrMFD (); // activate commander MFD controls

		ok = true;
		break;

	}

	if (ok) {
		// register the HUDs (synced)
		oapiVCRegisterHUD (&huds);
		// register all MFD displays
		for (int i = 0; i < 7; i++) {
			mfds.ngroup = mfdgrp[i];
			oapiRegisterMFD (MFD_LEFT+i, &mfds);
		}
		
	}

	SetCamPos();

	return ok;
}

bool dm2_capsule::clbkLoadGenericCockpit ()
{
	SetCamPos();
	return true;
}

void dm2_capsule::SetCamPos()
{
	//oapiDelAnnotation (hNote);
	//char cbuf[256];
	//VECTOR3 col = _V(1,1,1);

	campos = _V(0,0,0);
	double campolar = 0*RAD;
	double camazimuth = 0*RAD;
	VECTOR3 camdir = _V(0,0,1);
	SetCameraRotationRange(0.99*PI,0.99*PI,0.49*PI,0.49*PI);

	if(oapiCockpitMode() == COCKPIT_VIRTUAL)
	{
		if(cam_status == 0)
		{
		campos = _V(-0.273, 0.879, 1.815);
		campolar = -45*RAD;
		camazimuth = 0*RAD;
		camdir = _V(0, -0.406, 0.914);
		}
		else if(cam_status == 1)
		{
		campos = _V(0.243, 0.879, 1.815);
		campolar = -45*RAD;
		camazimuth = 0*RAD;
		camdir = _V(0, -0.406, 0.914);
		}
		else if(cam_status == 2)
		{
		campos = _V(-0.903, 0.516, 1.719);
		campolar = 24*RAD;
		camazimuth = 0*RAD;
		camdir = _V(0, -0.406, 0.914);
		}
		else if(cam_status == 3)
		{
		campos = _V(0,0,3.5);
		campolar = 0*RAD;
		camazimuth = 0*RAD;
		camdir = _V(0, 0, 1);
		}
		else if(cam_status == 4)
		{
		campos = _V(0, -1.041, 2.355);
		campolar = 65*RAD;
		camazimuth = 180*RAD;
		camdir = _V(0, 0.9, -0.4359);
		}

	}
	else
	{
	if (gc_cam_status == 1)camdir = _V(0,0,-1);
	}

	SetCameraOffset(campos);
	//oapiCameraSetCockpitDir(campolar, camazimuth);
	//oapiCameraAttach(GetHandle(), 2);

	SetCameraDefaultDirection (camdir);
	oapiCameraAttach(GetHandle(), 2);
	oapiCameraSetCockpitDir(0, 0);

	//if(oapiCameraInternal() == false)oapiDelAnnotation (hNote);

}

void dm2_capsule::CalcMass()
{
	double current_mass = DRYMASS;
	if (pd_status == 0)current_mass = current_mass+PARADOORMASS;
	if (tk_status == 0)current_mass = current_mass+TRUNKMASS;

	int npatts = AttachmentCount(false);
	if (npatts > 0)
	{
	pld_mass = 0;
	for (DWORD i = 0; i < npatts; i++)
		{
		ATTACHMENTHANDLE hpld_att =  GetAttachmentHandle (false, i);
		OBJHANDLE hpld = GetAttachmentStatus(hpld_att);
		if (hpld == NULL)continue;
		else pld_mass = oapiGetMass(hpld)+pld_mass;
		}
	}
	current_mass = current_mass+pld_mass;
	current_mass = current_mass+ballast;
	SetEmptyMass(current_mass);
}



void dm2_capsule::JettisonTK()
{
char name[256];
strcpy(name,GetName());
strcat(name, "_Trunk");
VESSELSTATUS vs;
 GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (TRUNK_OFF, vs.rpos);
  GlobalRot (_V(0,0,-0.3), rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.vrot.x = vs.vrot.x;
  vs.vrot.y = vs.vrot.y;
  vs.vrot.z = vs.vrot.z;
  OBJHANDLE hTrunk;
  if(skin_sw < 0) hTrunk = oapiCreateVessel (name, "DM2/dm2_trunk_sd", vs);
  else hTrunk = oapiCreateVessel (name, "DM2/dm2_trunk", vs);
//  hTrunk = oapiCreateVessel (name, "DM2/dm2_trunk", vs);
  //DelMesh(4);
  //DelMesh(3);
  DelMesh(5+crew_count);
  tk_status = 1;
  DeleteTrunkPayload();
  CalcMass();


  /*VESSEL3* v =(VESSEL3*)oapiGetVesselInterface(hTrunk);
  char abuf[256];
  sprintf (abuf, "%0.4f", solp_proc);
  v->clbkConsumeBufferedKey(COMM_SEND_SLPD, false, abuf);

  char bbuf[256];
  sprintf (bbuf, "%0.4f", solp_rot_proc);
  v->clbkConsumeBufferedKey(COMM_SEND_SLPR, false, bbuf);

  char cbuf[256];
  sprintf (cbuf, "%d", fr_status);
  v->clbkConsumeBufferedKey(COMM_SEND_FAIR, false, cbuf);*/

  //fr_status = 1;
  capvert = CreateAirfoil2 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeffNeg, 1.4, 3, 1); // 1 1 1
  caphrz = CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeffNeg, 1.4, 3, 1); // 1 1 1

  //// add reentry streams
	//SURFHANDLE pstr5 = oapiRegisterExhaustTexture ("contrail3");
	//PARTICLESTREAMSPEC rps2 = {
	//	0, 3, 30, 0, 0.004, 0.2, 2, 1, PARTICLESTREAMSPEC::EMISSIVE,
	//	PARTICLESTREAMSPEC::LVL_LIN, 0.25, 0.75, PARTICLESTREAMSPEC::ATM_PLIN, 4.5e7, 6e7,
	//	pstr5
	//};
	//AddReentryStream (&rps2);

	//PARTICLESTREAMSPEC rps = {
	//	0, 3, 20, 0, 0.008, 0.4, 20, 2, PARTICLESTREAMSPEC::DIFFUSE,
	//	PARTICLESTREAMSPEC::LVL_FLAT, 0.25, 0.75, PARTICLESTREAMSPEC::ATM_PLIN, 3e7, 6e7
	//};
	//AddReentryStream (&rps);

 // oapiVCTriggerRedrawArea (-1, SEP_BUTTONS);

}


void dm2_capsule::JettisonPayload()
{
	int npatts = AttachmentCount(false);
	if (npatts > 0)
	{
	for (DWORD i = 0; i < npatts; i++)
		{
		ATTACHMENTHANDLE hpld_att =  GetAttachmentHandle (false, i);
		OBJHANDLE hpld = GetAttachmentStatus(hpld_att);
		if (hpld == NULL)continue;
		else
			{
				DetachChild(hpld_att);
				pld_att_count = -1;
				return;
			}
		}
	}
	
}

void dm2_capsule::DeleteTrunkPayload()
{
	int npatts = AttachmentCount(false);
	if (npatts > 0)
	{
	for (DWORD i = 0; i < npatts; i++)
		{
		ATTACHMENTHANDLE hpld_att =  GetAttachmentHandle (false, i);
		OBJHANDLE hpld = GetAttachmentStatus(hpld_att);
		if (hpld == NULL)continue;
		else {
			VECTOR3 pld_glob_pos = _V(0,0,0);
			oapiGetGlobalPos(hpld, &pld_glob_pos);
			VECTOR3 pld_loc_pos = _V(0,0,0);
			Global2Local(pld_glob_pos, pld_loc_pos);
			if (pld_loc_pos.z < -0.6) oapiDeleteVessel(hpld);
			}
		}
	}
//	oapiVCTriggerRedrawArea (-1, SEP_BUTTONS);
	pld_att_count = -1;
}


void dm2_capsule::JettisonDrogue()
{
	char name[256];
	strcpy(name,GetName());
	strcat(name, "_Drogue");

	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	GlobalRot (_V(0,0,1), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	if (zforward.x > 1)zforward.x = 1;
	else if (zforward.x < -1)zforward.x = -1;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);

VESSELSTATUS vs;
 GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (DROGUE_OFF, vs.rpos);
  GlobalRot (_V(-0,-0,50), rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  vs.vrot.x = vs.vrot.x;
  vs.vrot.y = -vs.vrot.y;
  vs.vrot.z = -vs.vrot.z;
  oapiCreateVessel (name, "DM2/dm2_drogue", vs); // ("child name", "cfg path", vesselstatus vs)

  //sprintf(oapiDebugString(), "vs.vrot.x %0.3f y %0.3f z %0.3f", vs.vrot.x,vs.vrot.y,vs.vrot.z);

  DelMesh(2+crew_count);
  //DelMesh(1);
  drogue_deploy_status = 0;
  chute_deploy_status = 1;
  chutevert = CreateAirfoil2 (LIFT_VERTICAL, _V(0,1,2), VLiftCoeff, 0, 0.1, 0);
  chutehrz = CreateAirfoil2 (LIFT_HORIZONTAL, _V(0,1,2), HLiftCoeff, 0, 0.1, 0);
}

void dm2_capsule::JettisonChutes()
{
	char name[256];
	strcpy(name,GetName());
	strcat(name, "_MainChute");


	VECTOR3 pos;
	VECTOR3 vel;
	GlobalRot (_V(1,0,0), pos);
	//GlobalRot (_V(0,0.3420,0.9397), vel);
	GlobalRot (_V(0,-0.5,0.866), vel);
	VECTOR3 yup;
	VECTOR3 yupn;
	VECTOR3 zforward;
	VECTOR3 xright;
	yupn = crossp(vel, pos);
	double yupnlength;
	yupnlength = length(yupn);
	yup = _V(yupn.x/yupnlength, yupn.y/yupnlength, yupn.z/yupnlength);
	xright = crossp (yup, vel);
	zforward = vel;
	
	if (zforward.x > 1)zforward.x = 1;
	else if (zforward.x < -1)zforward.x = -1;

	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);

	VECTOR3 aspdv;
	GetAirspeedVector(FRAME_LOCAL, aspdv);

VESSELSTATUS vs;
 GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (CHUTES_OFF, vs.rpos);
  GlobalRot (-aspdv, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.eng_main = vs.eng_hovr = 0.0;
  vs.status = 0;
  vs.arot = _V(alpha, beta, (PI/2)-gamma);
  vs.vrot.x = vs.vrot.x;
  vs.vrot.y = -vs.vrot.y;
  vs.vrot.z = -vs.vrot.z;
  oapiCreateVessel (name, "DM2/dm2_chutes", vs); // ("child name", "cfg path", vesselstatus vs)

}

void dm2_capsule::KillAllRCS ()
{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT,0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT,0);
		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_BACK, 0);
}

void dm2_capsule::ApplySkin()
{	
	if(skin_sw == 0)
	{
	//bs3 = oapiGetTextureHandle (hMcap, 1);
	//oapiBlt(bs3, bs1, 0, 0, 0, 0, 2048, 2048);

	oapiSetTexture(dhMesh2,1,bs1);
	oapiSetTexture(dhMesh2,8,bs1);
	//oapiReleaseTexture (bs1);
	skin_sw = 1;
	}

	else if(skin_sw == 1)
	{
	//bs4 = oapiGetTextureHandle (hMcap, 2);
	//oapiBlt(bs4, bs2, 0, 0, 0, 0, 1024, 1024);
	oapiSetTexture(dhMesh0,1,bs1);
	oapiSetTexture(dhMesh1,1,bs1);
	oapiSetTexture(dhMesh2,4,bs2);
	//oapiReleaseTexture (bs2);
	skin_sw = 2;
	}

	if(skin_sw == -3)
	{
	oapiSetTexture(dhMesh2,8,sd1);
	if(tk_status == 0)oapiSetTexture(dhMesh3,1,sd3);
	//oapiReleaseTexture (bs1);
	skin_sw = -2;
	}
	else if(skin_sw == -2)
	{
	//bs4 = oapiGetTextureHandle (hMcap, 2);
	//oapiBlt(bs4, bs2, 0, 0, 0, 0, 1024, 1024);
	//oapiSetTexture(dhMesh0,1,sd1);
	//oapiSetTexture(dhMesh1,1,sd1);
	oapiSetTexture(dhMesh2,4,sd2);
	//oapiReleaseTexture (bs2);
	skin_sw = -1;
	}

	//dhMesh2
	//bs3 = oapiGetTextureHandle (hMdoor, 1);
	//oapiBlt(bs3, bs2, 0, 0, 0, 0, 1024, 1024);
}

void dm2_capsule::SetMeshMat()
{
double drf = post_step*0.02;
if (red_fact < rpse_lvl)red_fact = red_fact+drf;
else if (red_fact > rpse_lvl)red_fact = red_fact-drf;

if (red_fact > 1)red_fact = 1;
else if (red_fact < 0)red_fact = 0;
double yel_fact = red_fact*0.25;

MATERIAL *mat2 = oapiMeshMaterial (hMcap, 23);

	
	mat2->emissive.r = red_fact*red_fact_adjust;
	mat2->emissive.g = yel_fact*red_fact_adjust;
	mat2->emissive.b = 0;
	//mat2->emissive.a = 1;
	double dif_red_fct = 0.6-red_fact;
	if(red_fact_adjust == 3)dif_red_fct = 1.0-red_fact;
	if(dif_red_fct < 0)dif_red_fct = 0;

	//if(red_fact_adjust == 1)mat2->power = 0;

	mat2->diffuse.r = dif_red_fct;
	mat2->diffuse.g = dif_red_fct;
	mat2->diffuse.b = dif_red_fct;
	//mat2->emissive.a = 1;
	if(dhMesh2)oapiSetMaterial(dhMesh2,23, mat2);
	
}


void dm2_capsule::SetNightMat()
{
	MATERIAL *mat1n = oapiMeshMaterial (hMcap, 0);
	MATERIAL *mat15n = oapiMeshMaterial (hMcap, 14);
	MATERIAL *mat18n = oapiMeshMaterial (hMcap, 17);
	MATERIAL *mat22n = oapiMeshMaterial (hMcap, 21);

	mat1n->diffuse.r = 0;
	mat1n->diffuse.g = 0;
	mat1n->diffuse.b = 0;
	mat1n->diffuse.a = 1;
	mat1n->ambient.r = 0;
	mat1n->ambient.g = 0;
	mat1n->ambient.b = 0;
	mat1n->ambient.a = 1;
	mat1n->specular.r = 0;
	mat1n->specular.g = 0;
	mat1n->specular.b = 0;
	mat1n->specular.a = 1;
	mat1n->power = 128;
	mat1n->emissive.r = 0;
	mat1n->emissive.g = 0;
	mat1n->emissive.b = 0;
	mat1n->emissive.a = 1;

	mat15n->diffuse.r = 0;
	mat15n->diffuse.g = 0;
	mat15n->diffuse.b = 0;
	mat15n->diffuse.a = 1;
	mat15n->ambient.r = 0;
	mat15n->ambient.g = 0;
	mat15n->ambient.b = 0;
	mat15n->ambient.a = 1;
	mat15n->specular.r = 0;
	mat15n->specular.g = 0;
	mat15n->specular.b = 0;
	mat15n->specular.a = 1;
	mat15n->power = 5;
	mat15n->emissive.r = 0;
	mat15n->emissive.g = 0;
	mat15n->emissive.b = 0;
	mat15n->emissive.a = 1;

	mat18n->diffuse.r = 0;
	mat18n->diffuse.g = 0;
	mat18n->diffuse.b = 0;
	mat18n->diffuse.a = 1;
	mat18n->ambient.r = 0;
	mat18n->ambient.g = 0;
	mat18n->ambient.b = 0;
	mat18n->ambient.a = 1;
	mat18n->specular.r = 0;
	mat18n->specular.g = 0;
	mat18n->specular.b = 0;
	mat18n->specular.a = 1;
	mat18n->power = 22.9053;
	//mat18n->emissive.r = 0.163;
	//mat18n->emissive.g = 0.178;
	//mat18n->emissive.b = 0.225;
	mat18n->emissive.r = 0.106;
	mat18n->emissive.g = 0.116;
	mat18n->emissive.b = 0.146;
	mat18n->emissive.a = 1;

	mat22n->diffuse.r = 0;
	mat22n->diffuse.g = 0;
	mat22n->diffuse.b = 0;
	mat22n->diffuse.a = 0.5;
	mat22n->ambient.r = 0;
	mat22n->ambient.g = 0;
	mat22n->ambient.b = 0;
	mat22n->ambient.a = 0.773684;
	mat22n->specular.r = 0;
	mat22n->specular.g = 0;
	mat22n->specular.b = 0;
	mat22n->specular.a = 0.773684;
	mat22n->power = 128;
	mat22n->emissive.r = 0;
	mat22n->emissive.g = 0;
	mat22n->emissive.b = 0;
	mat22n->emissive.a = 1;

	if(dhMesh2)
	{
	oapiSetMaterial(dhMesh2,0, mat1n);
	oapiSetMaterial(dhMesh2,14, mat15n);
	oapiSetMaterial(dhMesh2,17, mat18n);
	oapiSetMaterial(dhMesh2,21, mat22n);

	/////textures
	oapiSetTexture(dhMesh2,6,hS_panel_n);
	oapiSetTexture(dhMesh2,5,hS_VC_n);
	}

	////// crew
	MATERIAL *mat_crew1n = oapiMeshMaterial (hMcrew, 0);
	MATERIAL *mat_crew2n = oapiMeshMaterial (hMcrew, 1);
	//mat_crew1n->emissive.r = 0.163;
	//mat_crew1n->emissive.g = 0.178;
	//mat_crew1n->emissive.b = 0.225;
	mat_crew1n->emissive.r = 0.118;
	mat_crew1n->emissive.g = 0.128;
	mat_crew1n->emissive.b = 0.162;
	mat_crew2n->emissive.r = 0;
	mat_crew2n->emissive.g = 0;
	mat_crew2n->emissive.b = 0;

	for (int i = 0; i < crew_count; i++)
	{
		if(dHMeshCrew[i]) 
		{
			oapiSetMaterial(dHMeshCrew[i],0, mat_crew1n);
			oapiSetMaterial(dHMeshCrew[i],1, mat_crew2n);
		}
	}
	
}

void dm2_capsule::SetDayMat()
{
	MATERIAL *mat1d = oapiMeshMaterial (hMcap, 0);
	MATERIAL *mat15d = oapiMeshMaterial (hMcap, 14);
	MATERIAL *mat18d = oapiMeshMaterial (hMcap, 17);
	MATERIAL *mat22d = oapiMeshMaterial (hMcap, 21);

	mat1d->diffuse.r = 0;
	mat1d->diffuse.g = 0;
	mat1d->diffuse.b = 0;
	mat1d->diffuse.a = 1;
	mat1d->ambient.r = 0.2;
	mat1d->ambient.g = 0.2;
	mat1d->ambient.b = 0.2;
	mat1d->ambient.a = 1;
	mat1d->specular.r = 0.2;
	mat1d->specular.g = 0.2;
	mat1d->specular.b = 0.2;
	mat1d->specular.a = 1;
	mat1d->power = 128;
	mat1d->emissive.r = 0;
	mat1d->emissive.g = 0;
	mat1d->emissive.b = 0;
	mat1d->emissive.a = 1;

	mat15d->diffuse.r = 0.32;
	mat15d->diffuse.g = 0.35;
	mat15d->diffuse.b = 0.4;
	mat15d->diffuse.a = 1;
	mat15d->ambient.r = 0;
	mat15d->ambient.g = 0;
	mat15d->ambient.b = 0;
	mat15d->ambient.a = 1;
	mat15d->specular.r = 0.6;
	mat15d->specular.g = 0.6;
	mat15d->specular.b = 0.6;
	mat15d->specular.a = 1;
	mat15d->power = 5;
	mat15d->emissive.r = 0.32;
	mat15d->emissive.g = 0.35;
	mat15d->emissive.b = 0.4;
	mat15d->emissive.a = 1;

	mat18d->diffuse.r = 0;
	mat18d->diffuse.g = 0;
	mat18d->diffuse.b = 0;
	mat18d->diffuse.a = 1;
	mat18d->ambient.r = 1;
	mat18d->ambient.g = 1;
	mat18d->ambient.b = 1;
	mat18d->ambient.a = 1;
	mat18d->specular.r = 0;
	mat18d->specular.g = 0;
	mat18d->specular.b = 0;
	mat18d->specular.a = 1;
	mat18d->power = 5;
	mat18d->emissive.r = 0.435;
	mat18d->emissive.g = 0.435;
	mat18d->emissive.b = 0.435;
	mat18d->emissive.a = 1;

	mat22d->diffuse.r = 0;
	mat22d->diffuse.g = 0;
	mat22d->diffuse.b = 0;
	mat22d->diffuse.a = 0.42;
	mat22d->ambient.r = 0;
	mat22d->ambient.g = 0;
	mat22d->ambient.b = 0;
	mat22d->ambient.a = 0.773684;
	mat22d->specular.r = 0;
	mat22d->specular.g = 0;
	mat22d->specular.b = 0;
	mat22d->specular.a = 0.773684;
	mat22d->power = 128;
	mat22d->emissive.r = 0.25;
	mat22d->emissive.g = 0.25;
	mat22d->emissive.b = 0.25;
	mat22d->emissive.a = 1;

	if(dhMesh2)
	{
	oapiSetMaterial(dhMesh2,0, mat1d);
	oapiSetMaterial(dhMesh2,14, mat15d);
	oapiSetMaterial(dhMesh2,17, mat18d);
	oapiSetMaterial(dhMesh2,21, mat22d);

	////// textures
	oapiSetTexture(dhMesh2,6,hS_panel_d);
	oapiSetTexture(dhMesh2,5,hS_VC_d);
	}

	////// crew
	MATERIAL *mat_crew1d = oapiMeshMaterial (hMcrew, 0);
	MATERIAL *mat_crew2d = oapiMeshMaterial (hMcrew, 1);
	mat_crew1d->emissive.r = 0.8;
	mat_crew1d->emissive.g = 0.8;
	mat_crew1d->emissive.b = 0.8;
	mat_crew2d->emissive.r = 0.1;
	mat_crew2d->emissive.g = 0.1;
	mat_crew2d->emissive.b = 0.1;

	for (int i = 0; i < crew_count; i++)
	{
	
		if(dHMeshCrew[i])
		{
			oapiSetMaterial(dHMeshCrew[i],0, mat_crew1d);
			oapiSetMaterial(dHMeshCrew[i],1, mat_crew2d);
		}
	}

}

//void dm2_capsule::SetInstMat()
//{
//	if(oapiCockpitMode() == COCKPIT_VIRTUAL)
//	{
//	MATERIAL *mat1 = oapiMeshMaterial (hMinst, 2);
//	mat1->emissive.r = 1;
//	mat1->emissive.g = 1;
//	mat1->emissive.b = 1;
//	mat1->emissive.a = 1;
//	if(dhMesh1)oapiSetMaterial(dhMesh1,2, mat1);
//	}
//else
//	{
//	MATERIAL *mat1 = oapiMeshMaterial (hMinst, 2);
//	mat1->emissive.r = 0;
//	mat1->emissive.g = 0;
//	mat1->emissive.b = 0;
//	mat1->emissive.a = 1;
//	if(dhMesh1)oapiSetMaterial(dhMesh1,2, mat1);
//	}
//}

void dm2_capsule::SetChutes()
{	
	VECTOR3 aspd = _V(0,0,-1);
	double angle_lim = 30*RAD;//20.00 //26.565*RAD;
	double angle_dt = angle_lim*step_time*0.5;
	curr_para_angle = curr_para_angle+angle_dt;
	if (curr_para_angle > angle_lim)curr_para_angle = angle_lim;
	aspd.x = 0;
	aspd.y = sin(curr_para_angle);
	aspd.z = -cos(curr_para_angle);
	normalise(aspd);

	VECTOR3 chtvctr = _V(0,0,1);
	chtvctr.x = (aspd.x*-5);
	chtvctr.y = (aspd.y*-5)-1.5;
	chtvctr.z = (aspd.z*-5)+2;

	if (landed_sw == 0)
	{
	EditAirfoil(chutevert,11101,chtvctr,VLiftCoeff,(afchord*afedit),(afarea*afedit),afedit);
	EditAirfoil(chutehrz,11101,chtvctr,VLiftCoeff,(afchord*afedit),(afarea*afedit),afedit);
	SetCrossSections (_V(((192*afedit)+8),((192*afedit)+8),((342*afedit)+8)));
	}

	if (GroundContact() == true)
	{
		////// dust off
		if(dust_sw < 1)
		{
		dust_sw = dust_sw+step_time;
		dust_lvl = 1;
			if (dust_sw > 0.4)
			{
				dust_sw = 1;
				dust_lvl = 0;
				DelExhaustStream(hDust);
			}
		}
		else DelExhaustStream(hDust);
		/////////////////////

	if (landed_sw == 0)
		{
		landed_sw = 1;

		DelAirfoil(chutevert);
		DelAirfoil(chutehrz);
		SetCrossSections (_V(8,8,8));
		SetMeshVisibilityMode(1+crew_count, MESHVIS_NEVER);
		JettisonChutes();
		SetSize(4);
		}
	
		/*double da = step_time*0.0625;
		double db = step_time*1.0;

		if (chute_roty_proc > 0.5) 
		{
			chute_roty_proc = chute_roty_proc-da;
			if (chute_roty_proc < 0.5)chute_roty_proc = 0.5;
		}
		else if (chute_roty_proc < 0.5)
		{
			chute_roty_proc = chute_roty_proc+da;
			if (chute_roty_proc > 0.5)chute_roty_proc = 0.5;
		}
		SetAnimation(anim_chute_roty, chute_roty_proc);


		chute_rotx_proc = chute_rotx_proc-da;
		if (chute_rotx_proc < 0.78)chute_rotx_proc = 0.78;
		SetAnimation(anim_chute_rotx, chute_rotx_proc);

		if (chute_rotx_proc == 0.78)
		{
			chute_collapse_proc = chute_collapse_proc+db;
			if (chute_collapse_proc > 1)chute_collapse_proc = 1;
		}
		SetAnimation(anim_chute_collapse, chute_collapse_proc);*/

	}


	else
	{
	VECTOR3 negaspdn = -aspd;

	double schk5 = (dotp(_V(1,0,0), negaspdn));
	if (schk5 < -1)schk5 = -1;
	else if (schk5 > 1)schk5 = 1;

	double yrot_angle = abs(acos(schk5));
	chute_roty_proc = yrot_angle/PI;
	if (chute_roty_proc > 1)chute_roty_proc = 1;
	else if (chute_roty_proc < 0)chute_roty_proc = 0;

	double xrot_angle = atan2(negaspdn.y,negaspdn.z);
	if(xrot_angle < 0) xrot_angle = xrot_angle+PI2;
	chute_rotx_proc = xrot_angle/PI2;
	if (chute_rotx_proc > 1)chute_rotx_proc = chute_rotx_proc - (int)(chute_rotx_proc);
	else if (chute_rotx_proc < 0)chute_rotx_proc = chute_rotx_proc - (int)(chute_rotx_proc) + 1;

	SetAnimation(anim_chute_roty, chute_roty_proc);
	SetAnimation(anim_chute_rotx, chute_rotx_proc);
	}
	
	//sprintf(oapiDebugString(), "chute_roty_proc %0.3f chute_rotx_proc %0.3f", chute_roty_proc,chute_rotx_proc);
	
}

//VECTOR3 dm2_capsule::RotMat(VECTOR3 CURR_VECT,VECTOR3 ROT_AXIS,double rot_angle)
//{
//	double ct = cos(rot_angle);
//	double st = sin(rot_angle);
//
//	double ux = ROT_AXIS.x;
//	double uy = ROT_AXIS.y;
//	double uz = ROT_AXIS.z;
//	double ux2 = ux*ux;
//	double uy2 = uy*uy;
//	double uz2 = uz*uz;
//
//	MATRIX3 RG;
//
//	RG.m11 = ux2+(1-ux2)*ct;
//	RG.m12 = ux*uy*(1-ct)-uz*st;
//	RG.m13 = ux*uz*(1-ct)+uy*st;
//	RG.m21 = ux*uy*(1-ct)+uz*st;
//	RG.m22 = uy2+(1-uy2)*ct;
//	RG.m23 = uy*uz*(1-ct)-ux*st;
//	RG.m31 = ux*uz*(1-ct)-uy*st;
//	RG.m32 = uy*uz*(1-ct)+ux*st;
//	RG.m33 = uz2+(1-uz2)*ct;
//
//	VECTOR3 result_vector = mul(RG, CURR_VECT);
//	return result_vector;
//
//}

//void dm2_capsule::GetFlightTime ()
//{
//	double flt_tgt_arc = tgt_arc;
//	int fttasw = 0;
//	if (flt_tgt_arc > PI2) 
//	{
//		flt_tgt_arc = flt_tgt_arc-PI2;
//		fttasw = 1;
//	}
//
//	double apd = prm.ApD;
//	double ped = prm.PeD;
//
//	double sma = 0.5*(apd + ped);
//	double ecc = 1-(2/((apd/ped)+1));
//	
//	double to = 0;
//	double tf = 0;
//	double vo = prm.TrA;
//	double vf = vo+flt_tgt_arc;
//
//	double schk6 = ((ecc + cos(vo))/(1+(ecc*cos(vo))));
//	if (schk6 < -1)schk6 = -1;
//	else if (schk6 > 1)schk6 = 1;
//
//	double eo = acos(schk6);
//	if (vo > PI) eo = PI2 - eo;
//
//	double schk7 = (ecc + cos(vf))/(1+(ecc*cos(vf)));
//	if (schk7 < -1)schk7 = -1;
//	else if (schk7 > 1)schk7 = 1;
//
//	double ef = acos(schk7);
//
//	if (vf > 3*PI) ef = 4*PI - ef;
//	else if (vf > PI2) ef = ef + PI2;
//	else if (vf > PI) ef = PI2 - ef;
//
//	
//	double mo = eo - (ecc*sin(eo));
//	double mf = ef - (ecc*sin(ef));
//	//sprintf(oapiDebugString(), "eca %0.2f eo %0.2f ef %0.1f mo %0.2f mf %0.1f",prm.EcA*DEG, eo*DEG, ef*DEG, mo*DEG, mf*DEG);
//	if (mf < mo) mf = mf + PI2;
//
//	double schk1 = (mu/(sma*sma*sma));
//	if (schk1 < 0)schk1 = -schk1;
//
//	double n = sqrt(schk1);
//	tf = to + ((mf-mo)/n);
//	flt_time = tf;
//
//	if (fttasw == 1) flt_time = flt_time+prm.T;
//
//}

//void dm2_capsule::DoNav ()
//{	
//	OBJHANDLE hasparent = GetAttachmentStatus(hF9attach);
//	if (hasparent)
//		{
//		flt_time = 0;
//		tgt_arc = 0;
//		req_inc = 0;
//		return;
//		}
//	else hasparent =  GetAttachmentStatus(hRMSattach);
//	if (hasparent)
//		{
//		flt_time = 0;
//		tgt_arc = 0;
//		req_inc = 0;
//		return;
//		}
//	else hasparent =  GetAttachmentStatus(hHoistattach);
//		if (hasparent)
//		{
//		flt_time = 0;
//		tgt_arc = 0;
//		req_inc = 0;
//		return;
//		}
//
//	//////////////////////////////
//	hGref = GetGravityRef();
//	GetElements (hGref,el,&prm,0,FRAME_EQU);
//
//	//////////////////// check if targeting valid
//	OBJHANDLE hTgt = oapiGetObjectByName(tgt_name);
//	if (hTgt == NULL)hTgt = oapiGetBaseByName(hGref, tgt_name);
//	if (hTgt == NULL)
//	{
//		flt_time = 0;
//		tgt_arc = 0;
//		req_inc = 0;
//		if (auto_reentry_sw == 1)auto_reentry_sw = 2;
//		return;
//	}
//
//	/////////////////////////////////
//	
//
//	double alti = GetAltitude();
//	
//	VECTOR3 rpos_glob = _V(0,0,0);
//	GetRelativePos(hGref,rpos_glob);
//
//	VECTOR3 tgt_rpos_glob = _V(0,0,0);
//	oapiGetRelativePos(hTgt,hGref,&tgt_rpos_glob);
//
//	VECTOR3 tgt_pos_glob = _V(0,0,0);
//	oapiGetGlobalPos(hTgt, &tgt_pos_glob);
//
//	double tgt_off_angle = flt_time*earth_rot_rate;
//	VECTOR3 tgt_rpos_glob_rot = RotMat(tgt_rpos_glob, EARTH_AXIS, -tgt_off_angle);
//	
//	VECTOR3 tgt_rpos_glob_rot_nzd = tgt_rpos_glob_rot;
//	normalise(tgt_rpos_glob_rot_nzd);
//	VECTOR3 rpos_glob_nzd = rpos_glob;
//	normalise(rpos_glob_nzd);
//	VECTOR3 tgt_rpos_glob_nzd = tgt_rpos_glob;
//	normalise(tgt_rpos_glob_nzd);
//
//	//tgt_arc = acos(dotp(tgt_rpos_glob_rot_nzd,rpos_glob_nzd));
//	
//
//	VECTOR3 rvel_glob = _V(0,0,0);
//	GetRelativeVel(hGref, rvel_glob);
//	VECTOR3 rvel_glob_nzd = rvel_glob;
//	normalise(rvel_glob_nzd);
//
//	VECTOR3 ptr_req_orb_nml_ecl = crossp(tgt_rpos_glob, rpos_glob);
//	normalise(ptr_req_orb_nml_ecl);
//
//	VECTOR3 orb_nml_ecl = crossp(rvel_glob_nzd,rpos_glob_nzd);
//	normalise(orb_nml_ecl);
//
//	VECTOR3 target_ref_nml = crossp(orb_nml_ecl,tgt_rpos_glob_rot_nzd);
//	normalise(target_ref_nml);
//
//	VECTOR3 target_ref_in_pln = crossp(target_ref_nml,orb_nml_ecl);
//	normalise(target_ref_in_pln);
//
//	double schk8 = (dotp(target_ref_in_pln,rpos_glob_nzd));
//	if (schk8 < -1)schk8 = -1;
//	else if (schk8 > 1)schk8 = 1;
//
//	tgt_arc = acos(schk8);
//
//	VECTOR3 ves_tgt_nml = crossp(target_ref_in_pln, rpos_glob_nzd);
//	normalise(ves_tgt_nml);
//
//	double app_away = dotp(ves_tgt_nml,orb_nml_ecl);
//	
//
//	if (alti < 80000)tgt_arc = tgt_arc;
//	else if (auto_burn_sw == 2)tgt_arc = tgt_arc;
//	else if(tgt_arc < (90*RAD) && (app_away > 0)) tgt_arc = tgt_arc + PI2;
//	else if (app_away < 0) tgt_arc = PI2 - tgt_arc;
//
//	VECTOR3 req_orb_nml_ecl = crossp(tgt_rpos_glob_rot_nzd, rpos_glob);
//	normalise(req_orb_nml_ecl);
//
//	VECTOR3 req_tan_vel_vect_glob = crossp(rpos_glob, req_orb_nml_ecl);
//	normalise(req_tan_vel_vect_glob);
//
//	VECTOR3 ptr_req_tan_vel_vect_glob = crossp(rpos_glob, ptr_req_orb_nml_ecl);
//	normalise(ptr_req_tan_vel_vect_glob);
//	
//	VECTOR3 pos_glob;
//	GetGlobalPos(pos_glob);
//
//	VECTOR3 ptr_req_tan_vel_vect_glob_pos = ptr_req_tan_vel_vect_glob+pos_glob;
//	VECTOR3 ptr_req_tan_vel_vect_loc = _V(0,0,0);
//	Global2Local(ptr_req_tan_vel_vect_glob_pos,ptr_req_tan_vel_vect_loc);
//	VECTOR3 ptr_req_tan_vel_vect_hrz = _V(0,0,0);
//	HorizonRot(ptr_req_tan_vel_vect_loc, ptr_req_tan_vel_vect_hrz);
//	tgt_hdg = atan2(ptr_req_tan_vel_vect_hrz.x, ptr_req_tan_vel_vect_hrz.z);
//
//	VECTOR3 req_tan_vel_vect_glob_pos = req_tan_vel_vect_glob+pos_glob;
//	VECTOR3 req_tan_vel_vect_loc = _V(0,0,0);
//	Global2Local(req_tan_vel_vect_glob_pos,req_tan_vel_vect_loc);
//	VECTOR3 req_tan_vel_vect_hrz = _V(0,0,0);
//	HorizonRot(req_tan_vel_vect_loc, req_tan_vel_vect_hrz);
//	req_hdg = atan2(req_tan_vel_vect_hrz.x, req_tan_vel_vect_hrz.z);
//
//	double landing_range = 0;
//	if(alti < 25000)
//	{
//	VECTOR3 tgt_pos_loc = _V(0,0,0);
//	Global2Local(tgt_pos_glob,tgt_pos_loc);
//	VECTOR3 tgt_pos_hrz = _V(0,0,0);
//	HorizonRot(tgt_pos_loc, tgt_pos_hrz);
//	req_hdg = atan2(tgt_pos_hrz.x, tgt_pos_hrz.z);
//	landing_range = length(_V(tgt_pos_hrz.x,0,tgt_pos_hrz.z));
//	}
//
//
//	VECTOR3 rvel_hrz = _V(0,0,0);
//	VECTOR3 rvel_loc = _V(0,0,0);
//	VECTOR3 rvel_glob_pos = rvel_glob+pos_glob;
//	Global2Local(rvel_glob_pos,rvel_loc);
//	HorizonRot(rvel_loc, rvel_hrz);
//	if(alti < 25000) GetGroundspeedVector(FRAME_HORIZON,rvel_hrz);
//	curr_hdg = atan2(rvel_hrz.x, rvel_hrz.z);
//
//	double schk9 = (dotp(EARTH_AXIS,req_orb_nml_ecl));
//	if (schk9 < -1)schk9 = -1;
//	else if (schk9 > 1)schk9 = 1;
//
//	req_inc = acos(schk9);
//	if (req_inc > PI)req_inc =  req_inc-PI;
//	else if (req_inc < -PI)req_inc =  req_inc+PI;
//	if (req_inc > PI05)req_inc =  PI-req_inc;
//
//	tgt_dist = tgt_arc*rt;
//	if(alti < 25000) tgt_dist = landing_range;
//
//	GetFlightTime();
//	if(alti < 25000) flt_time = 0;
//
//	double schk10 = (dotp(tgt_rpos_glob_rot_nzd,orb_nml_ecl));
//	if (schk10 < -1)schk10 = -1;
//	else if (schk10 > 1)schk10 = 1;
//
//	crossrange_error = acos(schk10);
//	crossrange_error = crossrange_error - PI05;
//	if (crossrange_error < 0)crossrange_error = -crossrange_error;
//	crossrange_error = crossrange_error*rt;
//
//	double arc_tgt_to_3000 = 26.9796*RAD;
//	double my_ta = prm.TrA;
//	double tgt_ta = my_ta+tgt_arc;
//	double ei_ta = tgt_ta-arc_tgt_to_3000;
//
//	double ecc = el.e;
//	double sma = el.a;
//
//	ei_radius = sma*(1-(ecc*ecc))/(1+(ecc*(cos(ei_ta))));
//	double ei_omega = atan((ecc*(sin(ei_ta)))/(1+(ecc*(cos(ei_ta)))));
//
//	double ang_past_lan = prm.TrA + el.omegab - el.theta;
//	double end_lan_ang = ang_past_lan+tgt_arc;//-arc_tgt_to_3000;
//	if (end_lan_ang > PI2)end_lan_ang = end_lan_ang - PI2;
//	if (end_lan_ang > PI)end_lan_ang = end_lan_ang - PI;
//
//	//sprintf(oapiDebugString(), "ei_radius %0.1f ei_omega %0.1f epl %0.1f", ei_radius*0.001, ei_omega*DEG, end_lan_ang*DEG);
//
//	//sprintf(oapiDebugString(), "tgtdst %0.1f tgt_hdg %0.1f req_hdg %0.1f crs_hdg %0.1f flttime %0.1f tgt_off_angle %0.1f cre %0.1f appretrl %d apprettg %d",tgt_dist*0.001, tgt_hdg*DEG, req_hdg*DEG, curr_hdg*DEG, flt_time, tgt_off_angle*DEG, crossrange_error*0.001, appret_real, appret_tgt);
//
//	////////////////////// reentry burn ///////////////////////////////
//
//	double schk11 = (dotp(rvel_glob_nzd, rpos_glob_nzd));
//	if (schk11 < -1)schk11 = -1;
//	else if (schk11 > 1)schk11 = 1;
//
//	double lambda_1 = acos(schk11);
//	double phi_1 = PI05 - lambda_1;
//	double v_1 = length(rvel_glob);
//	double r_1 = alti+rt;
//
//	double r_2 = sma*(1-(ecc*ecc))/(1+(ecc*(cos(my_ta+arc_to_burn))));
//
//	double schk2 = (mu*((2/r_2)-(1/sma)));
//	if (schk2 < 0)schk2 = -schk2;
//
//	v_2 = sqrt(schk2);
//	double phi_2 = atan((ecc*(sin(my_ta+arc_to_burn)))/(1+(ecc*(cos(my_ta+arc_to_burn)))));
//	lambda_2 = PI05 - phi_2;
//
//	double lambda_3 = 91.05*RAD;
//	double phi_3 = PI05 - lambda_3;
//	double r_3_tgt = r_3 + r_3_fact;
//
//	double c = (r_2*sin(lambda_2))/(r_3_tgt*sin(lambda_3));
//	double r_av = (r_2+r_3_tgt)/2;
//	double g_av = mu/(r_av*r_av);
//
//	double schk3 = (2*g_av*(r_2-r_3_tgt)/((c*c)-1));
//	if (schk3 < 0)schk3 = -schk3;
//
//	v_2pb = sqrt(schk3);
//	double v_3 = c*v_2pb;
//
//	double k_2 = r_2*v_2pb*v_2pb/mu;
//	double k_3 = r_3_tgt*v_3*v_3/mu;
//
//	double ta_2 = atan2((k_2*cos(phi_2)*sin(phi_2)),((k_2*cos(phi_2)*cos(phi_2))-1));
//	if (ta_2 < 0)ta_2 = ta_2+PI2;
//	double ta_3 = atan2((k_3*cos(phi_3)*sin(phi_3)),((k_3*cos(phi_3)*cos(phi_3))-1));
//	if (ta_3 < 0)ta_3 = ta_3+PI2;
//
//	double reentry_arc = abs(ta_3-ta_2);
//	double burn_ta = ei_ta - reentry_arc;
//	arc_to_burn = burn_ta - my_ta;
//
//	//sprintf(oapiDebugString(), "ei_rad %0.1f ei_om %0.1f arc_to_burn %0.1f v2 %0.1f v2pb %0.1f", ei_radius*0.001, ei_omega*DEG, arc_to_burn*DEG, v_2, v_2pb);
//
//}

//void dm2_capsule::CalculateBank()
//{
//	double altd = GetAltitude();
//
//	//if (tgt_dist < 5000000)DoEnergyManagement();
//
//	double angle_diff = req_hdg - curr_hdg;
//	if (angle_diff > PI) angle_diff = angle_diff-PI2;
//	else if (angle_diff < -PI) angle_diff = angle_diff+PI2;
//
//	double crossrange_deadband = 2;
//	double req_bank_angle = 80*RAD;
//
//	if(altd < 95000)
//	{
//		crossrange_deadband = 1.5;
//		req_bank_angle = (70+bank_correction)*RAD;
//	}
//
//	if(altd < 85000) 
//	{
//		if (tgt_dist < 5000000)DoEnergyManagement();
//	}
//
//	if(altd < 70000)
//	{
//		
//		crossrange_deadband = 1.25;
//		req_bank_angle = (55+bank_correction)*RAD;
//	}
//
//	if(altd < 50000)
//	{
//		crossrange_deadband = 1;
//		req_bank_angle = (60+bank_correction)*RAD;
//	}
//
//	if(altd < 30000)
//	{
//		crossrange_deadband = 0.5;
//		req_bank_angle = (30+bank_correction)*RAD;
//	}
//
//
//	if(altd < 25000)
//	{
//		req_bank_angle = 10*angle_diff;
//		if (req_bank_angle > 30*RAD) req_bank_angle = 30*RAD;
//		else if (req_bank_angle < -30*RAD) req_bank_angle = -30*RAD;
//		bank_phase_sw = 2;
//	}
//
//	if(altd < 18000)
//	{
//		bank_phase_sw = 2;
//		req_bank_angle = 0;
//	}
//
//	
//
//	if (req_bank_angle > 110*RAD) req_bank_angle = 110*RAD;
//	else if (req_bank_angle < 0) req_bank_angle = 0;
//
//	if (bank_phase_sw == 0)
//	{
//		if (angle_diff > (crossrange_deadband*RAD)) bank_phase_sw = 1;
//		bank_angle = -req_bank_angle;
//	}
//	else if (bank_phase_sw == 1)
//	{
//		if (angle_diff < -(crossrange_deadband*RAD)) bank_phase_sw = 0;
//		bank_angle = req_bank_angle;
//	}
//	else if (bank_phase_sw == 2)
//	{
//	bank_angle = req_bank_angle;
//	}
//	
//
//}

//void dm2_capsule::CalculateBankNoTarget(void)
//{
//	double req_bank_angle = 80*RAD;
//
//	if(GetAltitude() < 95000) req_bank_angle = -70*RAD;
//	if(GetAltitude() < 85000) req_bank_angle = 70*RAD;
//	if(GetAltitude() < 70000) req_bank_angle = -55*RAD;
//	if(GetAltitude() < 60000) req_bank_angle = 55*RAD;
//	if(GetAltitude() < 50000) req_bank_angle = -60*RAD;
//	if(GetAltitude() < 40000) req_bank_angle = 60*RAD;
//	if(GetAltitude() < 30000) req_bank_angle = -30*RAD;
//	if(GetAltitude() < 25000) req_bank_angle = 15*RAD;
//	if(GetAltitude() < 20000) req_bank_angle = 0*RAD;
//
//	bank_angle = req_bank_angle;
//}

////void dm2_capsule::CalculateFlightVector()
////{
////	VECTOR3 haspdv = _V(0,0,0);
////	GetGroundspeedVector(FRAME_HORIZON,haspdv);
////	haspdv = -haspdv;
////	normalise(haspdv);
////
////	VECTOR3 wv = _V(0,-1,0);
////	//GetWeightVector(wv);
////	//normalise(wv);
////
////	VECTOR3 vel_wv_nml = crossp(wv, haspdv);
////	normalise(vel_wv_nml);
////	double rotangle = 10*RAD;
////	VECTOR3 fv_hrz = RotMat(haspdv, vel_wv_nml, rotangle);
////	VECTOR3 fv_hrz_banked = RotMat(fv_hrz, haspdv, bank_angle);
////
////	HorizonInvRot(fv_hrz_banked,flight_vector);
////}
//
void dm2_capsule::GoToFlightVector()
{
	//VECTOR3 wv = _V(0,0,0);
	//GetWeightVector(wv);
	//wv = -wv;
	double vrot_lim = 5*RAD;
	if(abort_status > 0)vrot_lim = 20*RAD;
	
	//double angoa = atan2(-wv.y, wv.z);
	//double slipage = atan2(-wv.x, wv.z);
	DeactivateNavmode(NAVMODE_KILLROT);
	DeactivateNavmode(NAVMODE_HLEVEL);
	DeactivateNavmode(NAVMODE_PROGRADE);
	DeactivateNavmode(NAVMODE_RETROGRADE);
	DeactivateNavmode(NAVMODE_NORMAL);
	DeactivateNavmode(NAVMODE_ANTINORMAL);
	DeactivateNavmode(NAVMODE_HOLDALT);

	double angoa = atan2(-flight_vector.y, flight_vector.z);
	double slipage = atan2(flight_vector.x, flight_vector.z);

	if(angoa > PI)angoa = (PI*2)-angoa;
	else if(angoa < -PI)angoa = (PI*2)+angoa;

	if(slipage > PI)slipage = (PI*2)-slipage;
	else if(slipage < -PI)slipage = (PI*2)+slipage;

	double mass = GetMass();

	double max_rcs = RCSTH;
	double max_torque = 2*max_rcs*1;
	double inertia = 1.9*mass;
	
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double max_acc = max_torque/inertia;
	double max_drot = max_acc*step_time;
	/////////////////////////////////////////////////////////
	//double reqxv = 0.9*(sqrt(abs(angoa*max_acc*2))); //0.9
	//if (reqxv > vrot_lim)reqxv = vrot_lim;
	//if (angoa > 0)reqxv = -reqxv;
	//double req_drotx = reqxv - avel.x;
	//double thx_lvl = req_drotx/max_drot;
	//if((angoa*DEG) < 1 && (angoa*DEG) > -1)
	//{
	//reqxv = 1*-angoa;
	//double acc_reqx = (reqxv-avel.x)/4.0; // 0.1
	//thx_lvl = acc_reqx/max_acc;
	//}

	//double reqyv = 0.9*(sqrt(abs(slipage*max_acc*2))); //0.9
	//if (reqyv > vrot_lim)reqyv = vrot_lim;
	//if (slipage > 0)reqyv = -reqyv;
	//double req_droty = reqyv - avel.y;
	//double thy_lvl = req_droty/max_drot;
	//if((slipage*DEG) < 1 && (slipage*DEG) > -1)
	//{
	//reqyv = 1*-slipage;
	//double acc_reqy = (reqyv-avel.y)/4.0; //0.1
	//thy_lvl = acc_reqy/max_acc;
	//}
	/////////////////////////////////////////////////
	double reqxv = 1*abs(angoa); //1
	if (reqxv > vrot_lim)reqxv = vrot_lim;
	if (angoa > 0)reqxv = -reqxv;
	double req_drotx = reqxv - avel.x;
	double thx_lvl = req_drotx/max_drot;
	if((angoa*DEG) < 0.5 && (angoa*DEG) > -0.5) thx_lvl = 0.25*thx_lvl;

	double reqyv = 1*abs(slipage);//1
	if (reqyv > vrot_lim)reqyv = vrot_lim;
	if (slipage > 0)reqyv = -reqyv;
	double req_droty = reqyv - avel.y;
	double thy_lvl = req_droty/max_drot;
	if((slipage*DEG) < 0.5 && (slipage*DEG) > -0.5) thy_lvl = 0.25*thy_lvl;
	//sprintf(oapiDebugString(), "slip %0.4f yrot %0.4f reqyv%0.4f", (slipage*DEG),(avel.y*DEG), (reqyv*DEG));


	if(thx_lvl > 1)thx_lvl = 1;
	else if(thx_lvl < -1)thx_lvl = -1;

	if(thy_lvl > 1)thy_lvl = 1;
	else if(thy_lvl < -1)thy_lvl = -1;


	if(thx_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
		}
	else if(thx_lvl < 0)
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, -thx_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
	}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_PITCHUP, 0);
		SetThrusterGroupLevel(THGROUP_ATT_PITCHDOWN, 0);
	}


	if(thy_lvl > 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		}
	else if(thy_lvl < 0)
		{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, -thy_lvl);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
		}
	else 
	{
		SetThrusterGroupLevel(THGROUP_ATT_YAWRIGHT, 0);
		SetThrusterGroupLevel(THGROUP_ATT_YAWLEFT, 0);
	}


}

//void dm2_capsule::RollToBankAngle()
//{
//double vrot_lim = 10*RAD;
//
//VECTOR3 haspdv = _V(0,0,1);
//GetGroundspeedVector(FRAME_HORIZON,haspdv);
//haspdv = -haspdv;
//normalise(haspdv);
//
//if (auto_burn_sw == 1) haspdv = _V(0,-1,0);
//
//VECTOR3 haspdv_loc = _V(0,0,1);
//HorizonInvRot(haspdv, haspdv_loc);
//
//double roll_angle = atan2(haspdv_loc.x,-haspdv_loc.y);
//if (roll_angle > PI) roll_angle = roll_angle - PI2;
//else if (roll_angle < -PI) roll_angle = roll_angle + PI2;
//
////sprintf(oapiDebugString(), "roll_angle %0.1f", roll_angle*DEG);
//
//	double mass = GetMass();
//
//	double max_rcs = 1*RCSTH;
//	double max_torque = 2*max_rcs*2;
//	double inertia = 1.72*mass;
//
//	VECTOR3 avel = _V(0,0,0);
//	GetAngularVel(avel);
//
//
//	double max_acc = max_torque/inertia;
//	double max_drot = max_acc*step_time;
//
//
//	double reqzv = 1*abs(roll_angle);
//	if (reqzv > vrot_lim)reqzv = vrot_lim;
//	if (roll_angle > 0)reqzv = -reqzv;
//	double req_drot = reqzv - avel.z;
//	double thz_lvl = req_drot/max_drot;
//	if((roll_angle*DEG) < 0.5 && (roll_angle*DEG) > -0.5) thz_lvl = 0.25*thz_lvl;
//
//	/*if((roll_angle*DEG) < 1 && (roll_angle*DEG) > -1)
//	{
//	reqzv = 1*-roll_angle;
//	double acc_reqz = (reqzv-avel.z)/0.1;
//	thz_lvl = acc_reqz/max_acc;
//	}*/
//
//
//	if(thz_lvl > 1)thz_lvl = 1;
//	else if(thz_lvl < -1)thz_lvl = -1;
//
//
//	if(thz_lvl > 0)
//	{
//		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, thz_lvl);
//		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
//	}
//	else if(thz_lvl < 0)
//	{
//		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, -thz_lvl);
//		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
//	}
//	else 
//	{
//		SetThrusterGroupLevel(THGROUP_ATT_BANKRIGHT, 0);
//		SetThrusterGroupLevel(THGROUP_ATT_BANKLEFT, 0);
//	}
//
//
//}


//void dm2_capsule::DoReentryBurn()
//{
//
//	int no_tgt_sw = 0;
//	//////////////////// check if targeting valid
//	OBJHANDLE hTgt = oapiGetObjectByName(tgt_name);
//	if (hTgt == NULL)hTgt = oapiGetBaseByName(hGref, tgt_name);
//	if (hTgt == NULL)
//	{
//		no_tgt_sw = 1;
//		tgt_dist = 0;
//		crossrange_error = 0;
//	}
//
//	/////////////////////////////////
//
//	VECTOR3 rvel_glob = _V(0,0,0);
//	GetRelativeVel(hGref, rvel_glob);
//
//	VECTOR3 pos_glob;
//	GetGlobalPos(pos_glob);
//
//	VECTOR3 rvel_glob_pos = rvel_glob+pos_glob;
//	VECTOR3 rvel_glob_pos_loc = _V(0,0,0);
//	Global2Local(rvel_glob_pos,rvel_glob_pos_loc);
//
//	VECTOR3 rvel_glob_pos_loc_nzd = rvel_glob_pos_loc;
//	normalise(rvel_glob_pos_loc_nzd);
//
//	VECTOR3 rvel_glob_hrz = _V(0,0,0);
//	HorizonRot(rvel_glob_pos_loc, rvel_glob_hrz);
//	
//	flight_vector = rvel_glob_pos_loc_nzd;
//	bank_angle = PI;
//
//	double schk12 = (dotp(flight_vector,_V(0,0,1)));
//	if (schk12 < -1)schk12 = -1;
//	else if (schk12 > 1)schk12 = 1;
//
//	double pointing_error = acos(schk12);
//
//	if (no_tgt_sw == 1 && auto_burn_sw == 1) ////////////////////// no tgt
//	{
//		if (prm.PeD > (rt+50000))
//		{
//			if (pointing_error < 2*RAD) SetThrusterGroupLevel(THGROUP_ATT_BACK, 1);
//			else SetThrusterGroupLevel(THGROUP_ATT_BACK, 0);
//			burn_sw = 1;
//		}
//		else
//		{
//		SetThrusterGroupLevel(THGROUP_ATT_BACK, 0);
//		auto_burn_sw = 2;
//		burn_sw = 0;
//		oapiVCTriggerRedrawArea (-1, AUTO_BUTTONS);
//		}
//	}
//
//	else /////////////// targeting
//	{
//	double req_dv = length(rvel_glob) - v_2pb;
//	double tot_mass = GetMass();
//
//	double fraction = exp(req_dv/ISP);
//	double fuelmass = tot_mass - (tot_mass/fraction);
//	double fuelrate = 4*RCSTH/ISP;
//	double dvtime = fuelmass/fuelrate;
//
//	double ang_past_lan = prm.TrA + el.omegab - el.theta;
//	double sin_inc = sin(el.i);
//	double inc_factor = sin_inc*sin_inc*1.626; // 1 at 51.65 inc
//
//	//r_3_fact = ((sin(end_lan_ang))*38330*inc_factor*sma_factor)-11000;
//	//sprintf(oapiDebugString(), "r_3_fact %0.1f", r_3_fact);
//
//	//////////////////////////////////////////////////////////// alternative ending
//	double tgt_arc_fact = (tgt_arc-(RAD*26.9796))/(120*RAD);
//	double om_off = -13.7*RAD;
//	r_3_fact = inc_factor*tgt_arc_fact*(-10457*0.5*(cos((2*((ang_past_lan+tgt_arc)-(RAD*26.9796)))+ om_off) - cos((2*(ang_past_lan)) + om_off))) + 1930;
//	if (NonsphericalGravityEnabled() == false)r_3_fact = 0;
//
//
//	//sprintf(oapiDebugString(), "r_3_fact %0.1f", r_3_fact);
//
//	double r_3_lim = r_3+r_3_fact;
//
//	GetTimeToBurn ();
//	time_to_burn = time_to_burn - (0.5*dvtime);
//	if(no_tgt_sw == 1)time_to_burn = -0.001;
//
//	if (time_to_burn < 0 && auto_burn_sw == 1)
//	{
//	if (ttb_check_sw == 1)burn_sw = 1;
//	ttb_check_sw = ttb_check_sw+1;
//	}
//	else ttb_check_sw = 0;
//		
//
//	if (burn_sw == 1)
//		{
//		if (ei_radius > r_3_lim)
//			{
//			if (pointing_error < 2*RAD) SetThrusterGroupLevel(THGROUP_ATT_BACK, 1);
//			else SetThrusterGroupLevel(THGROUP_ATT_BACK, 0);
//			}
//		else
//			{
//			SetThrusterGroupLevel(THGROUP_ATT_BACK, 0);
//			burn_sw = 0;
//			auto_burn_sw = 2;
//			oapiVCTriggerRedrawArea (-1, AUTO_BUTTONS);
//			}
//		}
//	}
//
//}

//void dm2_capsule::GetTimeToBurn ()
//{
//	
//	double flt_tgt_arc = arc_to_burn;
//	int fttasw = 0;
//	if (flt_tgt_arc > PI2) 
//	{
//		flt_tgt_arc = flt_tgt_arc-PI2;
//		fttasw = 1;
//	}
//
//	double apd = prm.ApD;
//	double ped = prm.PeD;
//
//	double sma = 0.5*(apd + ped);
//	double ecc = 1-(2/((apd/ped)+1));
//	
//	double to = 0;
//	double tf = 0;
//	double vo = prm.TrA;
//	double vf = vo+flt_tgt_arc;
//
//	double schk13 = ((ecc + cos(vo))/(1+(ecc*cos(vo))));
//	if (schk13 < -1)schk13 = -1;
//	else if (schk13 > 1)schk13 = 1;
//
//	double eo = acos(schk13);
//	if (vo > PI) eo = PI2 - eo;
//
//	double schk14 = ((ecc + cos(vf))/(1+(ecc*cos(vf))));
//	if (schk14 < -1)schk14 = -1;
//	else if (schk14 > 1)schk14 = 1;
//
//	double ef = acos(schk14);
//
//	if (vf > 3*PI) ef = 4*PI - ef;
//	else if (vf > PI2) ef = ef + PI2;
//	else if (vf > PI) ef = PI2 - ef;
//
//	
//	double mo = eo - (ecc*sin(eo));
//	double mf = ef - (ecc*sin(ef));
//	//sprintf(oapiDebugString(), "eca %0.2f eo %0.2f ef %0.1f mo %0.2f mf %0.1f",prm.EcA*DEG, eo*DEG, ef*DEG, mo*DEG, mf*DEG);
//	if (mf < mo) mf = mf + PI2;
//
//	double schk4 = (mu/(sma*sma*sma));
//	if (schk4 < 0)schk4 = -schk4;
//
//	double n = sqrt(schk4);
//	tf = to + ((mf-mo)/n);
//	time_to_burn = tf;
//
//	if (fttasw == 1) time_to_burn = time_to_burn + prm.T;
//
//}

//void dm2_capsule::DoEnergyManagement()
//{
//		double altd = GetAltitude();
//		VECTOR3 rvelg;
//		GetRelativeVel(hGref,rvelg);
//		double v2 = length(rvelg);
//		v2 = v2*v2;
//		double ke = 0.5*v2;
//		double pe = mu/(rt+altd);
//		double tot_energy = ke+pe;
//
//		double req_energy = 0;
//		double energy_diff_abs = 0;
//		double energy_diff_ratio = 0;
//		int idx = 0;
//		double dst_past_tld = 0;
//		int upper_step = 0;
//		int lower_step = 0;
//		double gradient = 0;
//		double delta_j = 0;
//		double energy_correct_time = 15;
//		int inc_dec = 0;
//		//double bcf = 1;
//		
//
//	if (tgt_dist < 1000000)
//	{
//		dst_past_tld = 1000000 - tgt_dist;
//		upper_step = (int) (dst_past_tld/step_1000);
//		lower_step = upper_step + 1;
//		if (lower_step > (ndp_1000-1)) lower_step = upper_step;
//		gradient = (J_1000[lower_step] - J_1000[upper_step])/step_1000;
//		delta_j = (dst_past_tld - (upper_step*step_1000))*gradient;
//		req_energy = J_1000[upper_step] + delta_j;
//		energy_diff_abs = req_energy - tot_energy;
//		energy_diff_ratio = tot_energy/req_energy;
//	}
//
//	else if (tgt_dist < 3000000)
//	{
//		dst_past_tld = 3000000 - tgt_dist;
//		upper_step = (int) (dst_past_tld/step_3000);
//		lower_step = upper_step + 1;
//		if (lower_step > (ndp_3000-1)) lower_step = upper_step;
//		gradient = (J_3000[lower_step] - J_3000[upper_step])/step_3000;
//		delta_j = (dst_past_tld - (upper_step*step_3000))*gradient;
//		req_energy = J_3000[upper_step] + delta_j;
//		energy_diff_abs = req_energy - tot_energy;
//		energy_diff_ratio = tot_energy/req_energy;
//	}
//
//	else
//	{
//		dst_past_tld = 5000000 - tgt_dist;
//		upper_step = (int) (dst_past_tld/step_5000);
//		lower_step = upper_step + 1;
//		if (lower_step > (ndp_5000-1)) lower_step = upper_step;
//		gradient = (J_5000[lower_step] - J_5000[upper_step])/step_5000;
//		delta_j = (dst_past_tld - (upper_step*step_5000))*gradient;
//		req_energy = J_5000[upper_step] + delta_j;
//		energy_diff_abs = req_energy - tot_energy;
//		energy_diff_ratio = tot_energy/req_energy;
//	}
//	
//	if (tot_energy_old == 0)tot_energy_old = tot_energy;
//	double curr_energy_dt = (tot_energy - tot_energy_old)/post_step;
//	tot_energy_old = tot_energy;
//	double req_energy_adjust_dt = energy_diff_abs/energy_correct_time;
//
//	if (dst_past_tld_old == 0)dst_past_tld_old = dst_past_tld;
//	double nominal_energy_dt = gradient*(dst_past_tld - dst_past_tld_old)/post_step;
//	dst_past_tld_old = dst_past_tld;
//
//	double net_req_energy_dt =  nominal_energy_dt+req_energy_adjust_dt;
//
//	//if (abs(net_req_energy_dt-curr_energy_dt) < 1000)bcf = 0.5;
//	//else bcf = 1;
//
//	double diff_abs_ratio = abs(net_req_energy_dt-curr_energy_dt)/abs(curr_energy_dt);
//
//	if (net_req_energy_dt > curr_energy_dt)inc_dec =  -1;
//	else if (net_req_energy_dt < curr_energy_dt)inc_dec = 1;
//
//	//digi_dar = 0;
//	double new_digi_dar = ((int)(8*diff_abs_ratio))*2;
//	if (new_digi_dar == digi_dar_old) digi_dar_check_sw = digi_dar_check_sw + 1;
//	else digi_dar_check_sw = 0;
//	if (digi_dar_check_sw > 5)digi_dar_check_sw = 5;
//	
//	if (digi_dar_check_sw == 5) digi_dar = new_digi_dar;
//	digi_dar_old = new_digi_dar;
//
//	bank_correction = inc_dec*(0+digi_dar);
//	//bank_correction = 0;
//
//	//sprintf(oapiDebugString(), "bank_correction %0.1f net_req_energy_dt %0.4f curr_e_dt %0.4f diff_abs_ratio %0.4f", bank_correction, net_req_energy_dt, curr_energy_dt, diff_abs_ratio);
//
//}

void dm2_capsule::Abort(void)
{
	OBJHANDLE hLauncher = GetAttachmentStatus(hF9attach);

	if(hLauncher != NULL)DelAttachment(hF9attach);
	//else
	//{
	//	abort_status = 2;
	//	return;
	//}
	VECTOR3 flight_vector_hrz = _V(0.57735,0.57735,0.57735);

	double fuel_mass = GetPropellantMass(hProp);
	double comp_level = 0.85+((fuel_mass/1750)*0.15);
	if(fuel_mass == 0)comp_level = 0.85;

	if(fuel_mass < 1750)
	{
		
		SetThrusterLevel(th_sd[0], comp_level);
		SetThrusterLevel(th_sd[1], comp_level);
		SetThrusterLevel(th_sd[2], comp_level);
		SetThrusterLevel(th_sd[3], comp_level);
		SetThrusterLevel(th_sd[4], comp_level);
		SetThrusterLevel(th_sd[5], comp_level);
		SetThrusterLevel(th_sd[6], comp_level);
		SetThrusterLevel(th_sd[7], comp_level);

		HorizonInvRot(flight_vector_hrz, flight_vector);
		GoToFlightVector();
	}
	else
	{
		
		SetThrusterLevel(th_sd[0], 0.9);
		SetThrusterLevel(th_sd[1], 0.9);
		SetThrusterLevel(th_sd[2], 0.897);
		SetThrusterLevel(th_sd[3], 0.9);
		SetThrusterLevel(th_sd[4], 0.9);
		SetThrusterLevel(th_sd[5], 0.9);
		SetThrusterLevel(th_sd[6], 0.897);
		SetThrusterLevel(th_sd[7], 0.9);
		
//		HorizonInvRot(flight_vector_hrz, flight_vector);
//		GoToFlightVector();
	}

	if (GetPropellantMass(hProp) < 60)
	{
		if(umb_status == 2)umb_status = 1;
	}
	

	if (GetPropellantMass(hProp) < 40) 
		{
			//if(umb_status == 2)umb_status = 1;
			abort_status = 2;
			//flight_vector_hrz = _V(-0.7071,0.7071,0);
			SetThrusterLevel(th_sd[0], 0);
			SetThrusterLevel(th_sd[1], 0);
			SetThrusterLevel(th_sd[2], 0);
			SetThrusterLevel(th_sd[3], 0);
			SetThrusterLevel(th_sd[4], 0);
			SetThrusterLevel(th_sd[5], 0);
			SetThrusterLevel(th_sd[6], 0);
			SetThrusterLevel(th_sd[7], 0);
			KillAllRCS();
			//chute_arm_status = 1;
		}

		if(abort_status == 2)
		{
		VECTOR3 airspdvector;
		GetAirspeedVector(FRAME_LOCAL, airspdvector);
		normalise(airspdvector);
		double asp_zaxis_angle = acos(dotp(airspdvector, _V(0,0,-1)));

		flight_vector = -airspdvector;
		GoToFlightVector();


		if(asp_zaxis_angle < (60*RAD))
			{
			chute_arm_status = 1;
			abort_status = -1;
			KillAllRCS();
			}

		}
	

}

void dm2_capsule::RegisterVC_CdrMFD ()
{
	// activate MFD function buttons
	oapiVCRegisterArea (AID_CDR1_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR2_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR3_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR4_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR5_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR6_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR7_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);

	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR1_BUTTONS, _V(-0.187, 0.706, 2.302), _V(-0.373, 0.706, 2.302),  _V(-0.187, 0.594, 2.288), _V(-0.373, 0.594, 2.288));
	//oapiVCSetAreaClickmode_Quadrilateral (AID_CDR1_BUTTONS, _V(-0.398031, 0.742513, 2.30748), _V(-0.205193, 0.742513, 2.30748),  _V(-0.398031, 0.561643, 2.28527), _V(-0.205193, 0.561643, 2.28527));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR2_BUTTONS,  _V(0.186, 0.706, 2.302),  _V(0.372, 0.706, 2.302),   _V(0.186, 0.594, 2.288), _V(0.372, 0.594, 2.288));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR3_BUTTONS, _V(-0.093, 0.706, 2.302), _V(0.093, 0.706, 2.302),  _V(-0.093, 0.594, 2.288), _V(0.093, 0.594, 2.288));

	//oapiVCSetAreaClickmode_Quadrilateral (AID_CDR4_BUTTONS, _V(-0.501573, 0.652083, 2.29638), _V(-0.398031, 0.652083, 2.29638),  _V(-0.501573, 0.561643, 2.28527), _V(-0.398031, 0.561643, 2.28527));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR4_BUTTONS, _V(-0.499, 0.631, 2.293), _V(-0.4, 0.631, 2.293),  _V(-0.499, 0.575, 2.287), _V(-0.4, 0.575, 2.287));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR5_BUTTONS,  _V(-0.499, 0.723, 2.305), _V(-0.4, 0.723, 2.305),   _V(-0.499, 0.667, 2.298), _V(-0.4, 0.667, 2.298));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR6_BUTTONS, _V(0.4, 0.723, 2.305), _V(0.499, 0.723, 2.305),   _V(0.4, 0.667, 2.298), _V(0.499, 0.667, 2.298));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR7_BUTTONS, _V(0.4, 0.631, 2.293), _V(0.499, 0.631, 2.293),  _V(0.4, 0.575, 2.287), _V(0.499, 0.575, 2.287));
	//oapiVCSetAreaClickmode_Quadrilateral (HATCH_BUTTONS, _V(-0.121407, 1.07709, 1.97221), _V(0.119913, 1.07709, 1.97221),  _V(-0.121407, 1.06146, 1.96861), _V(0.119913, 1.06146, 1.96861));
	//oapiVCSetAreaClickmode_Quadrilateral (RCS_BUTTONS, _V(-0.407345, 0.734656, 1.89316), _V(-0.287299, 0.734656, 1.89316),  _V(-0.407345, 0.705211, 1.88637), _V(-0.287299, 0.705211, 1.88637));
	//oapiVCSetAreaClickmode_Quadrilateral (AUTO_BUTTONS, _V(-0.26034, 0.734656, 1.89316), _V(-0.140295, 0.734656, 1.89316),  _V(-0.26034, 0.705211, 1.88637), _V(-0.140295, 0.705211, 1.88637));
	//oapiVCSetAreaClickmode_Quadrilateral (TGT_BUTTONS, _V(-0.119214, 0.734861, 1.89321), _V(0.000831604, 0.734861, 1.89321),  _V(-0.119214, 0.705416, 1.88642), _V(0.000831604, 0.705416, 1.88642));
	//oapiVCSetAreaClickmode_Quadrilateral (NAV_BUTTONS, _V(-0.259484, 0.689028, 1.88262), _V(-0.139577, 0.689028, 1.88262),  _V(-0.259484, 0.673393, 1.87902), _V(-0.139577, 0.673393, 1.87902));
	//oapiVCSetAreaClickmode_Quadrilateral (PANEL_BUTTONS, _V(-0.0202412, 1.13432, 1.98543), _V(0.0207905, 1.13432, 1.98543),  _V(-0.0202412, 1.11869, 1.98183), _V(0.0207905, 1.11869, 1.98183));
	//oapiVCSetAreaClickmode_Quadrilateral (SEP_BUTTONS, _V(0.0204977, 0.734861, 1.89321), _V(0.140543, 0.734861, 1.89321),  _V(0.0204977, 0.705416, 1.88642), _V(0.140543, 0.705416, 1.88642));
	//oapiVCSetAreaClickmode_Quadrilateral (LIGHT_BUTTONS, _V(0.0208565, 0.689028, 1.88262), _V(0.140763, 0.689028, 1.88262),  _V(0.0208565, 0.673393, 1.87902), _V(0.140763, 0.673393, 1.87902));
	//oapiVCSetAreaClickmode_Quadrilateral (CHUTE_BUTTONS, _V(0.160757, 0.734861, 1.89321), _V(0.280802, 0.734861, 1.89321),  _V(0.160757, 0.705416, 1.88642), _V(0.280802, 0.705416, 1.88642));

    // D. Beachy: register+activate MFD power buttons
    const double powerButtonRadius = 0.02; // radius of power button on each MFD
	oapiVCRegisterArea (AID_CDR1_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR2_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR3_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR4_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR5_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR6_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCRegisterArea (AID_CDR7_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	//oapiVCRegisterArea (AID_CDR3_PWR, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	//oapiVCRegisterArea (AID_CDR4_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
    //oapiVCSetAreaClickmode_Spherical(AID_CDR1_PWR, _V(-0.381824, 0.79592, 1.9064), powerButtonRadius);  
   // oapiVCSetAreaClickmode_Spherical(AID_CDR2_PWR, _V(-0.118535, 0.79592, 1.9064), powerButtonRadius);
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR1_PWR, _V(-0.314, 0.560, 2.281), _V(-0.248, 0.560, 2.281),  _V(-0.314, 0.540, 2.281), _V(-0.248, 0.540, 2.281));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR2_PWR, _V(0.241, 0.560, 2.281), _V(0.313, 0.560, 2.281),  _V(0.241, 0.540, 2.281), _V(0.313, 0.540, 2.281));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR3_PWR, _V(-0.042, 0.560, 2.281), _V(0.042, 0.560, 2.281),  _V(-0.042, 0.540, 2.281), _V(0.042, 0.540, 2.281));
	
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR4_PWR, _V(-0.522, 0.622, 2.29), _V(-0.512, 0.622, 2.29),  _V(-0.522, 0.58, 2.285), _V(-0.512, 0.58, 2.285));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR5_PWR, _V(-0.522, 0.717, 2.302), _V(-0.512, 0.717, 2.302),  _V(-0.522, 0.675, 2.297), _V(-0.512, 0.675, 2.297));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR6_PWR, _V(0.512, 0.717, 2.302), _V(0.522, 0.717, 2.302),  _V(0.512, 0.675, 2.297), _V(0.522, 0.675, 2.297));
	oapiVCSetAreaClickmode_Quadrilateral (AID_CDR7_PWR, _V(0.512, 0.622, 2.29), _V(0.522, 0.622, 2.29),  _V(0.512, 0.58, 2.285), _V(0.522, 0.58, 2.285));
	
	
	//input buttons
	oapiVCRegisterArea (INPUT_BUTTONS, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCSetAreaClickmode_Quadrilateral (INPUT_BUTTONS, _V(-0.551, 0.5, 2.279), _V(0.551, 0.5, 2.279),  _V(-0.551, 0.421, 2.269), _V(0.551, 0.421, 2.269));

	// override sw
	oapiVCRegisterArea (OV_SWITCH, PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN|PANEL_MOUSE_ONREPLAY);
	oapiVCSetAreaClickmode_Quadrilateral (OV_SWITCH, _V(0.287, 0.476, 2.276), _V(0.31, 0.476, 2.276),  _V(0.287, 0.444, 2.272), _V(0.31, 0.444, 2.272));

}

// --------------------------------------------------------------
// Respond to virtual cockpit mouse event
// --------------------------------------------------------------
bool dm2_capsule::clbkVCMouseEvent (int id, int event, VECTOR3 &p)
{
	
	
	switch (id)
	{
	// handle non default MFD selection buttons
	
	case AID_CDR2_BUTTONS:
	case AID_CDR3_BUTTONS:
	case AID_CDR4_BUTTONS:
	case AID_CDR5_BUTTONS:
	case AID_CDR6_BUTTONS:
	case AID_CDR7_BUTTONS:
		{
		int mfd = id-AID_CDR1_BUTTONS+MFD_LEFT;
		int bt = (int)(p.y*5.99);
		int splt = (int)(p.x*1.99);
		oapiProcessMFDButton (mfd, (bt+(splt*6)), event);
		//sprintf(oapiDebugString(), "mfd %d", mfd);
		} return true;

	// handle Left MFD selection buttons
	case AID_CDR1_BUTTONS:
		{
		int mfd = id-AID_CDR1_BUTTONS+MFD_LEFT;
		int bt = (int)(p.y*5.99);
		int splt = (int)((1-p.x)*1.99);
		oapiProcessMFDButton (mfd, (bt+(splt*6)), event);
		//sprintf(oapiDebugString(), "mfd %d", mfd);
		} return  true;
		
   //handle power/sel buttons BIG screens
    case AID_CDR1_PWR:
    case AID_CDR2_PWR:
	case AID_CDR3_PWR:
		{
        int mfd = id - AID_CDR1_PWR+MFD_LEFT;
		int bt = (int)(p.x*2.99);
			{
			if(bt == 0)oapiSendMFDKey(mfd, OAPI_KEY_ESCAPE);
			else if(bt == 1)oapiSendMFDKey(mfd, OAPI_KEY_F1);
			else if(bt == 2)oapiSendMFDKey(mfd, OAPI_KEY_GRAVE);
			}
		} return true;

	//handle power/sel buttons SMALL screens
    case AID_CDR4_PWR:
    case AID_CDR5_PWR:
	case AID_CDR6_PWR:
	case AID_CDR7_PWR:
		{
        int mfd = id - AID_CDR1_PWR+MFD_LEFT;
		int bt = (int)(p.y*2.99);
			{
			if(bt == 0)oapiSendMFDKey(mfd, OAPI_KEY_ESCAPE);
			else if(bt == 1)oapiSendMFDKey(mfd, OAPI_KEY_F1);
			else if(bt == 2)oapiSendMFDKey(mfd, OAPI_KEY_GRAVE);
			}
		} return true;

	case OV_SWITCH: // custom functions
			{
			SendBufferedKey(OAPI_KEY_7,true,"KEYMOD_CONTROL");
			} return true;

	case INPUT_BUTTONS: // input switch covers
		{
			int bt = (int)(p.x*7.99);
			{
				if(bt == 0)SendBufferedKey(OAPI_KEY_1,true,"KEYMOD_CONTROL");
				else if(bt == 1)SendBufferedKey(OAPI_KEY_2,true,"KEYMOD_CONTROL");
				else if(bt == 2)SendBufferedKey(OAPI_KEY_3,true,"KEYMOD_CONTROL");
				else if(bt == 5)SendBufferedKey(OAPI_KEY_4,true,"KEYMOD_CONTROL");
				else if(bt == 6)SendBufferedKey(OAPI_KEY_5,true,"KEYMOD_CONTROL");
				else if(bt == 7)SendBufferedKey(OAPI_KEY_6,true,"KEYMOD_CONTROL");
			}
		} return true;
              
	

	//case HATCH_BUTTONS: // custom functions
	//		{
	//		if(panel_status == 0)
	//			{
	//			int bx = (int)(p.x*2.99);
	//			if(bx == 0)SendBufferedKey(OAPI_KEY_B);
	//			else if(bx == 1)SendBufferedKey(OAPI_KEY_K);
	//			else if(bx == 2)SendBufferedKey(OAPI_KEY_G);
	//			oapiVCTriggerRedrawArea(-1,HATCH_BUTTONS);
	//			}
	//		else SendBufferedKey(OAPI_KEY_V);
	//		} return true;

		//case RCS_BUTTONS: // custom functions
		//	{
		//	if(panel_status == 0)
		//		{
		//		int bx = (int)(p.x*2.99);
		//		if(bx == 0)SendBufferedKey(OAPI_KEY_DIVIDE);
		//		else if(bx > 0)SendBufferedKey(OAPI_KEY_D);
		//		oapiVCTriggerRedrawArea(-1,RCS_BUTTONS);
		//		}
		//	} return true;

		//case AUTO_BUTTONS: // custom functions
		//	{
		//	if(panel_status == 0)
		//	{
		//	int by = (int)(p.y*1.99);
		//	if(by == 0)SendBufferedKey(OAPI_KEY_U);
		//	else if(by == 1)SendBufferedKey(OAPI_KEY_E);
		//	oapiVCTriggerRedrawArea(-1,AUTO_BUTTONS);
		//	}
		//	} return true;

		//case TGT_BUTTONS: // custom functions
		//	{
		//	if(panel_status == 0)
		//	{
		//	SendBufferedKey(OAPI_KEY_S);
		//	oapiVCTriggerRedrawArea(-1,TGT_BUTTONS);
		//	}
		//	} return true;

			//case NAV_BUTTONS: // custom functions
			//{
			//if(panel_status == 0)
			//	{
			//int bx = (int)(p.x*2.99);
			//if(bx == 0)SendBufferedKey(OAPI_KEY_NUMPAD5);
			//else if(bx == 1)SendBufferedKey(OAPI_KEY_LBRACKET);
			//else if(bx ==2)SendBufferedKey(OAPI_KEY_RBRACKET);
			//oapiVCTriggerRedrawArea(-1,NAV_BUTTONS);
			//	}
			//} return true;

			//case PANEL_BUTTONS: // custom functions
			//{
			//SendBufferedKey(OAPI_KEY_V);
			//oapiVCTriggerRedrawArea(-1,PANEL_BUTTONS);
			//} return true;


			//case SEP_BUTTONS: // custom functions
			//{
			//if(panel_status == 0)
			//	{
			//int by = (int)(p.y*1.99);
			//if(by == 0)SendBufferedKey(OAPI_KEY_M);
			//else if(by == 1)SendBufferedKey(OAPI_KEY_J);
			//oapiVCTriggerRedrawArea(-1,SEP_BUTTONS);
			//	}
			//} return true;

			//case LIGHT_BUTTONS: // custom functions
			//{
			//if(panel_status == 0)
			//	{
			//int bx = (int)(p.x*2.99);
			//if(bx == 0)SendBufferedKey(OAPI_KEY_1);
			//else if(bx == 1)SendBufferedKey(OAPI_KEY_2);
			//else if(bx == 2)SendBufferedKey(OAPI_KEY_3);
			//oapiVCTriggerRedrawArea(-1,LIGHT_BUTTONS);
			//	}
			//} return true;

			//case CHUTE_BUTTONS: // custom functions
			//{
			//if(panel_status == 0)
			//	{
			//	SendBufferedKey(OAPI_KEY_L);
			//	oapiVCTriggerRedrawArea(-1,CHUTE_BUTTONS);
			//	}
			//} return true;

	}
	
	return false;
}

void dm2_capsule::RedrawPanel_MFDButton (SURFHANDLE surf, int mfd)
{
	//sprintf(oapiDebugString(), "RedrawPanel_MFDButton called");

	HDC hDC = oapiGetDC (surf);


	// D. Beachy: BUGFIX: if MFD powered off, cover separator lines and do not paint buttons
    if (oapiGetMFDMode(mfd) == MFD_NONE) {
       // RECT r = { 0,0,255,13 };
       // FillRect(hDC, &r, (HBRUSH)GetStockObject(BLACK_BRUSH));
		//sprintf(oapiDebugString(), "BUGFIX called");
    } else 
	{   // MFD powered on
		HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
		SetTextColor (hDC, RGB(225,255,216));
		SetTextAlign (hDC, TA_CENTER);
		SetBkMode (hDC, TRANSPARENT);
		//SetBkColor(hDC, RGB(255,0,0));
		const char *label;
		int y = 75;
		for (int bt = 0; bt < 6; bt++) {
			if (label = oapiMFDButtonLabel (mfd, bt)) {
				TextOut (hDC, 60, y, label, strlen(label));
				y += 73;
			} else break;
		}

		int y2 = 75;
		for (int bt = 6; bt < 12; bt++) {
			if (label = oapiMFDButtonLabel (mfd, bt)) {
				TextOut (hDC, 124, y2, label, strlen(label));
				y2 += 73;
			} else break;
		}
		//TextOut (hDC, 234, 1, "PG", 2);
		SelectObject (hDC, pFont);
	}

	oapiReleaseDC (surf, hDC);

}


//void dm2_capsule::RedrawHatchButton (SURFHANDLE surf)
//{
//	//sprintf(oapiDebugString(), "RedrawPanel_MFDButton called");
//
//	HDC hDC = oapiGetDC (surf);
//
//	//sprintf(oapiDebugString(), "nc_status %d",nc_status);
//
//	if (nc_status == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,200,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE CONE");
//		TextOut (hDC, 43, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (nc_status == 2)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE CONE");
//		TextOut (hDC, 43, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (nc_status == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(155,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE CONE");
//		TextOut (hDC, 43, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (nc_status == 3)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,100,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE CONE");
//		TextOut (hDC, 43, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	///////////////////////////////////////
//
//	if (hatch_status == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,200,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE HATCH");
//		TextOut (hDC, 129, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (hatch_status == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(155,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE HATCH");
//		TextOut (hDC, 129, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (hatch_status == 2)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		//SetTextColor (hDC, RGB(225,255,255));
//		//SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE HATCH");
//		TextOut (hDC, 129, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (hatch_status == 3)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		//SetTextColor (hDC, RGB(225,255,255));
//		//SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,100,0));
//		char cbuf[256];
//		strcpy(cbuf,"NOSE HATCH");
//		TextOut (hDC, 129, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	///////////////////////////////////////////////////////
//
//	if (door_status == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,200,0));
//		char cbuf[256];
//		strcpy(cbuf,"SIDE HATCH");
//		TextOut (hDC, 215, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (door_status == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(155,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"SIDE HATCH");
//		TextOut (hDC, 215, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (door_status == 2)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"SIDE HATCH");
//		TextOut (hDC, 215, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	else if (door_status == 3)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(0,100,0));
//		char cbuf[256];
//		strcpy(cbuf,"SIDE HATCH");
//		TextOut (hDC, 215, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	oapiReleaseDC (surf, hDC);
//}

//void dm2_capsule::RedrawRCSButton (SURFHANDLE surf)
//{
//	//sprintf(oapiDebugString(), "RedrawPanel_MFDButton called");
//
//	HDC hDC = oapiGetDC (surf);
//
//	//sprintf(oapiDebugString(), "nc_status %d",nc_status);
//
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"RCS");
//		TextOut (hDC, 22, 1, cbuf, strlen(cbuf));
//		strcpy(cbuf,"ROLL / YAW");
//		TextOut (hDC, 87, 1, cbuf, strlen(cbuf));
//
//	SelectObject (hDC, pFont);
//	}
//
//	int att_mode = GetAttitudeMode();
//
//	if (att_mode == RCS_NONE)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"OFF");
//		TextOut (hDC, 22, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (att_mode == RCS_LIN)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"LIN");
//		TextOut (hDC, 22, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"ROT");
//		TextOut (hDC, 22, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	/////////////////////////// reentry rcs
//
//	if (rcs_cfg == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"NORMAL");
//		TextOut (hDC, 87, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"REENTRY");
//		TextOut (hDC, 87, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	oapiReleaseDC (surf, hDC);
//}


//void dm2_capsule::RedrawAutoButton (SURFHANDLE surf)
//{
//	//sprintf(oapiDebugString(), "RedrawPanel_MFDButton called");
//
//	HDC hDC = oapiGetDC (surf);
//
//	//sprintf(oapiDebugString(), "nc_status %d",nc_status);
//
//	if (auto_burn_sw == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"AUTO DE-ORBIT BURN");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else if (auto_burn_sw == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"AUTO DE-ORBIT BURN");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(125,125,125));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"DE-ORBIT FINISHED");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	/////////////// reentry att
//
//	if (auto_reentry_sw == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"AUTO REENTRY ATT.");
//		TextOut (hDC, 64, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"AUTO REENTRY ATT.");
//		TextOut (hDC, 64, 14, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//	
//	oapiReleaseDC (surf, hDC);
//}


//void dm2_capsule::RedrawInfoTab (SURFHANDLE surf)
//{
//	HDC hDC = oapiGetDC (surf);
//
//	//sprintf(oapiDebugString(), "nc_status %d",nc_status);
//
//	if (auto_burn_sw == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		char cbuf[256];
//		strcpy(cbuf,"TtB");
//		TextOut (hDC, 21, 0, cbuf, strlen(cbuf));
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"TtB");
//		TextOut (hDC, 21, 0, cbuf, strlen(cbuf));
//	//sprintf(oapiDebugString(), "Black called");
//	SelectObject (hDC, pFont);
//	}
//
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(100,255,100));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"Rng(km)");
//		TextOut (hDC, 66, 0, cbuf, strlen(cbuf));
//		strcpy(cbuf,"X-Rng");
//		TextOut (hDC, 109, 0, cbuf, strlen(cbuf));
//
//		sprintf(cbuf, "%0.1f", tgt_dist*0.001);
//		TextOut (hDC, 66, 14, cbuf, strlen(cbuf));
//		sprintf(cbuf, "%0.1f", crossrange_error*0.001);
//		TextOut (hDC, 109, 14, cbuf, strlen(cbuf));
//
//		if (auto_burn_sw == 0)
//		{
//		sprintf(cbuf, "%d", (int)(time_to_burn));
//		TextOut (hDC, 21, 14, cbuf, strlen(cbuf));
//		}
//		else if (auto_burn_sw == 1)
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		if (burn_sw == 0)
//			{
//			sprintf(cbuf, "%d", (int)(time_to_burn));
//			TextOut (hDC, 21, 14, cbuf, strlen(cbuf));
//			}
//		else
//			{
//			sprintf(cbuf, "ACTIVE");
//			TextOut (hDC, 23, 14, cbuf, strlen(cbuf));
//			}
//		}
//		else if (auto_burn_sw == 2)
//		{
//			SetTextColor (hDC, RGB(125,125,125));
//			SetBkMode (hDC, TRANSPARENT);
//			sprintf(cbuf, "FINISHED");
//			TextOut (hDC, 23, 14, cbuf, strlen(cbuf));
//		}
//
//		
//	SelectObject (hDC, pFont);
//	}
//	
//	
//	
//
//	oapiReleaseDC (surf, hDC);
//}


//void dm2_capsule::RedrawTgtButton(SURFHANDLE surf)
//{
//	HDC hDC = oapiGetDC (surf);
//
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"REENTRY TARGET");
//		TextOut (hDC, 65, 1, cbuf, strlen(cbuf));
//		sprintf(cbuf,"%s", tgt_name);
//		TextOut (hDC, 65, 14, cbuf, strlen(cbuf));
//	SelectObject (hDC, pFont);
//	}
//	
//	
//	
//
//	oapiReleaseDC (surf, hDC);
//}

//void dm2_capsule::RedrawNavButton(SURFHANDLE surf)
//{
//	HDC hDC = oapiGetDC (surf);
//
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//
//		SetBkMode (hDC, TRANSPARENT);
//		if(GetNavmodeState(NAVMODE_KILLROT))
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		}
//		char cbuf[256];
//		strcpy(cbuf,"KILROT");
//		TextOut (hDC, 22, 1, cbuf, strlen(cbuf));
//
//		SetBkMode (hDC, TRANSPARENT);
//		if(GetNavmodeState(NAVMODE_PROGRADE))
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		}
//		strcpy(cbuf,"PRGRD");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//
//		SetBkMode (hDC, TRANSPARENT);
//		if(GetNavmodeState(NAVMODE_RETROGRADE))
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		}
//		strcpy(cbuf,"RETRG");
//		TextOut (hDC, 106, 1, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//
//	oapiReleaseDC (surf, hDC);
//}

////void dm2_capsule::RedrawPanelButton(SURFHANDLE surf)
////{
////	HDC hDC = oapiGetDC (surf);
////
////	{
////	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
////		SetTextColor (hDC, RGB(225,255,255));
////		SetTextAlign (hDC, TA_CENTER);
////		SetBkMode (hDC, TRANSPARENT);
////		char cbuf[256];
////		strcpy(cbuf,"PANEL");
////		TextOut (hDC, 22, 1, cbuf, strlen(cbuf));
////		
////	SelectObject (hDC, pFont);
////	}
////
////	oapiReleaseDC (surf, hDC);
////}
//
//void dm2_capsule::RedrawSepButton(SURFHANDLE surf)
//{
//	HDC hDC = oapiGetDC (surf);
//
//	///////////////// trunk
//	if(tk_status == 0)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		if(umb_status < 2)
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		}
//
//		char cbuf[256];
//		strcpy(cbuf,"JETTISON TRUNK");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(125,125,125));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"NO TRUNK");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//	/////////////// payloads
//
//	char pld_name[256];
//	OBJHANDLE hpld;
//	int npatts = AttachmentCount(false);
//	if (npatts > 0)
//	{
//	for (int i = 0; i < npatts; i++)
//		{
//		ATTACHMENTHANDLE hpld_att =  GetAttachmentHandle (false, i);
//		hpld = GetAttachmentStatus(hpld_att);
//		if (hpld == NULL)continue;
//		else
//			{
//				oapiGetObjectName(hpld, pld_name,60);
//				break;
//			}
//		}
//	}
//
//	if(tk_status == 1)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(125,125,125));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"NO PAYLOAD");
//		TextOut (hDC, 64, 14, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//	else if (hpld == NULL)
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(125,125,125));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"NO PAYLOAD");
//		TextOut (hDC, 64, 14, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//	else
//	{
//		HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_LEFT);
//		SetBkMode (hDC, TRANSPARENT);
//		char cbuf[256];
//		strcpy(cbuf,"JET.PAYLOAD: ");
//		strcat(cbuf, pld_name);
//		TextOut (hDC, 2, 14, cbuf, strlen(cbuf));
//	SelectObject (hDC, pFont);
//	}
//
//
//	oapiReleaseDC (surf, hDC);
//}

////void dm2_capsule::RedrawLightButton(SURFHANDLE surf)
////{
////	HDC hDC = oapiGetDC (surf);
////
////	{
////	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
////		SetTextColor (hDC, RGB(225,255,255));
////		SetTextAlign (hDC, TA_CENTER);
////
////		SetBkMode (hDC, TRANSPARENT);
////		if(beac_al_status == 1)
////		{
////		SetBkMode (hDC, OPAQUE);
////		SetBkColor(hDC, RGB(255,0,0));
////		}
////		char cbuf[256];
////		strcpy(cbuf,"LIGHTS");
////		TextOut (hDC, 22, 1, cbuf, strlen(cbuf));
////
////		SetBkMode (hDC, TRANSPARENT);
////		if(light_status == 1)
////		{
////		SetBkMode (hDC, OPAQUE);
////		SetBkColor(hDC, RGB(255,0,0));
////		}
////		strcpy(cbuf,"SPOT");
////		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
////
////		SetBkMode (hDC, TRANSPARENT);
////		if(laser_status == 1)
////		{
////		SetBkMode (hDC, OPAQUE);
////		SetBkColor(hDC, RGB(255,0,0));
////		}
////		strcpy(cbuf,"LIDAR");
////		TextOut (hDC, 106, 1, cbuf, strlen(cbuf));
////		
////	SelectObject (hDC, pFont);
////	}
////
////	oapiReleaseDC (surf, hDC);
////}
//
//
//void dm2_capsule::RedrawChuteButton(SURFHANDLE surf)
//{
//	HDC hDC = oapiGetDC (surf);
//
//	///////////////// header
//	{
//	HFONT pFont = (HFONT)SelectObject (hDC, g_Param.font[0]);
//		SetTextColor (hDC, RGB(225,255,255));
//		SetTextAlign (hDC, TA_CENTER);
//		SetBkMode (hDC, TRANSPARENT);
//		//if(umb_status < 2)
//		//{
//		//SetBkMode (hDC, OPAQUE);
//		//SetBkColor(hDC, RGB(255,0,0));
//		//}
//
//		char cbuf[256];
//		strcpy(cbuf,"DROGUE / MAIN CHUTES");
//		TextOut (hDC, 64, 1, cbuf, strlen(cbuf));
//
//		if (drogue_deploy_status != 3)strcpy(cbuf,"DEPLOYED");
//
//		else if (chute_arm_status == 0)strcpy(cbuf,"SAFE");
//
//		else if (chute_arm_status == 1)
//		{
//		SetBkMode (hDC, OPAQUE);
//		SetBkColor(hDC, RGB(255,0,0));
//		strcpy(cbuf,"ARMED");
//		}
//
//		TextOut (hDC, 64, 14, cbuf, strlen(cbuf));
//		
//	SelectObject (hDC, pFont);
//	}
//	
//
//	oapiReleaseDC (surf, hDC);
//}


// --------------------------------------------------------------
// Respond to virtual cockpit area redraw request
// --------------------------------------------------------------
bool dm2_capsule::clbkVCRedrawEvent (int id, int event, SURFHANDLE surf)
{
	switch (id) 
	{
	case AID_CDR3_BUTTONS:
		{
		int mfd = id-AID_CDR1_BUTTONS+MFD_LEFT;
		RedrawPanel_MFDButton (surf, mfd);
		} return true;

	case AID_CDR1_BUTTONS:
		{
		int mfd = id-AID_CDR1_BUTTONS+MFD_LEFT;
		RedrawPanel_MFDButton (surf, mfd);
		} return true;

	case AID_CDR2_BUTTONS:
		{
		int mfd = id-AID_CDR1_BUTTONS+MFD_LEFT;
		RedrawPanel_MFDButton (surf, mfd);
		} return true;

	}

	return false;
}

// --------------------------------------------------------------
// Respond to MFD mode change
// --------------------------------------------------------------
void dm2_capsule::clbkMFDMode (int mfd, int mode)
{
	oapiVCTriggerRedrawArea (-1, AID_CDR1_BUTTONS+mfd-MFD_LEFT);
}

//void dm2_capsule::clbkRCSMode (int mode)
//{
//	oapiVCTriggerRedrawArea (-1, RCS_BUTTONS);
//}

//void dm2_capsule::clbkNavMode (int mode, bool active)
//{
//	oapiVCTriggerRedrawArea (-1, NAV_BUTTONS);
//}